/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 * License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#undef LOCK_DEBUG
#include <marlin-block.h>

MarlinBlock *
marlin_block_new_from_data (gpointer data,
			    guint64 num_frames,
			    int scale_factor,
			    int frame_size)
{
	MarlinBlock *block;

	g_return_val_if_fail (scale_factor != 0, NULL);
	
	block = g_new0 (MarlinBlock, 1);

	block->lock = marlin_read_write_lock_new ();
	block->num_frames = num_frames;
	block->frame_data = data;
	block->scale_factor = scale_factor;
	block->size = frame_size;

	return block;
}

void
marlin_block_free (MarlinBlock *block)
{
	g_return_if_fail (block != NULL);

	marlin_read_write_lock_destroy (block->lock);
	g_free (block);
}

void
marlin_block_free_list (MarlinBlock *block)
{
	g_return_if_fail (block != NULL);

	while (block) {
		MarlinBlock *old_block;

		old_block = block;

		block = marlin_block_next (old_block);
		marlin_block_free (old_block);
	}
}

void
marlin_block_append (MarlinBlock *a,
		     MarlinBlock *b)
{
	g_return_if_fail (a != NULL);
	g_return_if_fail (b != NULL);

	WRITE_LOCK (a->lock);
	WRITE_LOCK (b->lock);

	if (a->next != NULL) {
		b->next == a->next;
		a->next->previous = b;
	}

	a->next = b;
	b->previous = a;

	WRITE_UNLOCK (b->lock);
	WRITE_UNLOCK (a->lock);
}

MarlinBlock *
marlin_block_next (MarlinBlock *block)
{
	MarlinBlock *next;

	READ_LOCK (block->lock);
	next = block->next;
	READ_UNLOCK (block->lock);
	
	return next;
}

MarlinBlock *
marlin_block_previous (MarlinBlock *block)
{
	MarlinBlock *prev;

	READ_LOCK (block->lock);
	prev = block->previous;
	READ_UNLOCK (block->lock);

	return prev;
}

MarlinBlock *
marlin_block_last (MarlinBlock *block)
{
	while (block->next) {
		block = block->next;
	}

	return block;
}

/**
 * marlin_block_split:
 * block:
 * split_frame:
 *
 * Splits @block on @split_frame and returns the newly created block.
 * Given a block
 * 0 |--------------------| 19 (20 frames)
 * and splitting on frame 10
 * will give you 2 blocks
 * 0 |----------| 9 (10 frames)
 * 10|----------| 19 (10 frames)
 */
MarlinBlock *
marlin_block_split (MarlinBlock *block,
		    guint64 split_frame)
{
	MarlinBlock *new_block;
	guint64 frame_in_block;
	
	g_return_val_if_fail (split_frame <= block->end, NULL);

	WRITE_LOCK (block->lock);

	frame_in_block = (guint64) ((split_frame - block->start) / block->scale_factor);

	new_block = marlin_block_new_from_data (block->frame_data + (frame_in_block * block->size),
						(block->end - split_frame) + 1,
						block->scale_factor, block->size);
	new_block->start = split_frame;
	new_block->end = block->end;
	block->end = split_frame - 1;
	block->num_frames = (block->end - block->start) + 1;
	new_block->num_frames = (new_block->end - new_block->start) + 1;

	WRITE_UNLOCK (block->lock);

	return new_block;
}

MarlinBlock *
marlin_block_get_last (MarlinBlock *block)
{
	while (block->next != NULL) {
		block = block->next;
	}

	return block;
}

MarlinBlock *
marlin_block_copy (MarlinBlock *b)
{
	MarlinBlock *block;

	g_return_val_if_fail (b != NULL, NULL);
	
	block = g_new0 (MarlinBlock, 1);

	block->lock = marlin_read_write_lock_new ();
	block->num_frames = b->num_frames;
	block->frame_data = b->frame_data;
	block->scale_factor = b->scale_factor;
	block->size = b->size;

	return block;
}

MarlinBlock *
marlin_block_copy_list (MarlinBlock *first)
{
	MarlinBlock *p;
	MarlinBlock *out_list = NULL, *ol;

	ol = out_list;
	for (p = first; p; p = p->next) {
		MarlinBlock *block = marlin_block_copy (p);

		if (ol) {
			marlin_block_append (ol, block);
			
			ol = block;
		} else {
			out_list = ol = block;
		}
	}

	return out_list;
}
