/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 * License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <math.h>

#include <glib.h>
#include <glib-object.h>

#include <marlin/marlin-block.h>
#undef LOCK_DEBUG
/* #define LOCK_DEBUG 1 */
#include <marlin/marlin-channel.h>
#include <marlin/marlin-utils.h>

static char *tmp_dir = "/tmp";

struct _MMapRegion {
	gpointer address;
	size_t length;
};

/* Fwd declarations */
static void lockless_unlink_range (MarlinChannel *channel,
				   guint64 start_frame,
				   guint64 finish_frame,
				   MarlinBlock **blocks,
				   MarlinBlock **peaks,
				   MarlinUndoContext *ctxt);


GQuark
marlin_channel_error_quark (void)
{
	static GQuark quark = 0;
	if (quark == 0) {
		quark = g_quark_from_static_string ("marlin-channel-error-quark");
	}

	return quark;
}

/**
 * marlin_channel_new:
 * sample:
 * filename:
 * error:
 * 
 * Creates a new channel for @sample, using @filename as the basename for its
 * temp files.
 *
 * Returns: Newly created MarlinChannel or NULL on error.
 */
MarlinChannel *
marlin_channel_new (MarlinSample *sample,
		    const char *filename,
		    GError **error)
{
	MarlinChannel *channel;
	
	channel = g_new0 (MarlinChannel, 1);
	channel->lock = marlin_read_write_lock_new ();
	channel->sample = sample;

	channel->mmapped = TRUE;
	channel->file_region.offset = -1;
	channel->file_region.length = 0;
	channel->file_region.frames = (guint64) 0;
	
	if (channel->mmap_fd == 0) {
		char *basename;
		basename = g_path_get_basename (filename);
		channel->mmap_file = g_strdup_printf ("%s/%s.XXXXXX",
						      tmp_dir, basename);
		g_free (basename);
		
		channel->mmap_fd = mkstemp (channel->mmap_file);
/* 		g_print ("Opening %s for sample\n", channel->mmap_file); */
		if (channel->mmap_fd == -1) {
			if (error == NULL) {
				g_warning ("Error opening %s", channel->mmap_file);
			} else {
				*error = g_error_new (MARLIN_CHANNEL_ERROR,
						      MARLIN_CHANNEL_ERROR_IO,
						      _("Error opening data file '%s'\nError: '%s'"),
						      channel->mmap_file,
						      g_strerror (errno));
			}

			marlin_channel_free (channel);
			return NULL;
		}
	}

	if (channel->peak_fd == 0) {
		char *basename;
		char *filename;
		
		basename = g_path_get_basename (channel->mmap_file);
		channel->peak_file = g_strdup_printf ("%s/%s.XXXXXX",
						      tmp_dir, basename);
		g_free (basename);
		
		channel->peak_fd = mkstemp (channel->peak_file);
/* 		g_print ("Opening %s for peaks\n", channel->peak_file); */
		if (channel->peak_fd == -1) {
			if (error == NULL) {
				g_warning ("Error opening %s\n", channel->peak_file);
			} else {
				*error = g_error_new (MARLIN_CHANNEL_ERROR,
						      MARLIN_CHANNEL_ERROR_IO,
						      _("Error opening peak file '%'\nError: '%'"),
						      channel->peak_file,
						      g_strerror (errno));
			}

			marlin_channel_free (channel);
			return NULL;
		}
	}

	return channel;
}

/**
 * marlin_channel_free:
 * channel:
 *
 * Frees all resources that @channel was using.
 */
void
marlin_channel_free (MarlinChannel *channel)
{
	MarlinBlock *block;
	GList *l;
	
	g_return_if_fail (channel != NULL);

	WRITE_LOCK (channel->lock);

	if (channel->first) {
		marlin_block_free_list (channel->first);
	}

	if (channel->peak_first) {
		marlin_block_free_list (channel->peak_first);
	}

	for (l = channel->mmap_regions; l; l = l->next) {
		struct _MMapRegion *region = l->data;

		munmap (region->address, region->length);

		g_free (region);
	}

	g_list_free (channel->mmap_regions);
	
	if (channel->mmap_fd != -1) {
		close (channel->mmap_fd);
	}

	if (channel->peak_fd != -1) {
		close (channel->peak_fd);
	}

/* 	g_print ("Unlinking files %s and %s\n", channel->mmap_file, channel->peak_file); */
	unlink (channel->mmap_file);
	g_free (channel->mmap_file);	

	unlink (channel->peak_file);
	g_free (channel->peak_file);

	WRITE_UNLOCK (channel->lock);
	marlin_read_write_lock_destroy (channel->lock);
	
	g_free (channel);
}

static off_t
write_data (MarlinChannel *channel,
	    float *data,
	    guint num_frames,
	    GError **error)
{
	size_t size;
	off_t offset;
	int ret, o;

	g_return_val_if_fail (channel != NULL, -1);

	/* Find the start of where the data is being written */
	offset = lseek (channel->mmap_fd, 0, SEEK_END);

	/* FIXME: MARLIN_BLOCK_SIZE can be quite wasteful :) */
	/* Even if we have < MARLIN_BLOCK_SIZE of data, we always want
	   to write a full block out, so that we can mmap a full block
	   to save having to munmap the last block, plus it makes sure
	   that the offset is a multiple of the page size. */
	size = MARLIN_BLOCK_SIZE * sizeof (float);
	o = 0;
	
 try_to_write:
	ret = write (channel->mmap_fd, data, size);
	if (ret == -1) {
		if (error == NULL) {
			g_warning ("Error writing sample to temp file %s\n%s (%d)",
				   channel->mmap_file,
				   g_strerror (errno), errno);
		} else {
			if (errno == ENOSPC) {
				*error = g_error_new (MARLIN_CHANNEL_ERROR,
						      MARLIN_CHANNEL_ERROR_NO_SPACE,
						      _("There was not enough space for '%s'"),
						      channel->mmap_file);
			} else {
				*error = g_error_new (MARLIN_CHANNEL_ERROR,
						      MARLIN_CHANNEL_ERROR_IO,
						      _("Error writing data to '%s'\nError: %s"),
						      channel->mmap_file,
						      g_strerror (errno));
			}
		}
		
		return -1;
	}

	if (ret != size) {
		size = size - ret;
		data += (ret / sizeof (float));

		goto try_to_write;
	}
	

	channel->frames += (guint64) num_frames;
	
	channel->mmapped = FALSE; /* We've written some data that hasn't
				     been mmapped */

	if (channel->file_region.offset == -1) {
		channel->file_region.offset = offset;
	}
	channel->file_region.length += size;
	channel->file_region.frames += num_frames;

	return offset;
}

/* Split the whole length into MARLIN_BLOCK_SIZE chunks */
static off_t
write_data_full (MarlinChannel *channel,
		 float *data,
		 guint64 num_frames,
		 GError **error)
{
	off_t offset = -1;
	float *buffer, *p;
	
	buffer = g_new0 (float, MARLIN_BLOCK_SIZE);

	p = data;
	while (num_frames > (guint64) 0) {
		guint64 len_to_copy;
		off_t o;
		
		len_to_copy = MIN (MARLIN_BLOCK_SIZE, num_frames);
		memcpy (buffer, p, len_to_copy * sizeof (float));

		o = write_data (channel, buffer, len_to_copy, error);
		if (o == -1) {
			g_free (buffer);
			return -1;
		}

		if (offset == -1) {
			offset = o;
		}
		
		/* Reset for next time around */
		memset (buffer, 0, MARLIN_BLOCK_SIZE * sizeof (float));
		p += len_to_copy;
		num_frames -= len_to_copy;
	}

	g_free (buffer);

	return offset;
}

/**
 * marlin_channel_write_data:
 * channel:
 * data:
 * num_frames:
 * error:
 *
 * Writes @num_frames starting at @data into @channel.
 *
 * Returns: Offset in the file at which @data was writen, -1 on failure
 * with details in @error.
 */
off_t
marlin_channel_write_data (MarlinChannel *channel,
			   float *data,
			   guint num_frames,
			   GError **error)
{
	off_t offset;
	
	g_return_val_if_fail (channel != NULL, -1);

	WRITE_LOCK (channel->lock);

	offset = write_data_full (channel, data, num_frames, error);
	
	WRITE_UNLOCK (channel->lock);

	return offset;
}

#define MAX_MMAP_SIZE 134217728
static struct _MMapRegion *
mmap_section (int fd,
	      size_t length,
	      off_t offset,
	      GError **error)
{
	static int mmap_count = 0;
	gpointer *mmap_mem;
	struct _MMapRegion *region;

	mmap_mem = mmap (0, length,
			 PROT_READ | PROT_WRITE,
			 MAP_SHARED, fd,
			 offset);
	if (mmap_mem == MAP_FAILED) {
		if (error == NULL) {
			g_warning ("Error trying to mmap data\n%s",
				   g_strerror (errno));
		} else {
			*error = g_error_new (MARLIN_CHANNEL_ERROR,
					      MARLIN_CHANNEL_ERROR_MEMORY,
					      _("Failed to allocate memory.\nError: '%s'"),
					      g_strerror (errno));
		}
		
		return NULL;
	}

	region = g_new (struct _MMapRegion, 1);
	region->address = mmap_mem;
	region->length = length;

	return region;
}

/* FIXME: This needs to be able to handle any block,
   not just the first in a list */
static void
recalculate_block_ranges (MarlinBlock *start)
{
	guint64 last_frame;

	last_frame = (guint64) 0;
	
	while (start) {
		start->start = last_frame;
		start->end = start->start + start->num_frames - 1;

		last_frame = start->end + 1;
		start = start->next;
	}
}

static gboolean
lockless_ensure_mmapped (MarlinChannel *channel,
			 MarlinOperation *operation,
			 GError **error)
{
	struct _MMapRegion *region;
	MarlinBlock *block;
	size_t size;
	off_t offset;
	guint64 mmapped_frames = 0, unmapped_frames, frame_block, length;
	
	if (channel->mmapped) {
		return TRUE;
	}

	unmapped_frames = channel->file_region.frames;
	offset = channel->file_region.offset;
/* 	length = channel->file_region.length; */

	length = unmapped_frames;
	while (length > 0) {
		if ((length * sizeof (float))> MAX_MMAP_SIZE) {
			size = MAX_MMAP_SIZE;
		} else {
			size = length * sizeof (float);
		}

		frame_block = size / sizeof (float);
		
		region = mmap_section (channel->mmap_fd, size, offset, error);
		if (region == NULL) {
			return FALSE;
		}

		channel->mmap_regions = g_list_append (channel->mmap_regions, region);
		
		block = marlin_block_new_from_data (region->address, frame_block, 1, sizeof (float));
		
		if (channel->last == NULL) {
			channel->last = block;
			channel->first = block;
			block->start = 0;
			block->end = frame_block;
		} else {
			channel->last->next = block;
			block->previous = channel->last;
			channel->last = block;
			block->start = block->previous->end + 1;
			block->end = block->start + frame_block;
		}
		
		mmapped_frames += frame_block;
		length = (unmapped_frames - mmapped_frames);
		offset += size;
	}

	channel->mmapped = TRUE;
	
	/* Now the file is all mmapped, we need to zero the file regions */
	channel->file_region.length = 0;
	channel->file_region.offset = -1;
	channel->file_region.frames = (guint64) 0;

	return TRUE;
}

/**
 * marlin_channel_ensure_mmapped:
 * channel:
 * operation:
 * error:
 *
 * Maps all the unmaped data in @channel into memory.
 * This needs to be called after data has been writen to the disk
 * so that it will be used.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_ensure_mmapped (MarlinChannel *channel,
			       MarlinOperation *operation,
			       GError **error)
{
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);

 	ret = lockless_ensure_mmapped (channel, operation, error);
	
	WRITE_UNLOCK (channel->lock);

	return ret;
}

/* Channel lock must be held before calling this function */
static MarlinBlock *
get_for_frame (MarlinBlock *block,
	       guint64 frame)
{
	/* Yes, this is even slower */
	for (; block; block = marlin_block_next (block)) {
		if (frame >= block->start && frame <= block->end) {
			return block;
		}
	}

	return NULL;
}

/**
 * marlin_channel_get_block_for_frame:
 * channel:
 * frame:
 *
 * Gets the block from @channel that @frame is found in.
 *
 * Returns: MarlinBlock or NULL on failure.
 */
MarlinBlock *
marlin_channel_get_block_for_frame (MarlinChannel *channel,
				    guint64 frame)
{
	MarlinBlock *block;
	g_return_val_if_fail (channel != NULL, NULL);

	READ_LOCK (channel->lock);

	if (frame == channel->frames) {
		MarlinBlock *b;

		b = channel->last;
		READ_UNLOCK (channel->lock);
		
		return b;
	}

	if (frame == (guint64) 0) {
		MarlinBlock *b;

		b = channel->first;
		READ_UNLOCK (channel->lock);

		return b;
	}
	
	block = get_for_frame (channel->first, frame);

	READ_UNLOCK (channel->lock);
		
	return block;
}

/**
 * marlin_channel_get_peak_for_frame:
 * channel:
 * frame:
 *
 * Gets the peak block from @channel that @frame is found in.
 *
 * Returns: MarlinBlock or NULL on failure.
 */
MarlinBlock *
marlin_channel_get_peak_for_frame (MarlinChannel *channel,
				   guint64 frame)
{
	MarlinBlock *block;
	
	g_return_val_if_fail (channel != NULL, NULL);

	READ_LOCK (channel->lock);

	if (frame == channel->frames) {
		MarlinBlock *b;

		b = channel->peak_last;
		READ_UNLOCK (channel->lock);
		
		return b;
	}

	if (frame == (guint64) 0) {
		MarlinBlock *b;

		b = channel->peak_first;
		READ_UNLOCK (channel->lock);

		return b;
	}
	
	block = get_for_frame (channel->peak_first, frame);

	READ_UNLOCK (channel->lock);

	return block;
}

#define PEAKS_PER_WRITE 256

/* Should hold a WRITE lock on the channel
   before calling */
static MarlinBlock *
generate_peak_for_block (MarlinChannel *channel,
			 MarlinBlock *block,
			 MarlinOperation *operation,
			 GError **error)
{
	MarlinPeak peak[PEAKS_PER_WRITE];
	MarlinBlock *new_block;
	struct _MMapRegion *region;
	guint64 total, p, num_peaks;
	float *frame_data;
	int i, ret, peak_num;
	size_t size;
	off_t offset;
	char *desc;
	int percent, last_percent;
	
	last_percent = 0;

	READ_LOCK (block->lock);

	offset = lseek (channel->peak_fd, 0, SEEK_END);
	num_peaks = block->num_frames / (guint64) MARLIN_FRAMES_PER_PEAK;
	if (block->num_frames % (guint64) MARLIN_FRAMES_PER_PEAK != 0) {
		num_peaks++;
	}
	
	total = (guint64) 0;
	frame_data = (float *) block->frame_data;
	memset (peak, 0, sizeof (MarlinPeak) * PEAKS_PER_WRITE);
	for (p = (guint64) 0, peak_num = 0; p < num_peaks; p++) {
		int j, num_frames;

		num_frames = 0;

		for (j = 0; j < MARLIN_FRAMES_PER_PEAK && total < block->num_frames;
		     j++, total++) {
			guint64 total_in_block;
			float d;

			d = frame_data[total];
			
			if (d < -1.0 || d > 1.0) {
/* 				g_print ("d: %f\n", d); */
				d = frame_data[total] = 0.0;
			}
			
			if (d < 0.0) {
				peak[peak_num].avg_low = peak[peak_num].avg_low + d;
			} else {
				peak[peak_num].avg_high = peak[peak_num].avg_high + d;
			}
			
			peak[peak_num].high = MAX (peak[peak_num].high, d);
			peak[peak_num].low = MIN (peak[peak_num].low, d);

			num_frames++;
		}
		
		peak[peak_num].avg_low /= num_frames;
		peak[peak_num].avg_high /= num_frames;
		
		peak_num++;
		if (peak_num == PEAKS_PER_WRITE || p == num_peaks - (guint64) 1) {
			/* Write out this block */
/* 			offset = lseek (channel->peak_fd, 0, SEEK_CUR); */
/* 			size = sizeof (MarlinPeak) * peak_num; */
			size = sizeof (MarlinPeak) * PEAKS_PER_WRITE;

			ret = write (channel->peak_fd, peak, size);
			if (ret == -1 || ret != size) {
				if (error == NULL) {
					g_warning ("Error writing to peak file %s\n%s (%d)",
						   channel->peak_file,
						   g_strerror (errno), errno);
				} else {

					/* Ewww */
					if (errno = -ENOSPC) {
						*error = g_error_new (MARLIN_CHANNEL_ERROR,
								      MARLIN_CHANNEL_ERROR_NO_SPACE,
								      _("There was not enough space for '%s'"),
								      channel->mmap_file);
					} else {
						*error = g_error_new (MARLIN_CHANNEL_ERROR,
								      MARLIN_CHANNEL_ERROR_IO,
								      _("Error writing data to '%'\nError: %s"),
								      channel->mmap_file,
								      g_strerror (errno));
					}
					
				}
				
				READ_UNLOCK (block->lock);
				return NULL;
			}

			if (operation) {
				percent = (int) ((total * 100) / channel->frames);
				/* This is limited to firing every 10% because
				   otherwise g_signal_emit deadlocks... */
				if (percent != last_percent && percent % 10 == 0) {
					marlin_operation_progress (operation, percent);
					last_percent = percent;
				}
			}

			peak_num = 0;
			memset (peak, 0, sizeof (MarlinPeak) * PEAKS_PER_WRITE);
		}
	}

	/* MMap this data in */
	region = mmap_section (channel->peak_fd, num_peaks * sizeof (MarlinPeak),
			       offset, error);
	if (region == NULL) {
		READ_UNLOCK (block->lock);
		return NULL;
	}
	
	channel->mmap_regions = g_list_append (channel->mmap_regions, region);

	new_block = marlin_block_new_from_data (region->address, block->num_frames,
						MARLIN_FRAMES_PER_PEAK, sizeof (MarlinPeak));

	new_block->start = block->start;
	new_block->end = block->end;
	READ_UNLOCK (block->lock);

	return new_block;
}

static gboolean
lockless_generate_peaks (MarlinChannel *channel,
			 MarlinOperation *operation,
			 GError **error)
{
	MarlinBlock *block;

	for (block = channel->first; block;) {
		MarlinBlock *peak;

		peak = generate_peak_for_block (channel, block, operation, error);
		if (peak == NULL) {
			return FALSE;
		}
		
		/* Append peak */
		if (channel->peak_last == NULL) {
			channel->peak_last = peak;
			channel->peak_first = peak;
		} else {
			channel->peak_last->next = peak;
			peak->previous = channel->peak_last;
			channel->peak_last = peak;
		}

		block = block->next;
	}

	return TRUE;
}
/**
 * marlin_channel_generate_peaks:
 * channel:
 * operation:
 * error:
 *
 * Generates the peaks for the data in @channel.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_generate_peaks (MarlinChannel *channel,
			       MarlinOperation *operation,
			       GError **error)
{
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

#if 0
	if (operation) {
		g_object_set (G_OBJECT (operation),
			      "description", "Generating peak data",
			      NULL);
		
		marlin_operation_progress (operation, 0);
	}
#endif
	WRITE_LOCK (channel->lock);

	ret = lockless_generate_peaks (channel, operation, error);
	
	WRITE_UNLOCK (channel->lock);

#if 0
	if (operation) {
		g_object_set (G_OBJECT (operation),
			      "description", "",
			      NULL);
	}
#endif
	return ret;
}

/*
 * splits the channel at split_frame
 * 0 |--------------| 20
 * Split at 10
 * 0 |----| 9 10 |------| 20
 */
static void
lockless_split_block (MarlinChannel *channel,
		      guint64 split_frame)
{
	MarlinBlock *first, *second;
	MarlinBlock *pf, *ps;

	g_return_if_fail (split_frame <= channel->frames);

	if (split_frame == channel->frames || split_frame == 0) {
		return;
	}

	first = get_for_frame (channel->first, split_frame);
	g_return_if_fail (first != NULL);

	if (split_frame != first->start) {

		second = marlin_block_split (first, split_frame);
		
		/* Insert second into the list after first */
		if (first->next == NULL) {
			first->next = second;
			second->previous = first;
			second->next = NULL;
			
			/* If first->next was NULL,
			   then first was the last block
			   Set the channel last point to second */
			channel->last = second;
		} else {
			second->next = first->next;
			second->next->previous = second;
			second->previous = first;
			first->next = second;
		}
	}

	pf = get_for_frame (channel->peak_first, split_frame);
	if (split_frame == pf->start) {

		/* Don't need to split or regenerate */
		return;
	}

	ps = marlin_block_split (pf, split_frame);
	/* Insert ps into list after pf */
	if (pf->next == NULL) {
		pf->next = ps;
		ps->previous = pf;
		
		channel->peak_last = ps;
	} else {
		ps->next = pf->next;
		ps->next->previous = ps;
		ps->previous = pf;
		pf->next = ps;
	}
}

/**
 * marlin_channel_split_block:
 * @channel:
 * @split_frame:
 *
 * Splits the block in @channel located at @split_frame into 2 separate blocks.
 */
void
marlin_channel_split_block (MarlinChannel *channel,
			    guint64 split_frame)
{
	g_return_if_fail (channel != NULL);
	
	WRITE_LOCK (channel->lock);
	lockless_split_block (channel, split_frame);
	WRITE_UNLOCK (channel->lock);
}

struct _unlink_closure {
	MarlinChannel *channel;
	MarlinBlock *original, *original_peaks;
	guint64 start_frame, end_frame;
};

static void
unlink_undo (gpointer data)
{
	struct _unlink_closure *c = data;
	MarlinBlock *prev_block, *prev_peak;
	MarlinBlock *next_block, *next_peak;
	MarlinBlock *first_block, *first_peak;
	MarlinBlock *last_block, *last_peak;

	WRITE_LOCK (c->channel->lock);

	if (c->channel->first == NULL) {
		c->channel->first = marlin_block_copy_list (c->original);
		c->channel->peak_first = marlin_block_copy_list (c->original_peaks);
		c->channel->last = marlin_block_last (c->channel->first);
		c->channel->peak_last = marlin_block_last (c->channel->peak_first);
	} else {
		prev_block = get_for_frame (c->channel->first, c->start_frame - 1);
		first_block = marlin_block_copy_list (c->original);
		last_block = marlin_block_last (first_block);

		if (prev_block == NULL) {
			next_block = c->channel->first;
			c->channel->first = first_block;
			first_block->previous = NULL;
		} else {
			next_block = prev_block->next;
			prev_block->next = first_block;
			first_block->previous = prev_block;
		}

		last_block->next = next_block;
		if (next_block) {
			next_block->previous = last_block;
		}
		
		prev_peak = get_for_frame (c->channel->peak_first, c->start_frame - 1);
		first_peak = marlin_block_copy_list (c->original_peaks);
		last_peak = marlin_block_last (first_peak);

		if (prev_peak == NULL) {
			next_peak = c->channel->peak_first;
			c->channel->peak_first = first_peak;
			first_peak->previous = NULL;
		} else {
			next_peak = prev_peak->next;
			prev_peak->next = first_peak;
			first_peak->previous = prev_peak;
		}

		last_peak->next = next_peak;
		if (next_peak) {
			next_peak->previous = last_peak;
		}
	}

	c->channel->frames += ((c->end_frame - c->start_frame) + 1);

	recalculate_block_ranges (c->channel->first);
	recalculate_block_ranges (c->channel->peak_first);

	WRITE_UNLOCK (c->channel->lock);
}

static void
unlink_redo (gpointer data)
{
	struct _unlink_closure *c = data;
	MarlinBlock *b, *p;
	
	WRITE_LOCK (c->channel->lock);
	lockless_unlink_range (c->channel, c->start_frame, c->end_frame,
			       &b, &p, NULL);
	WRITE_UNLOCK (c->channel->lock);

	marlin_block_free_list (b);
	marlin_block_free_list (p);
}

static void
unlink_destroy (gpointer data)
{
	struct _unlink_closure *c = data;

	marlin_block_free_list (c->original);
	marlin_block_free_list (c->original_peaks);
	g_free (c);
}

/* Removes data between start_frame and finish_frame inclusive
   by unlinking the blocks that contain it, and then returns
   the lists of removed blocks and peaks in &blocks and &peaks. */
static void
lockless_unlink_range (MarlinChannel *channel,
		       guint64 start_frame,
		       guint64 finish_frame,
		       MarlinBlock **blocks,
		       MarlinBlock **peaks,
		       MarlinUndoContext *ctxt)
{
	MarlinBlock *first_block, *last_block;
	MarlinBlock *first_peak, *last_peak;

	g_return_if_fail (start_frame <= channel->frames);
	g_return_if_fail (finish_frame <= channel->frames);
	g_return_if_fail (finish_frame >= start_frame);

	/* Split on the start and finish frames */
	if (finish_frame != channel->frames - 1) {
		lockless_split_block (channel, finish_frame + 1);
	}
	lockless_split_block (channel, start_frame);

	first_block = get_for_frame (channel->first, start_frame);
	g_assert (first_block != NULL);

	last_block = get_for_frame (channel->first, finish_frame);
	g_assert (last_block != NULL);

	first_peak = get_for_frame (channel->peak_first, start_frame);
	g_assert (first_peak != NULL);

	last_peak = get_for_frame (channel->peak_first, finish_frame);
	g_assert (last_peak != NULL);

	/* Unlink these blocks */
	if (channel->first == first_block) {
		channel->first = marlin_block_next (last_block);
	}

	if (channel->last == last_block) {
		channel->last = marlin_block_previous (first_block);
	}

	if (channel->peak_first == first_peak) {
		channel->peak_first = marlin_block_next (last_peak);
	}

	if (channel->peak_last == last_peak) {
		channel->peak_last = marlin_block_previous (first_peak);
	}

	/* Need to lock the blocks */
	if (first_block->previous != NULL) {
		first_block->previous->next = last_block->next;
	}
	
	if (last_block->next != NULL) {
		last_block->next->previous = first_block->previous;
	}
	
	if (first_peak->previous != NULL) {
		first_peak->previous->next = last_peak->next;
	}
	
	if (last_peak->next != NULL) {
		last_peak->next->previous = first_peak->previous;
	}

	last_block->next = NULL;
	last_peak->next = NULL;

	channel->frames -= ((finish_frame - start_frame) + 1);
	/* Need to unlock the blocks */

	recalculate_block_ranges (channel->first);
	recalculate_block_ranges (channel->peak_first);

	if (blocks) {
		*blocks = first_block;
	}

	if (peaks) {
		*peaks = first_peak;
	}

	if (ctxt) {
		MarlinUndoable *u;
		struct _unlink_closure *c;

		c = g_new (struct _unlink_closure, 1);
		c->channel = channel;
		c->original = marlin_block_copy_list (first_block);
		c->original_peaks = marlin_block_copy_list (first_peak);
		c->start_frame = start_frame;
		c->end_frame = finish_frame;

		u = marlin_undoable_new (unlink_undo,
					 unlink_redo,
					 unlink_destroy,
					 c);
		marlin_undo_context_add (ctxt, u);
	}
}

/* Deletes between start_frame and finish_frame inclusive */
static void
lockless_delete_range (MarlinChannel *channel,
		       guint64 start_frame,
		       guint64 finish_frame,
		       MarlinUndoContext *ctxt)
{
	MarlinBlock *blocks, *peaks;

	lockless_unlink_range (channel, start_frame, finish_frame, 
			       &blocks, &peaks, ctxt);
	marlin_block_free_list (blocks);
	marlin_block_free_list (peaks);
}

/* Hold write lock for channel
   Block should be read locked */
/* FIXME: I'm not sure if the whole block list thing is needed... */
static MarlinBlock *
copy_data_into_channel (MarlinChannel *channel,
			float *data,
			guint64 num_frames,
			GError **error)
{
	MarlinBlock *dest_list = NULL, *end = NULL;
	off_t offset;
	struct _MMapRegion *region;
	size_t size;
	guint64 mmapped_frames = 0, unmapped_frames, length;
	
	offset = write_data_full (channel, data, num_frames, error);
	if (offset == -1) {
		return NULL;
	}

	unmapped_frames = channel->file_region.frames;
	length = unmapped_frames;

	while (length > 0) {
		MarlinBlock *block;
		
		if ((length * sizeof (float)) > MAX_MMAP_SIZE) {
			size = MAX_MMAP_SIZE;
		} else {
			size = length * sizeof (float);
		}

		region = mmap_section (channel->mmap_fd, size, offset, error);
		if (region == NULL) {
			return NULL;
		}

		channel->mmap_regions = g_list_append (channel->mmap_regions, region);
		block = marlin_block_new_from_data (region->address, size / sizeof (float), 1, sizeof (float));

		if (dest_list == NULL) {
			dest_list = block;
			end = block;
		} else {
			end->next = block;
			block->previous = end;
			end = block;
		}

		mmapped_frames += (size / sizeof (float));
		length = (unmapped_frames - mmapped_frames);
		offset += size;
	}

	channel->mmapped = TRUE;
	channel->file_region.length = 0;
	channel->file_region.offset = -1;
	channel->file_region.frames = (guint64) 0;

	return dest_list;
}

static MarlinBlock *
copy_block_into_channel (MarlinChannel *channel,
			 MarlinBlock *src,
			 GError **error)
{
	return copy_data_into_channel (channel, (float *) src->frame_data,
				       src->num_frames, error);
}

struct _insert_data_closure {
	MarlinChannel *channel;
	MarlinBlock *blocks, *peaks;
	guint64 insert_frame;
	guint64 num_frames;
};

static void
insert_data_undo (gpointer data)
{
	struct _insert_data_closure *c = data;

	WRITE_LOCK (c->channel->lock);
	lockless_unlink_range (c->channel, c->insert_frame,
			       c->insert_frame + (c->num_frames - 1),
			       NULL, NULL, NULL);
	WRITE_UNLOCK (c->channel->lock);
}

static void
insert_data_redo (gpointer data)
{
	struct _insert_data_closure *c = data;
	MarlinBlock *prev_block, *prev_peak;
	MarlinBlock *next_block, *next_peak;
	MarlinBlock *first_block, *first_peak;
	MarlinBlock *last_block, *last_peak;
	
	WRITE_LOCK (c->channel->lock);

	if (c->channel->first == NULL) {
		c->channel->first = marlin_block_copy_list (c->blocks);
		c->channel->peak_first = marlin_block_copy_list (c->peaks);
		c->channel->last = marlin_block_last (c->channel->first);
		c->channel->peak_last = marlin_block_last (c->channel->peak_first);
	} else {
		prev_block = get_for_frame (c->channel->first, c->insert_frame - 1);
		first_block = marlin_block_copy_list (c->blocks);
		last_block = marlin_block_last (first_block);

		if (prev_block == NULL) {
			next_block = c->channel->first;
			c->channel->first = first_block;
			first_block->previous = NULL;
		} else {
			next_block = prev_block->next;
			prev_block->next = first_block;
			first_block->previous = prev_block;
		}

		last_block->next = next_block;
		if (next_block) {
			next_block->previous = last_block;
		}
		
		prev_peak = get_for_frame (c->channel->peak_first, c->insert_frame - 1);
		first_peak = marlin_block_copy_list (c->peaks);
		last_peak = marlin_block_last (first_peak);

		if (prev_peak == NULL) {
			next_peak = c->channel->peak_first;
			c->channel->peak_first = first_peak;
			first_peak->previous = NULL;
		} else {
			next_peak = prev_peak->next;
			prev_peak->next = first_peak;
			first_peak->previous = prev_peak;
		}

		last_peak->next = next_peak;
		if (next_peak) {
			next_peak->previous = last_peak;
		}
	}

	c->channel->frames += c->num_frames;

	recalculate_block_ranges (c->channel->first);
	recalculate_block_ranges (c->channel->peak_first);

	WRITE_UNLOCK (c->channel->lock);	
}

static void
insert_data_destroy (gpointer data)
{
	struct _insert_data_closure *c = data;

	marlin_block_free_list (c->blocks);
	marlin_block_free_list (c->peaks);
	g_free (c);
}

/* Splits the block on the insert_frame, so insert_frame will be first in
   new block. Then it gets the block for insert_frame - 1, and then inserts 
   data */
static gboolean
lockless_insert_data (MarlinChannel *channel,
		      float *data,
		      guint64 num_frames,
		      guint64 insert_frame,
		      MarlinUndoContext *ctxt,
		      GError **error)
{
	MarlinBlock *prev, *next;
	MarlinBlock *p_peak, *n_peak;
	MarlinBlock *block_list, *start, *end, *b;
	MarlinBlock *ps = NULL, *pe = NULL;
	struct _insert_data_closure *c = NULL;
	MarlinUndoable *u;
	guint64 frames;
	
	g_return_val_if_fail (insert_frame <= channel->frames, FALSE);

	/* FIXME: copy_data_into_channel adjusts the frame count
	   but the frames haven't actually been put into the
	   channel, they've just been writen to disk. */
	frames = channel->frames;
	block_list = copy_data_into_channel (channel, data, num_frames, error);
	if (block_list == NULL) {
		return FALSE;
	}

	start = block_list;
	end = marlin_block_get_last (block_list);

	if (ctxt) {
		c = g_new (struct _insert_data_closure, 1);
		c->channel = channel;
		c->num_frames = num_frames;
		c->insert_frame = insert_frame;
		c->blocks = marlin_block_copy_list (start);
	}

	if (channel->first == NULL) {
		channel->first = start;
		channel->last = end;
	} else {
		if (insert_frame != frames && insert_frame != 0) {
			lockless_split_block (channel, insert_frame);
		}

		if (insert_frame != 0) {
			prev = get_for_frame (channel->first, insert_frame - 1);
			g_assert (prev != NULL);
			
			next = prev->next;
			
			WRITE_LOCK (prev->lock);
			
			prev->next = start;
			start->previous = prev;
			WRITE_UNLOCK (prev->lock);
		} else {
			next = channel->first;
			channel->first = start;
		}

		end->next = next;
		if (next != NULL) {
			WRITE_LOCK (next->lock);
			next->previous = end;
			WRITE_UNLOCK (next->lock);
		} else {
			channel->last = end;
		}
	}

	recalculate_block_ranges (channel->first);

	/* Generate a list of peak blocks to be inserted between
	   p_peak and n_peak */
	b = start;
	while (b != end->next) {
		MarlinBlock *peak;

		peak = generate_peak_for_block (channel, b, NULL, error);
		if (peak == NULL) {
			return FALSE;
		}

		if (ps == NULL) {
			ps = peak;
			pe = peak;
		} else {
			pe->next = peak;
			peak->previous = pe;
			pe = peak;
		}

		b = b->next;
	}

	if (ctxt) {
		MarlinUndoable *u;

		/* Finish setting up the closure */
		c->peaks = marlin_block_copy_list (ps);

		u = marlin_undoable_new (insert_data_undo,
					 insert_data_redo,
					 insert_data_destroy,
					 c);

		marlin_undo_context_add (ctxt, u);
	}

	if (channel->peak_first == NULL) {
		channel->peak_first = ps;
		channel->peak_last = pe;
	} else {
		if (insert_frame != 0) {
			p_peak = get_for_frame (channel->peak_first, insert_frame - 1);
			g_assert (p_peak != NULL);
			
			n_peak = p_peak->next;
		
			WRITE_LOCK (p_peak->lock);
		
			p_peak->next = ps;
			ps->previous = p_peak;
			WRITE_UNLOCK (p_peak->lock);
		} else {
			n_peak = channel->peak_first;
			channel->peak_first = ps;
		}

		pe->next = n_peak;
		if (n_peak != NULL) {
			WRITE_LOCK (n_peak->lock);
			n_peak->previous = pe;
			WRITE_UNLOCK (n_peak->lock);
		} else {
			channel->peak_last = pe;
		}
	}
	recalculate_block_ranges (channel->peak_first);

	return TRUE;
}

/**
 * marlin_channel_insert_data:
 * channel:
 * data:
 * num_frames:
 * insert_frame:
 * error:
 *
 * Inserts @num_frames pointed to by @data into @channel starting at
 * @insert_frame.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_insert_data (MarlinChannel *channel,
			    float *data,
			    guint64 num_frames,
			    guint64 insert_frame,
			    MarlinUndoContext *ctxt,
			    GError **error)
{
	gboolean ret;

	WRITE_LOCK (channel->lock);

	ret = lockless_insert_data (channel, data, num_frames, insert_frame, ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_insert_block:
 * channel:
 * block:
 * insert_frame:
 * error:
 *
 * Inserts the data in @block into @channel starting at @insert_frame.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_insert_block (MarlinChannel *channel,
			     MarlinBlock *block,
			     guint64 insert_frame,
			     MarlinUndoContext *ctxt,
			     GError **error)
{
	gboolean ret;

	WRITE_LOCK (channel->lock);

	READ_LOCK (block->lock);
	ret = lockless_insert_data (channel, (float *) block->frame_data, block->num_frames, insert_frame, ctxt, error);
	READ_UNLOCK (block->lock);
	
	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_delete_range:
 * channel:
 * start_frame:
 * finish_frame:
 *
 * Deletes the data between @start_frame and @finish_frame from @channel.
 */
void
marlin_channel_delete_range (MarlinChannel *channel,
			     guint64 start_frame,
			     guint64 finish_frame,
			     MarlinUndoContext *ctxt)
{
	g_return_if_fail (channel != NULL);

	WRITE_LOCK (channel->lock);
	lockless_delete_range (channel, start_frame, finish_frame, ctxt);
	WRITE_UNLOCK (channel->lock);
}

static gboolean
lockless_insert_silence (MarlinChannel *channel,
			 guint64 insert_frame,
			 guint64 num_frames,
			 MarlinUndoContext *ctxt,
			 GError **error)
{
	float *data;
	guint64 length;
	
	g_return_val_if_fail (insert_frame <= channel->frames, FALSE);

	if (num_frames == 0) {
		return TRUE;
	}

	/* Make a space for the 0 */
	data = g_new0 (float, MARLIN_BLOCK_SIZE);

	/* Make blocks out of the silence */
	length = num_frames;
	while (length > (guint64) 0) {
		guint64 len_to_copy;
		MarlinBlock *s_block;
		gboolean ret;

		len_to_copy = MIN (MARLIN_BLOCK_SIZE, length);
	
		s_block = marlin_block_new_from_data (data, len_to_copy, 1, sizeof (float));

		ret = lockless_insert_data (channel, (float *) s_block->frame_data, s_block->num_frames, insert_frame, ctxt, error);
		if (ret == FALSE) {
			return FALSE;
		}

		insert_frame += s_block->num_frames;
		length -= len_to_copy;

		marlin_block_free (s_block);
	}

	g_free (data);
}

/**
 * marlin_channel_insert_silence:
 * channel:
 * insert_frame:
 * num_frames:
 * ctxt:
 * error:
 *
 * Inserts @num_frames of silence into @channel at @insert_frame.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_insert_silence (MarlinChannel *channel,
			       guint64 insert_frame,
			       guint64 num_frames,
			       MarlinUndoContext *ctxt,
			       GError **error)
{
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);
	ret = lockless_insert_silence (channel, insert_frame, num_frames, ctxt, error);
	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_clear_range:
 * channel:
 * start_frame:
 * end_frame:
 * error:
 *
 * Clears the @channel between @start_frame and @end_frame
 * In reality it deletes the range and then adds silence
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_clear_range (MarlinChannel *channel,
			    guint64 start_frame,
			    guint64 finish_frame,
			    MarlinUndoContext *ctxt,
			    GError **error)
{
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);

	lockless_delete_range (channel, start_frame, finish_frame, ctxt);
	ret = lockless_insert_silence (channel, start_frame, 
				       (finish_frame - start_frame) + 1,
				       ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_insert:
 * dest:
 * src:
 * start_frame:
 * error:
 *
 * Inserts the data in @src into @dest at @start_frame.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_insert (MarlinChannel *dest,
		       MarlinChannel *src,
		       guint64 start_frame,
		       MarlinUndoContext *ctxt,
		       GError **error)
{
	MarlinBlock *block;
	guint64 start = start_frame;
	gboolean ret = TRUE;
	
	g_return_val_if_fail (src != NULL, FALSE);
	g_return_val_if_fail (dest != NULL, FALSE);

	READ_LOCK (src->lock);
	WRITE_LOCK (dest->lock);

	block = src->first;
	/* FIXME: Lock the blocks */
	while (block != NULL) {
		/* FIXME: We can speed this up a lot
		   1: Special case it so that we copy all the blocks into
		   a few larger blocks
		   2: Insert these blocks
		   3: Only update the peaks etc after all the blocks have
		   been added
		*/
		ret = lockless_insert_data (dest, (float *) block->frame_data, block->num_frames, start, ctxt, error);
		if (ret == FALSE) {
			goto err;
		}
	
		start += block->num_frames;
		block = block->next;
	}

 err:
	WRITE_UNLOCK (dest->lock);
	READ_UNLOCK (src->lock);

	return ret;
}

/**
 * marlin_channel_crop_range:
 * channel:
 * start_frame:
 * finish_frame:
 *
 * Crops the area defined by @start_frame and @finish_frame (inclusive)
 * from @channel.
 */
void
marlin_channel_crop_range (MarlinChannel *channel,
			   guint64 start_frame,
			   guint64 finish_frame,
			   MarlinUndoContext *ctxt) 
{
	MarlinBlock *first_block, *last_block;
	MarlinBlock *first_peak, *last_peak;
	MarlinBlock *current;

	g_return_if_fail (start_frame <= channel->frames);
	g_return_if_fail (finish_frame <= channel->frames);
	g_return_if_fail (finish_frame >= start_frame);

	WRITE_LOCK (channel->lock);

	if (finish_frame + 1 < channel->frames) {
		lockless_delete_range (channel, finish_frame + 1, channel->frames - 1, ctxt);
	}

	if (start_frame >= (guint64) 1) {
		lockless_delete_range (channel, (guint64) 0, start_frame - 1, ctxt);
	}
	
	WRITE_UNLOCK (channel->lock);
}

/**
 * marlin_channel_copy_data:
 * src_channel:
 * dest_channel:
 * start_frame:
 * end_frame:
 * error:
 *
 * Copy the data between @start_frame and @end_frame (inclusive) from 
 * @dest_channel into @src_channel.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_copy_data (MarlinChannel *src_channel,
			  MarlinChannel *dest_channel,
			  guint64 start_frame,
			  guint64 end_frame,
			  GError **error)
{
	MarlinBlock *src_block;
	guint64 frames_needed;
	float *buf;
	gboolean ret;
	off_t offset;
	
	g_return_val_if_fail (src_channel != dest_channel, FALSE);

	if (end_frame >= src_channel->frames) {
		end_frame = src_channel->frames - 1;
	}

	frames_needed = (end_frame - start_frame) + 1;
	
	READ_LOCK (src_channel->lock);
	WRITE_LOCK (dest_channel->lock);
	
	src_block = get_for_frame (src_channel->first, start_frame);
	g_assert (src_block);

	buf = g_new0 (float, MARLIN_BLOCK_SIZE);

	READ_LOCK (src_block->lock);
	
	while (frames_needed > (guint64) 0) {
		guint64 length_got;

		length_got = (guint64) 0;
		
		while (length_got < MARLIN_BLOCK_SIZE &&
		       frames_needed > (guint64) 0) {
			guint64 buf_needed, length, pos_in_block;
			guint64 length_needed, left_in_block;
			
			float *data;

			if (start_frame > src_block->end) {
				/* We need a new block */
				READ_UNLOCK (src_block->lock);
				
				src_block = marlin_block_next (src_block);
				g_assert (src_block != NULL);
				
				READ_LOCK (src_block->lock);
				start_frame = src_block->start;
			}
			
			data = (float *) src_block->frame_data;

			pos_in_block = start_frame - src_block->start;
			left_in_block = (src_block->end - start_frame) + 1;
			
			buf_needed = MARLIN_BLOCK_SIZE - length_got;
			
			/* length_needed is the MIN of what is needed to fill
			   buffer and what is needed to fill the whole thing */
			length_needed = MIN (buf_needed, frames_needed);

			/* Length is the MIN of the amount needed and the
			   amount we can copy */
			length = MIN (length_needed, left_in_block);
			
			/* Now we copy length into the buf */
			memcpy (buf + length_got, data + pos_in_block,
				length * sizeof (float));

			start_frame += length;
			length_got += length;
			frames_needed -= length;

			/* Check if we've filled frames_needed */
			if (frames_needed == 0) {
				break;
			}
		}
				
		offset = write_data_full (dest_channel, buf,
					  length_got, error);
		if (offset == -1) {
			READ_UNLOCK (src_block->lock);
			g_free (buf);

			READ_UNLOCK (src_channel->lock);
			WRITE_UNLOCK (dest_channel->lock);
			return FALSE;
		}
		
		memset (buf, 0, MARLIN_BLOCK_SIZE * sizeof (float));
	}

	READ_UNLOCK (src_block->lock);
	
	g_free (buf);
	READ_UNLOCK (src_channel->lock);
	
	ret = lockless_ensure_mmapped (dest_channel, NULL, error);
	if (ret == FALSE) {
		WRITE_UNLOCK (dest_channel->lock);
		return FALSE;
	}

	/* FIXME: Should copy the peaks from the src_channel */
	ret = lockless_generate_peaks (dest_channel, NULL, error);

	WRITE_UNLOCK (dest_channel->lock);
	return ret;
}

/**
 * marlin_channel_mix:
 * @dest:
 * @src:
 * @src_db:
 * @dest_db:
 * @start_frame:
 * @end_frame:
 * @clip:
 * @operation:
 * @ctxt:
 * @error:
 *
 * Takes the data from @src and mixes it into @dest at @start_frame.
 * If @clip is TRUE, the data is only mixed up until @end_frame,
 * otherwise all data in @src is mixed in.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_mix (MarlinChannel *dest,
		    MarlinChannel *src,
		    double src_db,
		    double dest_db,
		    guint64 start_frame,
		    guint64 end_frame,
		    gboolean clip,
		    MarlinOperation *operation,
		    MarlinUndoContext *ctxt,
		    GError **error)
{
	MarlinBlock *s_block, *d_block, *d, *p;
	float s_ratio, d_ratio;
	guint64 dest_length, range, limit;
	guint64 pos_in_src, pos_in_dest, current, position;
	float *s_data, *d_data, *buf;
	gboolean ret = TRUE;

	g_return_val_if_fail (dest != NULL, FALSE);
	g_return_val_if_fail (src != NULL, FALSE);

	READ_LOCK (src->lock);
	WRITE_LOCK (dest->lock);

	/* If clip == TRUE, then we're only going to end_frame
	   otherwise we go to the end of the sample */
	limit = clip ? end_frame : MIN (dest->frames - 1, start_frame + src->frames - 1);
	dest_length = (limit - start_frame);
	range = clip ? dest_length : src->frames;

	/* Unlink all the blocks we're going to mix. */
	lockless_unlink_range (dest, start_frame, limit, &d, &p, ctxt);
	d_block = d;

	if (src_db == MARLIN_INFINITE_DB) {
		s_ratio = 0;
	} else {
		s_ratio = pow (10.0, src_db / 10.0);
	}

	if (dest_db == MARLIN_INFINITE_DB) {
		d_ratio = 0;
	} else {
		d_ratio = pow (10.0, dest_db / 10.0);
	}

	s_block = src->first;
	pos_in_src = (guint64) 0;

	READ_LOCK (s_block->lock);

	s_data = (float *) s_block->frame_data;

	buf = g_new (float, MARLIN_BLOCK_SIZE);
	current = start_frame;
	while (range > 0) {
		guint64 amount = MIN (MARLIN_BLOCK_SIZE, range);
		
		range -= amount;

		if (d_block != NULL) {
			WRITE_LOCK (d_block->lock);
			
			pos_in_dest = start_frame - d_block->start;
			d_data = (float *) d_block->frame_data;
		} else {
			pos_in_dest = 0;
			d_data = NULL;
		}

		for (position = 0; position < amount; position++) {
			float d;

			if (d_block && pos_in_dest + d_block->start > d_block->end) {
				WRITE_UNLOCK (d_block->lock);
				
				/* Get next block */
				d_block = marlin_block_next (d_block);

				/* This is not allowed if we're clipping */
				if (clip == TRUE) {
					g_assert (d_block != NULL);
					
					WRITE_LOCK (d_block->lock);
				
					pos_in_dest = 0;
					d_data = (float *) d_block->frame_data;
				} else {
					d_data = NULL;
				}
			}
		
			if (pos_in_src + s_block->start > s_block->end) {
				READ_UNLOCK (s_block->lock);
				
				/* Get next block */
				s_block = marlin_block_next (s_block);
				g_assert (s_block != NULL);
			
				READ_LOCK (s_block->lock);
				
				pos_in_src = 0;
				s_data = (float *) s_block->frame_data;
			}
		
			/* Combine the frames */
			if (d_data == NULL && clip == FALSE) {
				d = 0.0;
			} else {
				d = d_data[pos_in_dest] * d_ratio;
			}

			buf[position] = d + (s_data[pos_in_src] * s_ratio);

			pos_in_src++;
			pos_in_dest++;
		}
		
		if (d_block != NULL) {
			WRITE_UNLOCK (d_block->lock);
		}

		ret = lockless_insert_data (dest, buf, amount, current, ctxt, error);
		if (ret == FALSE) {
			goto err;
		}

		current += amount;
	}

 err:
	marlin_block_free_list (d);
	marlin_block_free_list (p);

	READ_UNLOCK (s_block->lock);
	g_free (buf);

	WRITE_UNLOCK (dest->lock);
	READ_UNLOCK (src->lock);

	return ret;
}

static gboolean
lockless_channel_multiply (MarlinChannel *channel,
			   guint64 start,
			   guint64 finish,
			   float ratio,
			   MarlinUndoContext *ctxt,
			   GError **error)
{
	MarlinBlock *block;
	guint64 range, pos_in_block, position;
	float *data, *buf;
	guint64 current;
	gboolean ret;
	
	/* FIXME: Check this */
	range = finish - start;

	buf = g_new (float, MARLIN_BLOCK_SIZE);
	current = start;
	while (range > 0) {
		guint64 amount = MIN (MARLIN_BLOCK_SIZE, range);
		
		range -= amount;
		
		block = get_for_frame (channel->first, start);
		
		WRITE_LOCK (block->lock);
		
		pos_in_block = start - block->start;
		data = (float *) block->frame_data;
		
		for (position = 0; position < amount; position++) {
			if (pos_in_block + block->start > block->end) {
				WRITE_UNLOCK (block->lock);
				
				block = marlin_block_next (block);
				g_assert (block != NULL);
				
				WRITE_LOCK (block->lock);
				
				pos_in_block = (guint64) 0;
				data = (float *) block->frame_data;
			}
			
			/* Apply the ratio */
			buf[position] = data[pos_in_block] * ratio;
			pos_in_block++;
		}
		
		WRITE_UNLOCK (block->lock);
		
		/* Now we need to delete the old block that we've just
		   adjusted and then insert the new block */
		lockless_delete_range (channel, current, current + amount, ctxt);
		ret = lockless_insert_data (channel, buf, amount, current, ctxt, error);
		if (ret == FALSE) {
			g_free (buf);
			return FALSE;
		}

		current += amount;
	}

	g_free (buf);
	return TRUE;
}

/**
 * marlin_channel_invert:
 * channel:
 * start:
 * finish:
 * error:
 *
 * Inverts the data between @start and @finish of @channel.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_invert (MarlinChannel *channel,
		       guint64 start,
		       guint64 finish,
		       MarlinUndoContext *ctxt,
		       GError **error)
{
	gboolean ret;
	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);

	ret = lockless_channel_multiply (channel, start, finish, -1.0, ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_adjust_volume:
 * channel:
 * db: The new volume in decibels
 * start:
 * finish:
 * error:
 *
 * Adjusts the volume of the data between @start and @finish of @channel
 * by @db decibels.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_adjust_volume (MarlinChannel *channel,
			      float db,
			      guint64 start,
			      guint64 finish,
			      MarlinUndoContext *ctxt,
			      GError **error)
{
	float ratio;
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	if (db == MARLIN_INFINITE_DB) {
		ratio = 0.0;
	} else {
		ratio = pow (10.0, db / 10.0);
	}

	WRITE_LOCK (channel->lock);

	ret = lockless_channel_multiply (channel, start, finish, ratio, ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_is_ready:
 * @channel: The channel.
 *
 * Checks if @channel is mmapped and peaks are generated.
 *
 * Returns: TRUE if @channel is ready, FALSE otherwise.
 */
gboolean
marlin_channel_is_ready (MarlinChannel *channel)
{
	g_return_val_if_fail (channel != NULL, FALSE);

	READ_LOCK (channel->lock);
	if (channel->peak_first == NULL) {
		READ_UNLOCK (channel->lock);
		return FALSE;
	} else {
		READ_UNLOCK (channel->lock);
		return TRUE;
	}
}
