/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2003 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>

#include <gdk/gdkkeysyms.h>

#include <gtk/gtkstock.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkimagemenuitem.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtktable.h>
#include <gtk/gtkentry.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkimage.h>

#include <libgnome/gnome-i18n.h>

#include <marlin/marlin-cursors.h>
#include <marlin/marlin-sample.h>
#include <marlin/marlin-marker-model.h>
#include <marlin/marlin-marker-view.h>
#include <marlin/marlin-marshal.h>
#include <marlin/marlin-utils.h>
#include <marlin/marlin-x-utils.h>
#include <marlin/marlin-position-spinner.h>

enum {
	PROP_0,
	PROP_MODEL,
	PROP_FRAMES_PER_PIXEL,
	PROP_TOTAL_FRAMES,
	PROP_SAMPLE,
};

enum {
	POPUP_ADD,
	POPUP_REMOVE,
	POPUP_GOTO,
	POPUP_EDIT
};

enum {
	MOVE_CURSOR,
	LAST_SIGNAL
};

struct _ViewMarker {
	MarlinMarker *marker; /* The marker in the model */

	guint64 real_position; /* The position it is currently at.
				  This may be different from marker->position
				  if we're in the process of moving the
				  marker */
	gboolean hidden; /* Is the marker visible or not */
};

struct _dialog_data {
	GtkWidget *dialog;
	GtkWidget *entry;
	GtkWidget *position;
};

struct _MarlinMarkerViewPrivate {
	MarlinMarkerModel *model;
	guint32 add_id, remove_id, move_id;

	/* Sample */
	MarlinSample *sample;
	MarlinUndoManager *undo;
	
	/* Pixmap to save redrawing */
	GdkPixmap *backing_store;
	GdkGC *non_gr_exp_gc;
	
	GList *markers; /* List of struct _ViewMarker */
	GHashTable *marker_to_view;
	GHashTable *position_to_markers; /* Get the struct _ViewMarker at
					    a position */
	
	/* To update correctly. */
	guint fpp;
	guint64 total_frames;

	int xofs; /* x offset in pixels */
	GtkAdjustment *hadj, *vadj; /* Adjustments for scrolling */

	/* To keep track of the mouse and marker */
	struct _ViewMarker *current_marker;
	struct _ViewMarker *focus_marker, *real_focus; /* Real focus is for
							  storing the
							  focus_marker while
							  we are out of
							  focus */

	/* Keeping an eye on the grab */
	gboolean in_drag_marker;
	struct _ViewMarker *drag_marker;
	guint64 floating_marker_pos;

	/* Popup menu */
	GtkWidget *popup;
	GtkWidget *popup_items[4]; /* For each of the popup items */

	struct _dialog_data *add_data, *edit_data;
};

static GtkWidgetClass *parent_class = NULL;
static signals[LAST_SIGNAL] = { 0 };

#define VIEW_HEIGHT 32
#define EXPANDER_SIZE 10
#define EXPANDER_SIZE_HALF 5 /* This should be half of the EXPANDER_SIZE */
#define DEFAULT_FRAMES_PER_PIXEL 4096

static void
clear_markers (MarlinMarkerView *view)
{
	GList *p;

	for (p = view->priv->markers; p; p = p->next) {
		struct _ViewMarker *vm = p->data;

		g_free (vm);
	}
	g_list_free (view->priv->markers);
	view->priv->markers = NULL;

	g_hash_table_destroy (view->priv->marker_to_view);
	view->priv->marker_to_view = NULL;
	g_hash_table_destroy (view->priv->position_to_markers);
	view->priv->position_to_markers = NULL;
}

static void
finalize (GObject *object)
{
	MarlinMarkerView *view;
	MarlinMarkerViewPrivate *priv;

	view = MARLIN_MARKER_VIEW (object);
	priv = view->priv;

	if (priv == NULL) {
		return;
	}

	if (priv->add_id > 0) {
		g_signal_handler_disconnect (G_OBJECT (priv->model), priv->add_id);
	}

	if (priv->remove_id > 0) {
		g_signal_handler_disconnect (G_OBJECT (priv->model), priv->remove_id);
	}

	if (priv->move_id > 0) {
		g_signal_handler_disconnect (G_OBJECT (priv->model), priv->move_id);
	}
	g_object_unref (G_OBJECT (priv->model));
	g_object_unref (G_OBJECT (priv->undo));
	g_object_unref (G_OBJECT (priv->sample));

	g_object_unref (G_OBJECT (priv->popup));
	
	clear_markers (view);

	if (priv->add_data != NULL) {
		gtk_widget_destroy (priv->add_data->dialog);
		g_free (priv->add_data);
	}

	if (priv->edit_data != NULL) {
		gtk_widget_destroy (priv->edit_data->dialog);
		g_free (priv->edit_data);
	}
	
	g_free (priv);
	view->priv = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
change_focus_marker (MarlinMarkerView *view,
		     struct _ViewMarker *new_focus)
{
	GtkWidget *widget = (GtkWidget *) view;
	GdkRectangle area;
	struct _ViewMarker *vm = view->priv->focus_marker;
	
	if (vm != NULL) {
		int x;

		x = ((vm->real_position / view->priv->fpp) - view->priv->xofs - EXPANDER_SIZE_HALF) - 2;
		
		area.x = MAX (x, 0);
		area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE - 2;
		area.width = EXPANDER_SIZE + 4;
		area.height = EXPANDER_SIZE + 4;

		gdk_window_invalidate_rect (widget->window, &area, FALSE);
	}

	vm = view->priv->focus_marker = new_focus;
	if (vm != NULL) {
		int x;
		
		x = ((vm->real_position / view->priv->fpp) - view->priv->xofs - EXPANDER_SIZE_HALF) - 2;
		area.x = MAX (x, 0);
		
		area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE - 2;
		area.width = EXPANDER_SIZE + 4;
		area.height = EXPANDER_SIZE + 4;
	
		gdk_window_invalidate_rect (widget->window, &area, FALSE);
	}
}

static int
compare_markers (gconstpointer a,
		 gconstpointer b)
{
	struct _ViewMarker *vm_a = (struct _ViewMarker *) a;
	struct _ViewMarker *vm_b = (struct _ViewMarker *) b;

	/* Sanity check just in case */
	if (a == b) {
		return 0;
	}

	/* This should never happen... */
	g_assert (a != NULL);
	g_assert (b != NULL);

	/* The threat of 64 bit arithmetic scares me
	   so I'm doing this the long way */
	if (vm_a->real_position < vm_b->real_position) {
		return -1;
	} else if (vm_a->real_position == vm_b->real_position) {
		return 0;
	} else {
		return 1;
	}
}

static void
marker_added (MarlinMarkerModel *model,
	      MarlinMarker *marker,
	      MarlinMarkerView *view)
{
	struct _ViewMarker *vm;
	GtkWidget *widget = (GtkWidget *) view;
	
	vm = g_new (struct _ViewMarker, 1);
	vm->marker = marker;
	vm->real_position = marker->position;
	vm->hidden = FALSE;

	g_hash_table_insert (view->priv->marker_to_view, marker, vm);
	view->priv->markers = g_list_insert_sorted (view->priv->markers, vm,
						    (GCompareFunc) compare_markers);

	/* The newest added marker gets the focus */
	change_focus_marker (view, vm);
}

static void
marker_removed (MarlinMarkerModel *model,
		MarlinMarker *marker,
		MarlinMarkerView *view)
{
	GtkWidget *widget = (GtkWidget *) view;
	struct _ViewMarker *vm;
	gboolean was_focus;
	guint64 pos;

	vm = g_hash_table_lookup (view->priv->marker_to_view, marker);
	g_assert (vm != NULL);

	g_hash_table_remove (view->priv->marker_to_view, marker);
	view->priv->markers = g_list_remove (view->priv->markers, vm);

	if (vm == view->priv->focus_marker) {
		was_focus = TRUE;
	} else {
		was_focus = FALSE;
	}
	
	if (GTK_WIDGET_DRAWABLE (widget)) {
		change_focus_marker (view, NULL);
	}

	g_free (vm);
}

static void
marker_changed (MarlinMarkerModel *model,
		MarlinMarker *marker,
		MarlinMarkerView *view)
{
	GtkWidget *widget = (GtkWidget *) view;
	GdkRectangle area;
	struct _ViewMarker *vm;
	int x;
	
	vm = g_hash_table_lookup (view->priv->marker_to_view, marker);
	g_assert (vm != NULL);
	
	x = ((vm->real_position / view->priv->fpp) - EXPANDER_SIZE_HALF) - 2;
	area.x = MAX (x, 0);
	
	area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE - 2;
	area.width = EXPANDER_SIZE + 4;
	area.height = EXPANDER_SIZE + 4;
	
	gdk_window_invalidate_rect (widget->window, &area, FALSE);

	/* Move the marker */
	vm->real_position = marker->position;
	x = ((vm->real_position / view->priv->fpp) - EXPANDER_SIZE_HALF) - 2;

	area.x = MAX (x, 0);
	area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE - 2;
	area.width = EXPANDER_SIZE + 4;
	area.height = EXPANDER_SIZE + 4;
	
	gdk_window_invalidate_rect (widget->window, &area, FALSE);
}

static void
set_popup_sensitive (MarlinMarkerView *view,
		     gboolean sensitive)
{
	if (view->priv->current_marker != NULL) {
		gtk_widget_set_sensitive (view->priv->popup_items[POPUP_REMOVE], sensitive);
		gtk_widget_set_sensitive (view->priv->popup_items[POPUP_GOTO], sensitive);
		gtk_widget_set_sensitive (view->priv->popup_items[POPUP_EDIT], sensitive);
	} else {
		gtk_widget_set_sensitive (view->priv->popup_items[POPUP_REMOVE], FALSE);
		gtk_widget_set_sensitive (view->priv->popup_items[POPUP_GOTO], FALSE);
		gtk_widget_set_sensitive (view->priv->popup_items[POPUP_EDIT], FALSE);
	}
}

static gboolean
find_marker (MarlinMarkerView *view,
	     guint64 position,
	     struct _ViewMarker **vm)
{
	GList *p;

	for (p = view->priv->markers; p; p = p->next) {
		struct _ViewMarker *vmarker = p->data;
		gint64 minus = ((gint64)vmarker->real_position) - (EXPANDER_SIZE_HALF *view->priv->fpp);

		if (( ((gint64)position) >= minus) &&
		    (position <= vmarker->real_position +
		     (EXPANDER_SIZE_HALF * view->priv->fpp))) {
			*vm = vmarker;
			return TRUE;
		}
	}

	*vm = NULL;
	return FALSE;
}

static void
add_markers (MarlinMarkerView *view,
	     GList *marks)
{
	GList *p;

	if (view->priv->marker_to_view == NULL) {
		view->priv->marker_to_view = g_hash_table_new (NULL, NULL);
	}

	if (view->priv->position_to_markers == NULL) {
		view->priv->position_to_markers = g_hash_table_new (NULL, NULL);
	}
	
	for (p = marks; p; p = p->next) {
		struct _ViewMarker *vm = g_new (struct _ViewMarker, 1);

		vm->marker = p->data;
		vm->real_position = vm->marker->position;
		vm->hidden = FALSE;

		view->priv->markers = g_list_prepend (view->priv->markers, vm);
		g_hash_table_insert (view->priv->marker_to_view, p->data, vm);
	}
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	GtkWidget *widget;
	MarlinMarkerView *view;
	GList *marks;
	
	view = MARLIN_MARKER_VIEW (object);
	widget = GTK_WIDGET (object);
	
	switch (prop_id) {
	case PROP_MODEL:
		/* Disconnect the signal handlers and unref the model */
		if (view->priv->add_id > 0) {
			g_signal_handler_disconnect (G_OBJECT (view->priv->model),
						     view->priv->add_id);
		}
		if (view->priv->remove_id > 0) {
			g_signal_handler_disconnect (G_OBJECT (view->priv->model),
						     view->priv->remove_id);
		}
		if (view->priv->move_id > 0) {
			g_signal_handler_disconnect (G_OBJECT (view->priv->model),
						     view->priv->move_id);
		}

		if (view->priv->model != NULL) {
			g_object_unref (G_OBJECT (view->priv->model));
		}

		clear_markers (view);
		view->priv->model = g_value_get_object (value);

		/* Add a reference to the model */
		g_object_ref (G_OBJECT (view->priv->model));

		g_object_get (G_OBJECT (view->priv->model),
			      "markers", &marks,
			      NULL);

		add_markers (view, marks);
		
		view->priv->add_id = g_signal_connect (G_OBJECT (view->priv->model),
						       "marker-added",
						       G_CALLBACK (marker_added),
						       view);
		view->priv->remove_id = g_signal_connect (G_OBJECT (view->priv->model),
							  "marker-removed",
							  G_CALLBACK (marker_removed),
							  view);
		view->priv->move_id = g_signal_connect (G_OBJECT (view->priv->model),
							"marker-changed",
							G_CALLBACK (marker_changed),
							view);

		break;

	case PROP_FRAMES_PER_PIXEL:
		view->priv->fpp = g_value_get_uint (value);
		
		if (GTK_WIDGET_DRAWABLE (widget)) {
			GdkRectangle rect;

			rect.x = 0;
			rect.y = 0;
			rect.width = widget->allocation.width;
			rect.height = widget->allocation.height;

			gdk_window_invalidate_rect (widget->window,
						    &rect, FALSE);
		}
		break;

	case PROP_TOTAL_FRAMES:
		/* FIXME: Should this not come from the sample? */
		view->priv->total_frames = g_value_get_uint64 (value);
		break;

	case PROP_SAMPLE:
		if (view->priv->sample != NULL) {
			g_object_unref (G_OBJECT (view->priv->sample));
		}

		view->priv->sample = g_value_get_object (value);
		g_object_ref (G_OBJECT (view->priv->sample));

		g_object_get (G_OBJECT (view->priv->sample),
			      "undo-manager", &view->priv->undo,
			      NULL);
		gtk_widget_set_sensitive (view->priv->popup_items[POPUP_ADD], TRUE);
		break;
		
	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinMarkerView *view;

	view = MARLIN_MARKER_VIEW (object);

	switch (prop_id) {
	case PROP_MODEL:
		g_value_set_object (value, view->priv->model);
		break;

	case PROP_SAMPLE:
		g_value_set_object (value, view->priv->sample);
		break;
		
	default:
		break;
	}
}

static void
make_pixmap (MarlinMarkerView *view,
	     int aw,
	     int ah)
{
	GtkWidget *widget;
	int width, height;

	widget = GTK_WIDGET (view);

	if (view->priv->backing_store) {
		gdk_drawable_get_size (view->priv->backing_store, &width, &height);
		if ((width == aw) &&
		    (height == ah)) {
			return;
		}

		g_object_unref (view->priv->backing_store);
	}

	view->priv->backing_store = gdk_pixmap_new (widget->window,
						    aw, ah, -1);

	if (view->priv->non_gr_exp_gc == NULL) {
		view->priv->non_gr_exp_gc = gdk_gc_new (widget->window);

		gdk_gc_set_exposures (view->priv->non_gr_exp_gc, FALSE);
	}
}

static void
size_allocate (GtkWidget *widget,
	       GtkAllocation *allocation)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);

	if (GTK_WIDGET_REALIZED (widget)) {
		GdkRectangle rect;
		
		gdk_window_move_resize (widget->window,
					allocation->x, allocation->y,
					allocation->width, allocation->height);
		make_pixmap (view, allocation->width, allocation->height);

		rect.x = 0;
		rect.y = 0;
		rect.width = allocation->width;
		rect.height = allocation->height;

		gdk_window_invalidate_rect (widget->window, &rect, FALSE);
	}

	/* Chain up */
	parent_class->size_allocate (widget, allocation);
}

static void
size_request (GtkWidget *widget,
	      GtkRequisition *req)
{
	int font_size;
	PangoFontDescription *fd;

	fd = widget->style->font_desc;
	
	font_size = pango_font_description_get_size (fd);

	req->height = PANGO_PIXELS (font_size) + widget->style->ythickness * 2 + 22;
}

static void
realize (GtkWidget *widget)
{
	MarlinMarkerView *view;
	MarlinMarkerViewPrivate *priv;
	GdkWindowAttr attr;
	int attr_mask;

	view = MARLIN_MARKER_VIEW (widget);
	priv = view->priv;

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

	attr.window_type = GDK_WINDOW_CHILD;
	attr.x = widget->allocation.x;
	attr.y = widget->allocation.y;
	attr.width = widget->allocation.width;
	attr.height = widget->allocation.height;
	attr.wclass = GDK_INPUT_OUTPUT;
	attr.visual = gtk_widget_get_visual (widget);
	attr.colormap = gtk_widget_get_colormap (widget);
	attr.event_mask = (gtk_widget_get_events (widget)
			   | GDK_EXPOSURE_MASK
			   | GDK_BUTTON_PRESS_MASK
			   | GDK_BUTTON_RELEASE_MASK
			   | GDK_POINTER_MOTION_MASK
			   | GDK_LEAVE_NOTIFY_MASK);

	attr_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_COLORMAP | GDK_WA_VISUAL;

	widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
					 &attr, attr_mask);
	gdk_window_set_user_data (widget->window, widget);

	widget->style = gtk_style_attach (widget->style, widget->window);
	gdk_window_set_back_pixmap (widget->window, NULL, FALSE);

	make_pixmap (view, widget->allocation.width, widget->allocation.height);
}

static void
unrealize (GtkWidget *widget)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);
	
	if (view->priv->backing_store) {
		g_object_unref (view->priv->backing_store);
	}

	if (view->priv->non_gr_exp_gc) {
		g_object_unref (view->priv->non_gr_exp_gc);
	}

	view->priv->backing_store = NULL;
	view->priv->non_gr_exp_gc = NULL;

	parent_class->unrealize (widget);
}

static void
draw_ticks (MarlinMarkerView *view,
	    GdkRectangle *area,
	    GtkStateType state_type)
{
	MarlinMarkerViewPrivate *priv = view->priv;
	GtkWidget *widget = (GtkWidget *) view;
	GdkGC *gc;
	int width, height;
	int xthickness, ythickness;
	guint64 ruler_upper, ruler_lower, start, first_mark;
	guint64 point, width_in_frames;
	int increment, gap, s, first_pixel, tenth, i;
	int digit_offset;
	char unit_str[32];
	PangoLayout *layout;
	PangoRectangle logical_rect, ink_rect;

	int incrs[] = {
		100, 200, 500, 500, 1000, 2000, 5000, 10000, 20000,
		50000, 100000, 200000, 500000, 1000000, 2000000
	};
	
	xthickness = widget->style->xthickness;
	ythickness = widget->style->ythickness;

	width = widget->allocation.width;
	height = widget->allocation.height - ythickness * 2;

	layout = gtk_widget_create_pango_layout (widget, "01234556789");
	pango_layout_get_extents (layout, &ink_rect, &logical_rect);

	digit_offset = ink_rect.y;
	
	gc = widget->style->fg_gc[GTK_STATE_NORMAL];
	
	gtk_paint_box (widget->style, view->priv->backing_store,
		       GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		       NULL, widget, "hruler",
		       0, 0,
		       widget->allocation.width, widget->allocation.height);

	gdk_draw_line (view->priv->backing_store, gc,
		       xthickness,
		       (height - 14) + ythickness,
		       widget->allocation.width - xthickness,
		       (height - 14) + ythickness);

	ruler_lower = priv->xofs * priv->fpp;
	ruler_upper = (priv->xofs + width) * priv->fpp;

	/* Find x in 2^x = fpp */
	increment = 100;
	for (i = 0; i < 15; i++) {
		if (pow (2, i) == priv->fpp) {
			increment = incrs[i];
		}
	}

	gap = increment / 10;
	
	start = (priv->xofs * priv->fpp);
	s = start % increment;
	if (s == 0) {
		first_mark = 0;
	} else {
/* 		first_mark = start + (increment - s); */
		first_mark = start - s;
	}
	
	width_in_frames = (width + priv->xofs) * priv->fpp;

	for (point = first_mark; point <= width_in_frames; point += increment) {
		int mark;
		int j;
		guint64 count;

		mark = point / priv->fpp;
		gdk_draw_line (view->priv->backing_store, gc,
			       mark - priv->xofs, (height - 14) + ythickness,
			       mark - priv->xofs, (height - 14) - 5 + ythickness);

		for (j = 1; j < 10; j++) {
			int pos = (point + (j * gap)) / priv->fpp;
			
			gdk_draw_line (view->priv->backing_store, gc,
				       pos - priv->xofs,
				       (height - 14) + ythickness,
				       pos - priv->xofs,
				       (height - 14) - 3 + ythickness);
		}

		/* Draw the text */
		count = point;
		if (count % increment != 0) {
			count = ((count / increment) + 1) * increment;
		}
			
		g_snprintf (unit_str, sizeof (unit_str), "%llu", count);

		pango_layout_set_text (layout, unit_str, -1);
		pango_layout_get_extents (layout, &logical_rect, NULL);

		gtk_paint_layout (widget->style, view->priv->backing_store,
				  GTK_WIDGET_STATE (widget), FALSE,
				  NULL, widget, "hruler",
				  mark - priv->xofs - (PANGO_PIXELS (logical_rect.width) / 2), ythickness + PANGO_PIXELS (logical_rect.y - digit_offset) - 3,
				  layout);
	}
}

static void
draw_markers (MarlinMarkerView *view,
	      GdkRectangle *area,
	      GtkStateType state_type)
{
	GtkWidget *widget = (GtkWidget *) view;
	GList *m;
	guint64 first, last;

	first = view->priv->xofs * view->priv->fpp;
	last = first + (widget->allocation.width * view->priv->fpp);
	
	for (m = view->priv->markers; m; m = m->next) {
		struct _ViewMarker *marker = m->data;
		int x;
		
		if (marker->hidden) {
			continue;
		}
		
		if (marker->real_position < first ||
		    marker->real_position > last) {
			continue;
		}

		if (marker == view->priv->current_marker) {
			state_type = GTK_STATE_PRELIGHT;
		} else {
			state_type = GTK_STATE_NORMAL;
		}

		x = ((int)(marker->real_position / view->priv->fpp)) - view->priv->xofs;
		gtk_paint_expander (widget->style, widget->window,
				    state_type, area, widget,
				    "marlin-marker-view",
				    MAX (x, 0),
				    widget->allocation.height - (widget->style->ythickness + 4),
				    GTK_EXPANDER_EXPANDED);

		if (GTK_WIDGET_HAS_FOCUS (widget) &&
		    marker == view->priv->focus_marker) {
			gtk_paint_focus (widget->style, widget->window,
					 state_type, area, widget,
					 "marlin-marker-view",
					 (marker->real_position / view->priv->fpp) - view->priv->xofs - EXPANDER_SIZE_HALF - 2,
					 widget->allocation.height - (widget->style->ythickness + EXPANDER_SIZE) - 2,
					 EXPANDER_SIZE + 4,
					 EXPANDER_SIZE + 4);
		}
	}

	if (view->priv->in_drag_marker) {
		/* FIXME: Should we only draw this if the floating marker
		   is inside area? */
		gtk_paint_expander (widget->style, widget->window,
				    GTK_STATE_ACTIVE, area, widget,
				    "marlin-marker-view",
				    (view->priv->floating_marker_pos / view->priv->fpp) - view->priv->xofs,
				    widget->allocation.height - (widget->style->ythickness + 4),
				    GTK_EXPANDER_EXPANDED);
	}
}

static gboolean
expose_event (GtkWidget *widget,
	      GdkEventExpose *event)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);
	
	if (GTK_WIDGET_DRAWABLE (widget)) {
		MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);

		draw_ticks (view, &event->area, GTK_STATE_NORMAL);
		gdk_draw_drawable (widget->window,
				   view->priv->non_gr_exp_gc,
				   view->priv->backing_store,
				   0, 0, 0, 0,
				   widget->allocation.width,
				   widget->allocation.height);
		draw_markers (view, &event->area, GTK_STATE_NORMAL);
		
	}

	return FALSE;
}

static void
marker_set_hidden (MarlinMarkerView *view,
		   struct _ViewMarker *marker,
		   gboolean hidden)
{
	GdkRectangle area;
	GtkWidget *widget = GTK_WIDGET (view);
	int x;
	
	marker->hidden = hidden;
	
	x = ((marker->real_position / view->priv->fpp) - EXPANDER_SIZE_HALF) - 2;

	area.x = MAX (x, 0);
	area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE - 2;
	area.width = EXPANDER_SIZE + 4;
	area.height = EXPANDER_SIZE + 4;
	
	gdk_window_invalidate_rect (widget->window, &area, FALSE);
}

/* The floating marker is a "virtual" marker drawn on the
   marker view when a real marker is being dragged.
   This virtual marker only exists as a position in the view
   and doesn't have any real representation as a struct _ViewMarker
   or in the MarlinMarkerView */
static void
redraw_floating_marker (MarlinMarkerView *view,
			guint64 new_position)
{
	GtkWidget *widget = GTK_WIDGET (view);
	GdkRectangle area;
	guint64 total_frames;
	int x;
	
	x = ((view->priv->floating_marker_pos / view->priv->fpp) - EXPANDER_SIZE_HALF) - 2;

	area.x = MAX (x, 0);
	area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE - 2;
	area.width = EXPANDER_SIZE + 4;
	area.height = EXPANDER_SIZE + 4;

	gdk_window_invalidate_rect (widget->window, &area, FALSE);

	g_object_get (G_OBJECT (view->priv->sample),
		      "total_frames", &total_frames,
		      NULL);

	view->priv->floating_marker_pos = new_position < total_frames ?
		new_position : total_frames - view->priv->fpp;
	
	x = ((view->priv->floating_marker_pos / view->priv->fpp) - EXPANDER_SIZE_HALF) - 2;

	area.x = MAX (x, 0);
	area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE - 2;
	area.width = EXPANDER_SIZE + 4;
	area.height = EXPANDER_SIZE + 4;

	gdk_window_invalidate_rect (widget->window, &area, FALSE);	
}
	
static gboolean
button_press_event (GtkWidget *widget,
		    GdkEventButton *event)
{
	MarlinMarkerView *view;
	struct _ViewMarker *vm;
	guint64 real;
	
	view = MARLIN_MARKER_VIEW (widget);

	real = (event->x + view->priv->xofs) * view->priv->fpp;

	gtk_widget_grab_focus (widget);

	switch (event->button) {
	case 1:
		if (find_marker (view, real, &vm)) {
			GdkCursor *grab_hand;
			
			change_focus_marker (view, vm);
			
			/* In drag */
			view->priv->in_drag_marker = TRUE;
			view->priv->drag_marker = vm;
			
			marker_set_hidden (view, vm, TRUE);
			redraw_floating_marker (view, real);
			
			grab_hand = marlin_cursor_get (widget, HAND_CLOSED);
			
			gdk_pointer_grab (widget->window, FALSE,
					  GDK_BUTTON_RELEASE_MASK |
					  GDK_BUTTON1_MOTION_MASK, NULL,
					  grab_hand, event->time);
			gdk_cursor_unref (grab_hand);
		}
		break;

	case 2:
		break;

	case 3:
		gtk_menu_popup (GTK_MENU (view->priv->popup), NULL, NULL,
				NULL, NULL, 3, event->time);
		break;
	}

	return FALSE;
}

static gboolean
button_release_event (GtkWidget *widget,
		      GdkEventButton *event)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);
	MarlinMarkerViewPrivate *priv = view->priv;
	int x;
	guint64 position, total;
	struct _ViewMarker *vm;

	switch (event->button) {
	case 1:
		x = event->x;
		position = (x + view->priv->xofs) * view->priv->fpp;
		
		g_object_get (G_OBJECT (view->priv->sample),
			      "total_frames", &total,
			      NULL);
		
		if (view->priv->in_drag_marker) {
			MarlinUndoContext *ctxt;

			
			gdk_pointer_ungrab (event->time);
			position = position < total ? position : total - view->priv->fpp;

			ctxt = marlin_undo_manager_context_begin (priv->undo,
								  _("Move Marker"));
			marlin_marker_model_move_marker (view->priv->model, view->priv->drag_marker->marker, position, ctxt);
			
			marlin_undo_manager_context_end (priv->undo, ctxt);
			
			marker_set_hidden (view, view->priv->drag_marker, FALSE);
			view->priv->in_drag_marker = FALSE;
			view->priv->drag_marker = NULL;
			
			return FALSE;
		}
		
		if (position > total) {
			return FALSE;
		}
		
		if (find_marker (view, position, &vm) == FALSE) {
			MarlinUndoContext *ctxt;

			ctxt = marlin_undo_manager_context_begin (priv->undo,
								  _("Add Marker"));

			marlin_marker_model_add_marker (view->priv->model, position, NULL, ctxt);
			marlin_undo_manager_context_end (priv->undo, ctxt);
		}
		
		return FALSE;

	case 2:
		break;

	case 3:
		break;
	}
}

static void
dragging_marker (MarlinMarkerView *view,
		 GdkEventMotion *event)
{
	redraw_floating_marker (view, (event->x + view->priv->xofs) * view->priv->fpp);
}

static gboolean
motion_notify_event (GtkWidget *widget,
		     GdkEventMotion *event)
{
	MarlinMarkerView *view;
	MarlinMarkerViewPrivate *priv;
	struct _ViewMarker *vm;
	guint64 px;

	view = MARLIN_MARKER_VIEW (widget);
	priv = view->priv;

	if (priv->in_drag_marker) {
		dragging_marker (view, event);
		return FALSE;
	}
	
	if (event->y < (widget->allocation.height - (14 + widget->style->ythickness))) {
		if (priv->current_marker != NULL) {
			if (GTK_WIDGET_DRAWABLE (widget)) {
				GdkRectangle area;
				
				area.x = (priv->current_marker->real_position / view->priv->fpp) - 5;
				area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE;
				area.width = EXPANDER_SIZE;
				area.height = EXPANDER_SIZE;
				
				gdk_window_invalidate_rect (widget->window, &area, FALSE);
			}
		}

		priv->current_marker = NULL;
		return TRUE;
	}

	px = (priv->xofs + event->x) * priv->fpp;

	if (find_marker (view, px, &vm)) {
		GdkCursor *hand = marlin_cursor_get (widget, HAND_OPEN);

		gdk_window_set_cursor (widget->window, hand);
		gdk_cursor_unref (hand);

		/* On top of a marker so set the menus to sensitive */
		set_popup_sensitive (view, TRUE);
		
		priv->current_marker = vm;

		/* Draw it in a different colour */
		if (GTK_WIDGET_DRAWABLE (widget)) {
			GdkRectangle area;
			int x;
			
			x = (vm->real_position / view->priv->fpp) - 5;
			
			area.x = MAX (x, 0);
			area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE;
			area.width = EXPANDER_SIZE;
			area.height = EXPANDER_SIZE;
			
			gdk_window_invalidate_rect (widget->window, &area, FALSE);
		}
	} else if (priv->current_marker != NULL) {
		if (GTK_WIDGET_DRAWABLE (widget)) {
			GdkRectangle area;
			int x;
			
			x = (priv->current_marker->real_position / view->priv->fpp) - 5;

			area.x = MAX (x, 0);
			area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE;
			area.width = EXPANDER_SIZE;
			area.height = EXPANDER_SIZE;
			
			gdk_window_invalidate_rect (widget->window, &area, FALSE);
		}

		/* Out of a marker so turn the menus off again */
		set_popup_sensitive (view, FALSE);
		
		gdk_window_set_cursor (widget->window, NULL);
		priv->current_marker = NULL;
	} else {
		/* Out of the marker, so turn the menus off again */
		set_popup_sensitive (view, FALSE);
		
		gdk_window_set_cursor (widget->window, NULL);
		priv->current_marker = NULL;
	}

	return FALSE;
}

static gboolean
leave_notify_event (GtkWidget *widget,
		    GdkEventCrossing *event)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);
	MarlinMarkerViewPrivate *priv = view->priv;

	if (priv->current_marker != NULL) {
		if (GTK_WIDGET_DRAWABLE (widget)) {
			GdkRectangle area;
			int x;
			
			x = (priv->current_marker->real_position / view->priv->fpp) - 5;

			area.x = MAX (x, 0);
			area.y = widget->allocation.height - widget->style->ythickness - EXPANDER_SIZE;
			area.width = EXPANDER_SIZE;
			area.height = EXPANDER_SIZE;

			if (priv->current_marker == priv->focus_marker) {
				area.x = MAX (0, ((int) area.x) - 2);
				area.y -= 2;
				area.width += 4;
				area.height += 4;
			}
			
			gdk_window_invalidate_rect (widget->window, &area, FALSE);
		}

/* 		priv->current_marker = NULL; */
	}

	return FALSE;
}

static gboolean
focus_in_event (GtkWidget *widget,
		GdkEventFocus *event)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);

	change_focus_marker (view, view->priv->real_focus);

	return FALSE;
}
	
static gboolean
focus_out_event (GtkWidget *widget,
		 GdkEventFocus *event)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);

	view->priv->real_focus = view->priv->focus_marker;
	change_focus_marker (view, NULL);

	return FALSE;
}

static gboolean
key_press_event (GtkWidget *widget,
		 GdkEventKey *event)
{
	MarlinMarkerView *view = MARLIN_MARKER_VIEW (widget);
	GList *focus_marker, *new_focus, *p;
	
	switch (event->keyval) {
	case GDK_Left:
		/* Select previous marker */
		focus_marker = g_list_find (view->priv->markers,
					    view->priv->focus_marker);
		if (focus_marker == NULL) {
			/* Erk? */
			return FALSE;
		}

		new_focus = focus_marker->prev;
		if (new_focus == NULL) {
			/* We've passed the first marker */
			return FALSE;
		}

		change_focus_marker (view, (struct _ViewMarker *) new_focus->data);
		break;

	case GDK_Right:
		/* Select next marker */
		focus_marker = g_list_find (view->priv->markers,
					    view->priv->focus_marker);
		if (focus_marker == NULL) {
			/* Again, I say "Erk?" */
			return FALSE;
		}

		new_focus = focus_marker->next;
		if (new_focus == NULL) {
			/* Passed the last marker out into the world */
			return FALSE;
		}

		change_focus_marker (view, (struct _ViewMarker *) new_focus->data);
		break;

	default:
		return FALSE;
	}

	return TRUE;
}

static void
scroll_to (MarlinMarkerView *view,
	   int x)
{
	MarlinMarkerViewPrivate *priv;
	GdkRectangle area;
	GtkWidget *widget;
	GdkGC *gc;
	int xofs;
	int width, height;
	int src_x, dest_x;

	widget = GTK_WIDGET (view);
	priv = view->priv;

	xofs = x - priv->xofs;

	if (xofs == 0) {
		return;
	}

	priv->xofs = x;

	if (GTK_WIDGET_DRAWABLE (widget) == FALSE) {
		return;
	}

	width = widget->allocation.width;
	height = widget->allocation.height;

	if (abs (xofs) >= width) {
		goto redraw;
	}

	/* Copy the window area */
	src_x = xofs < 0 ? 0 : xofs;
	dest_x = xofs < 0 ? -xofs : 0;

 redraw:
	area.x = 0;
	area.y = 0;
	area.width = width;
	area.height = height;
	
	gdk_window_invalidate_rect (widget->window, &area, FALSE);
}

static void
adjustment_changed_cb (GtkAdjustment *adj,
		       MarlinMarkerView *view)
{
	scroll_to (view, view->priv->hadj->value);
}

static void
set_scroll_adjustments (MarlinMarkerView *view,
			GtkAdjustment *hadj,
			GtkAdjustment *vadj)
{
	MarlinMarkerViewPrivate *priv;
	gboolean need_adjust;
	
	priv = view->priv;

	if (hadj == NULL ||
	    vadj == NULL) {
		return;
	}

	if (priv->hadj && priv->hadj != hadj) {
		g_signal_handlers_disconnect_matched (G_OBJECT (priv->hadj),
						      G_SIGNAL_MATCH_DATA,
						      0, 0, NULL,
						      NULL, view);
		g_object_unref (G_OBJECT (priv->hadj));
	}

	if (priv->vadj && priv->vadj != vadj) {
		g_object_unref (G_OBJECT (priv->vadj));
	}

	need_adjust = FALSE;

	if (priv->hadj != hadj) {
		priv->hadj = hadj;
		g_object_ref (G_OBJECT (hadj));

		g_signal_connect (G_OBJECT (priv->hadj), "value-changed",
				  G_CALLBACK (adjustment_changed_cb), view);
		need_adjust = TRUE;
	}

	if (priv->vadj != vadj) {
		priv->vadj = vadj;
		g_object_ref (G_OBJECT (priv->vadj));
	}

	if (need_adjust == TRUE) {
		adjustment_changed_cb (NULL, view);
	}
}
		
static void
class_init (MarlinMarkerViewClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
	
	object_class->finalize = finalize;
	object_class->set_property = set_property;
	object_class->get_property = get_property;

	widget_class->size_allocate = size_allocate;
	widget_class->size_request = size_request;
	widget_class->realize = realize;
	widget_class->unrealize = unrealize;
	widget_class->expose_event = expose_event;
	widget_class->button_press_event = button_press_event;
	widget_class->button_release_event = button_release_event;
	widget_class->motion_notify_event = motion_notify_event;
	widget_class->leave_notify_event = leave_notify_event;
	widget_class->focus_in_event = focus_in_event;
	widget_class->focus_out_event = focus_out_event;
	widget_class->key_press_event = key_press_event;
	
	klass->set_scroll_adjustments = set_scroll_adjustments;
	
	parent_class = g_type_class_peek_parent (klass);

	g_object_class_install_property (object_class,
					 PROP_MODEL,
					 g_param_spec_object ("model",
							      "", "",
							      MARLIN_MARKER_MODEL_TYPE,
							      G_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_FRAMES_PER_PIXEL,
					 g_param_spec_uint ("frames_per_pixel",
							    "", "",
							    0, G_MAXUINT, 0,
							    G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_TOTAL_FRAMES,
					 g_param_spec_uint64 ("total_frames",
							      "", "",
							      0,
							      G_MAXUINT64, 0,
							      G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SAMPLE,
					 g_param_spec_object ("sample",
							      "", "",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));
	
	gtk_widget_class_install_style_property (widget_class,
						 g_param_spec_int ("expander_size",
								   "", "",
								   0, G_MAXINT, 10,
								   G_PARAM_READABLE));
	widget_class->set_scroll_adjustments_signal =
		g_signal_new ("set_scroll_adjustments",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (MarlinMarkerViewClass, set_scroll_adjustments),
			      NULL, NULL,
			      marlin_marshal_VOID__OBJECT_OBJECT,
			      G_TYPE_NONE, 2,
			      GTK_TYPE_ADJUSTMENT,
			      GTK_TYPE_ADJUSTMENT);

	signals[MOVE_CURSOR] = g_signal_new ("move-cursor",
					     G_TYPE_FROM_CLASS (klass),
					     G_SIGNAL_RUN_FIRST |
					     G_SIGNAL_NO_RECURSE,
					     G_STRUCT_OFFSET (MarlinMarkerViewClass, move_cursor),
					     NULL, NULL,
					     marlin_marshal_VOID__UINT64,
					     G_TYPE_NONE,
					     1, G_TYPE_UINT64);
}

static void
add_marker_response (GtkDialog *dialog,
		     guint response_id,
		     MarlinMarkerView *view)
{
	const char *name;
	char *real_name;
	guint64 pos;
	GtkAdjustment *adj;
	MarlinUndoContext *ctxt;
	
	switch (response_id) {
	case GTK_RESPONSE_OK:
		name = gtk_entry_get_text (GTK_ENTRY (view->priv->add_data->entry));
		g_object_get (G_OBJECT (view->priv->add_data->position),
			      "adjustment", &adj,
			      NULL);
		
		pos = (guint64) adj->value;
		
		if (name == NULL) {
			real_name = g_strdup_printf (_("Marker %llu"), pos);
		} else {
			real_name = g_strdup (name);
		}

		ctxt = marlin_undo_manager_context_begin (view->priv->undo,
							  _("Add Marker"));
		marlin_marker_model_add_marker (view->priv->model, pos, real_name, ctxt);
		marlin_undo_manager_context_end (view->priv->undo, ctxt);

		g_free (real_name);
		break;

	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-add-marker-dialog");
		return;

	case GTK_RESPONSE_CANCEL:
	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (view->priv->add_data);
	view->priv->add_data = NULL;
}

static void
marker_add (GtkWidget *item,
	    MarlinMarkerView *view)
{
	char *name, *title;
	GtkWidget *vbox, *toplevel, *table, *label;
	struct _dialog_data *ad;
	guint rate;
	guint64 max_frames;
	
	if (view->priv->add_data != NULL) {
		gtk_window_present (GTK_WINDOW (view->priv->add_data->dialog));
		return;
	}

	g_object_get (G_OBJECT (view->priv->sample),
		      "name", &name,
		      NULL);

	title = g_strdup_printf ("Add Marker: %s", name);
	g_free (name);

	ad = view->priv->add_data = g_new (struct _dialog_data, 1);
	ad->dialog = gtk_dialog_new ();
	gtk_window_set_title (GTK_WINDOW (ad->dialog), title);
	g_free (title);
	
	gtk_dialog_set_has_separator (GTK_DIALOG (ad->dialog), FALSE);
	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (view));
	gtk_window_set_transient_for (GTK_WINDOW (ad->dialog), GTK_WINDOW (toplevel));
	gtk_window_set_resizable (GTK_WINDOW (ad->dialog), FALSE);
	
	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (ad->dialog)->vbox), vbox);

	table = marlin_make_table (2, 3, FALSE);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);

	label = marlin_make_title_label (_("_Name:"));
	gtk_widget_show (label);
	PACK (table, label, 0, 0, GTK_FILL);

	ad->entry = gtk_entry_new ();
	gtk_widget_show (ad->entry);
	gtk_table_attach (GTK_TABLE (table), ad->entry,
			  1, 3, 0, 1,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	gtk_label_set_mnemonic_widget (GTK_LABEL (label), ad->entry);
	marlin_add_paired_relations (ad->entry, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);
	
	label = marlin_make_title_label (_("_Position:"));
	gtk_widget_show (label);
	PACK (table, label, 0, 1, GTK_FILL);

	ad->position = marlin_position_spinner_new ();

	g_object_get (G_OBJECT (view->priv->sample),
		      "sample_rate", &rate,
		      "total_frames", &max_frames,
		      NULL);
	g_object_set (G_OBJECT (ad->position),
		      "rate", rate,
		      "max_frames", max_frames,
		      NULL);
	
	gtk_widget_show (ad->position);
	PACK (table, ad->position, 1, 1, GTK_FILL | GTK_EXPAND);

	gtk_label_set_mnemonic_widget (GTK_LABEL (label), ad->position);
	marlin_add_paired_relations (ad->position, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);

	label = marlin_position_spinner_label (MARLIN_POSITION_SPINNER (ad->position));
	gtk_widget_show (label);
	PACK (table, label, 2, 1, GTK_FILL);

	gtk_dialog_add_button (GTK_DIALOG (ad->dialog),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (ad->dialog),
			       _("Add Marker"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (ad->dialog),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (ad->dialog),
					 GTK_RESPONSE_OK);

	g_signal_connect (ad->dialog, "response",
			  G_CALLBACK (add_marker_response), view);
	
	gtk_widget_show (ad->dialog);
}

static void
marker_remove (GtkWidget *item,
	       MarlinMarkerView *view)
{
	struct _ViewMarker *vm;
	MarlinUndoContext *ctxt;

	vm = view->priv->current_marker;
	g_assert (vm != NULL);
	
	ctxt = marlin_undo_manager_context_begin (view->priv->undo,
						  _("Remove Marker"));
	marlin_marker_model_remove_marker (view->priv->model, vm->marker, ctxt);
	marlin_undo_manager_context_end (view->priv->undo, ctxt);
}

static void
marker_goto (GtkWidget *item,
	     MarlinMarkerView *view)
{
	struct _ViewMarker *vm;

	vm = view->priv->current_marker;
	g_assert (vm != NULL);
	
	g_signal_emit (G_OBJECT (view), signals[MOVE_CURSOR], 0, vm->real_position);
}

static void
edit_marker_response (GtkDialog *dialog,
		      guint response_id,
		      MarlinMarkerView *view)
{
	MarlinUndoContext *ctxt;
	const char *name;
	guint64 pos;
	GtkAdjustment *adj;

	switch (response_id) {
	case GTK_RESPONSE_OK:
		ctxt = marlin_undo_manager_context_begin (view->priv->undo,
							  _("Edit Marker"));
		
		name = gtk_entry_get_text (GTK_ENTRY (view->priv->edit_data->entry));
		g_object_get (G_OBJECT (view->priv->edit_data->position),
			      "adjustment", &adj,
			      NULL);
		
		pos = (guint64) adj->value;

		if (name != NULL) {			
			marlin_marker_model_rename_marker (view->priv->model,
							   view->priv->current_marker->marker,
							   name, ctxt);
		}

		
		marlin_marker_model_move_marker (view->priv->model,
						 view->priv->current_marker->marker,
						 pos, ctxt);
		marlin_undo_manager_context_end (view->priv->undo, ctxt);
		break;
			
	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-edit-marker-dialog");
		return;
		
	case GTK_RESPONSE_CANCEL:
	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (view->priv->edit_data);
	view->priv->edit_data = NULL;
}

static void
marker_edit (GtkWidget *item,
	     MarlinMarkerView *view)
{
	char *name, *title;
	GtkWidget *vbox, *toplevel, *table, *label;
	struct _dialog_data *ad;
	struct _ViewMarker *vm;
	guint64 max_frames;
	guint rate;
	MarlinMarker *marker;
	GtkAdjustment *adj;
	
	if (view->priv->edit_data != NULL) {
		gtk_window_present (GTK_WINDOW (view->priv->edit_data->dialog));
		return;
	}

	vm = view->priv->current_marker;
	g_assert (vm != NULL);
	
	marker = vm->marker;
	
	ad = view->priv->edit_data = g_new (struct _dialog_data, 1);
	g_object_get (G_OBJECT (view->priv->sample),
		      "name", &name,
		      NULL);

	title = g_strdup_printf ("Edit Marker: %s", name);
	g_free (name);

	ad->dialog = gtk_dialog_new ();
	gtk_window_set_title (GTK_WINDOW (ad->dialog), title);
	g_free (title);
	
	gtk_dialog_set_has_separator (GTK_DIALOG (ad->dialog), FALSE);
	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (view));
	gtk_window_set_transient_for (GTK_WINDOW (ad->dialog), GTK_WINDOW (toplevel));
	gtk_window_set_resizable (GTK_WINDOW (ad->dialog), FALSE);
	
	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (ad->dialog)->vbox), vbox);

	table = marlin_make_table (2, 3, FALSE);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);

	label = marlin_make_title_label (_("_Name:"));
	gtk_widget_show (label);
	PACK (table, label, 0, 0, GTK_FILL);

	ad->entry = gtk_entry_new ();
	gtk_entry_set_text (GTK_ENTRY (ad->entry), marker->name);
	gtk_entry_set_activates_default (GTK_ENTRY (ad->entry), TRUE);
	
	gtk_widget_show (ad->entry);
	gtk_table_attach (GTK_TABLE (table), ad->entry,
			  1, 3, 0, 1,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	gtk_label_set_mnemonic_widget (GTK_LABEL (label), ad->entry);
	marlin_add_paired_relations (ad->entry, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);
	
	label = marlin_make_title_label (_("_Position:"));
	gtk_widget_show (label);
	PACK (table, label, 0, 1, GTK_FILL);

	g_object_get (G_OBJECT (view->priv->sample),
		      "total_frames", &max_frames,
		      "sample_rate", &rate,
		      NULL);
	
	ad->position = marlin_position_spinner_new ();
	g_object_set (G_OBJECT (ad->position),
		      "rate", rate,
		      "max_frames", max_frames,
		      NULL);
	g_object_get (G_OBJECT (ad->position),
		      "adjustment", &adj,
		      NULL);

	adj->value = (double) marker->position;
	
	gtk_widget_show (ad->position);
	PACK (table, ad->position, 1, 1, GTK_FILL | GTK_EXPAND);

	gtk_label_set_mnemonic_widget (GTK_LABEL (label), ad->position);
	marlin_add_paired_relations (ad->position, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);

	label = marlin_position_spinner_label (ad->position);
	gtk_widget_show (label);
	PACK (table, label, 2, 1, GTK_FILL);

	gtk_dialog_add_button (GTK_DIALOG (ad->dialog),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (ad->dialog),
			       _("Change Details"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (ad->dialog),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (ad->dialog),
					 GTK_RESPONSE_OK);

	g_signal_connect (ad->dialog, "response",
			  G_CALLBACK (edit_marker_response), view);
	
	gtk_widget_show (ad->dialog);
	
}

struct _popupmenuitem {
	char *title;
	char *icon;
	GCallback callback;
};

struct _popupmenuitem menuitems[] = {
	{ N_("_Add Marker..."), GTK_STOCK_ADD, G_CALLBACK (marker_add) },
	{ N_("_Remove Marker"), GTK_STOCK_REMOVE, G_CALLBACK (marker_remove) },
	{ N_("_Move Cursor to Marker"), GTK_STOCK_JUMP_TO, G_CALLBACK (marker_goto) },
	{ N_("_Edit Marker..."), NULL, G_CALLBACK (marker_edit) },
	{ NULL, NULL, NULL }
};

static GtkWidget *
make_marker_popup (MarlinMarkerView *view)
{
	GtkWidget *menu;
	int i;
	
	menu = gtk_menu_new ();
	g_object_ref (menu);
	gtk_object_sink (GTK_OBJECT (menu));
	
	for (i = 0; menuitems[i].title != NULL; i++) {
		GtkWidget *item, *image;

		item = gtk_image_menu_item_new_with_mnemonic (_(menuitems[i].title));
		view->priv->popup_items[i] = item;
		
		if (menuitems[i].icon != NULL) {
			image = gtk_image_new_from_stock (menuitems[i].icon,
							  GTK_ICON_SIZE_MENU);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item),
						       image);
			gtk_widget_show (image);
		}

		g_signal_connect (G_OBJECT (item), "activate",
				  menuitems[i].callback, view);

		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);

		gtk_widget_show (item);
	}

	gtk_widget_set_sensitive (view->priv->popup_items[POPUP_ADD], FALSE);
	set_popup_sensitive (view, FALSE);
	
	return menu;
}
		
static void
init (MarlinMarkerView *view)
{
	GTK_WIDGET_SET_FLAGS (view, GTK_CAN_FOCUS);
	
	view->priv = g_new0 (MarlinMarkerViewPrivate, 1);
	view->priv->marker_to_view = g_hash_table_new (NULL, NULL);
	view->priv->position_to_markers = g_hash_table_new (NULL, NULL);
	view->priv->fpp = DEFAULT_FRAMES_PER_PIXEL;
	view->priv->popup = make_marker_popup (view);
}

GType
marlin_marker_view_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinMarkerViewClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinMarkerView), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (GTK_TYPE_WIDGET,
					       "MarlinMarkerView",
					       &info, 0);
	}

	return type;
}

/**
 * marlin_marker_view_new:
 * @model: The #MarlinMarkerModel that we use.
 * @sample: The #MarlinSample that we use.
 *
 * Creates a new #MarlinMarkerView based on @model and @sample.
 *
 * Returns: Newly created #MarlinMarkerView.
 */
MarlinMarkerView *
marlin_marker_view_new (MarlinMarkerModel *model,
			MarlinSample *sample)
{
	return g_object_new (MARLIN_MARKER_VIEW_TYPE,
			     "model", model,
			     "sample", sample,
			     NULL);
}
