/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gst/gst.h>
#include <gst/gconf/gconf.h>

#include <marlin-play-pipeline.h>
#include <marlin-marshal.h>
#include <marlin-sample.h>
#include <marlin-sample-element.h>
#include <marlin-mt.h>

enum {
	PROP_0,
	PROP_SAMPLE,
	PROP_POSITION
};

struct _link {
	GstPad *pad;
	guint32 link_id;
};

struct _MarlinPlayPipelinePrivate {
	MarlinSample *sample;
	MarlinSampleElementSrc *src;

	GstElement *interleave, *audioconvert, *sink;
	
	guint64 position;
	guint32 tick_id;
	guint32 state_id;

	gint time_seconds;

	GList *links;
};

static MarlinPipeline *parent_class = NULL;

static void
finalize (GObject *object)
{
	MarlinPlayPipeline *pipeline;
	GList *p;
	
	pipeline = MARLIN_PLAY_PIPELINE (object);

	if (pipeline->priv == NULL) {
		return;
	}

	if (pipeline->priv->state_id > 0) {
		g_signal_handler_disconnect (G_OBJECT (pipeline->priv->src), 
					     pipeline->priv->state_id);
		pipeline->priv->state_id = 0;
	}

	if (pipeline->priv->sample != NULL) {
		g_object_unref (G_OBJECT (pipeline->priv->sample));
	}

	if (pipeline->priv->tick_id > 0) {
		g_source_remove (pipeline->priv->tick_id);
		pipeline->priv->tick_id = 0;
	}

	/* Free all the links */
	for (p = pipeline->priv->links; p; p = p->next) {
		struct _link *l = p->data;

 		g_free (l);
	}
	g_list_free (pipeline->priv->links);
	
	g_free (pipeline->priv);
	
	pipeline->priv = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
dispose (GObject *object)
{
	MarlinPlayPipeline *pipeline;
	GList *p;
	
	pipeline = MARLIN_PLAY_PIPELINE (object);
	
	/* Loop though the channels, turning off the unlink signal */
	for (p = pipeline->priv->links; p; p = p->next) {
		struct _link *l = p->data;

		g_signal_handler_disconnect (l->pad, l->link_id);
	}
	
 	G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
ipad_unlinked (GstPad *pad,
	       GstPad *peer,
	       MarlinPlayPipeline *pipeline)
{
	GstElement *parent;
	GList *needle;

	parent = GST_PAD_PARENT (pad);

	/* Find the pad in the links list */
	for (needle = pipeline->priv->links; needle; needle = needle->next) {
		struct _link *l = needle->data;

		if (l->pad == pad) {
			g_free (l);
			break;
		}
	}

	pipeline->priv->links = g_list_remove_link (pipeline->priv->links, needle);
	g_list_free (needle);

	gst_element_remove_pad (parent, pad);
}

static void
sample_src_new_pad (GstElement *src,
		    GstPad *pad,
		    MarlinPlayPipeline *pipeline)
{
	GstPad *isink;
	struct _link *link;

	isink = gst_element_get_request_pad (pipeline->priv->interleave,
					     "sink%d");
	g_assert (isink != NULL);

	link = g_new (struct _link, 1);
	link->pad = isink;
	
	link->link_id = g_signal_connect (G_OBJECT (isink), "unlinked",
					  G_CALLBACK (ipad_unlinked), pipeline);
	pipeline->priv->links = g_list_prepend (pipeline->priv->links, link);

	if (!gst_pad_link (pad, isink)) {
		g_warning ("Caps nego in new_pad failed");
		return;
	}
}

static void
connect_pipeline (MarlinPlayPipeline *pipeline)
{
	guint channels, i;
	
	g_object_get (G_OBJECT (pipeline->priv->sample),
		      "channels", &channels,
		      NULL);

	for (i = 0; i < channels; i++) {
		GstPad *srcpad, *isink;
		struct _link *link;
		char *name;

		name = g_strdup_printf ("src%d", i);
		srcpad = gst_element_get_pad (GST_ELEMENT (pipeline->priv->src),
					      name);
		g_assert (srcpad != NULL);
		g_free (name);
		
		isink = gst_element_get_request_pad (pipeline->priv->interleave,
						     "sink%d");
		g_assert (isink != NULL);

		link = g_new (struct _link, 1);
		link->pad = isink;
		
		link->link_id = g_signal_connect (G_OBJECT (isink), "unlinked",
						  G_CALLBACK (ipad_unlinked), pipeline);
		pipeline->priv->links = g_list_prepend (pipeline->priv->links, link);
		
		if (!gst_pad_link (srcpad, isink)) {
			g_warning ("Caps nego fail 1\n");
			
			return;
		}
	}
}

static void
disconnect_pipeline (MarlinPlayPipeline *pipeline)
{
	gst_element_unlink (GST_ELEMENT (pipeline->priv->src),
			    pipeline->priv->interleave);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinPlayPipeline *pipeline;
	MarlinPlayPipelinePrivate *priv;
	
	pipeline = MARLIN_PLAY_PIPELINE (object);
	priv = pipeline->priv;

	switch (prop_id) {
	case PROP_POSITION:
		priv->position = g_value_get_ulong (value);
		break;

	case PROP_SAMPLE:
		if (priv->sample != NULL) {
			/* Sample is connected */
			disconnect_pipeline (pipeline);
			g_object_unref (priv->sample);
		}

		priv->sample = g_value_get_object (value);
		if (priv->sample == NULL) {
			/* Setting the sample to NULL just means there's
			   no sample, so don't try to connect it */
			return;
		}
		
		g_object_ref (G_OBJECT (priv->sample));
		/* Set the sample on the src */
		g_object_set (G_OBJECT (priv->src),
			      "sample", priv->sample,
			      NULL);
		
/* 		connect_pipeline (pipeline); */
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinPlayPipeline *pipeline;
	MarlinPlayPipelinePrivate *priv;

	pipeline = MARLIN_PLAY_PIPELINE (object);
	priv = pipeline->priv;

	switch (prop_id) {
	case PROP_POSITION:
		g_value_set_ulong (value, priv->position);
		break;

	case PROP_SAMPLE:
		g_value_set_object (value, priv->sample);
		break;

	default:
		break;
	}
}

static void
class_init (GObjectClass *klass)
{
	klass->finalize = finalize;
	klass->dispose = dispose;
	klass->set_property = set_property;
	klass->get_property = get_property;

	parent_class = g_type_class_peek_parent (klass);

	/* Properties */
	g_object_class_install_property (klass,
					 PROP_POSITION,
					 g_param_spec_ulong ("position",
							      "Position",
							      "Position of playback",
							      0, G_MAXULONG, 0,
							      G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_SAMPLE,
					 g_param_spec_object ("sample",
							      "Sample",
							      "The sample to be played",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));
	
}

/* Multi-threaded stuff */
static gboolean
pipeline_tick (MarlinPlayPipeline *pipeline)
{
	GstClock *clock;
	gint64 position;
	gint64 nanos;
	int secs;
	GstFormat format;

	g_return_val_if_fail (IS_MARLIN_PLAY_PIPELINE (pipeline), FALSE);

#if 0	
	clock = gst_bin_get_clock (GST_BIN (pipeline->priv->pipeline));
	nanos = gst_clock_get_time (clock);
	secs = (int) (nanos / GST_SECOND);

	pipeline->priv->time_seconds = secs;
#endif
 	position = marlin_sample_element_src_get_position (pipeline->priv->src, 0);
	marlin_pipeline_position_changed (MARLIN_PIPELINE (pipeline), (guint64) position);

	return TRUE;
}

static void
state_change (GstElement *element,
	      GstElementState old,
	      GstElementState state,
	      MarlinPlayPipeline *pipeline)
{
	switch (state) {
	case GST_STATE_PLAYING:
		if (pipeline->priv->tick_id > 0) {
			g_source_remove (pipeline->priv->tick_id);
		}

		pipeline->priv->tick_id = g_timeout_add (50, (GSourceFunc) pipeline_tick, pipeline);
		break;

	case GST_STATE_PAUSED:
	case GST_STATE_READY:
		if (pipeline->priv->tick_id > 0) {
			g_source_remove (pipeline->priv->tick_id);
			pipeline->priv->tick_id = 0;
		}
		break;
		
	default:
		break;
	}
}

static void
eos_reached (GstElement *element,
	     MarlinPlayPipeline *pipeline)
{
	marlin_pipeline_set_eos (MARLIN_PIPELINE (pipeline));
}

static void
init (MarlinPlayPipeline *pipeline)
{
	gboolean ret; 
	GstCaps *filtercaps;

	pipeline->priv = g_new (MarlinPlayPipelinePrivate, 1);

	/* Make pipeline */
	pipeline->priv->interleave = gst_element_factory_make ("interleave", "play-interleave");
	g_assert (pipeline->priv->interleave != NULL);

	pipeline->priv->audioconvert = gst_element_factory_make ("audioconvert", "play-convert");
	g_assert (pipeline->priv->audioconvert != NULL);
	
	pipeline->priv->sink = gst_gconf_get_default_audio_sink ();
	g_signal_connect (G_OBJECT (pipeline->priv->sink), "eos",
			  G_CALLBACK (eos_reached), pipeline);
	
	pipeline->priv->src = marlin_sample_element_src_new (NULL);
	gst_element_set_name (GST_ELEMENT (pipeline->priv->src), "play-src");
	pipeline->priv->state_id = g_signal_connect (G_OBJECT (pipeline->priv->src), 
						     "state-change",
						     G_CALLBACK (state_change),
						     pipeline);
	g_signal_connect (G_OBJECT (pipeline->priv->src), "new-pad",
			  G_CALLBACK (sample_src_new_pad), pipeline);

	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->interleave);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->audioconvert);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->sink);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), GST_ELEMENT (pipeline->priv->src));

	/* We connect the src when the sample is set */
	filtercaps = gst_caps_new_simple ("audio/x-raw-float", NULL);
	ret = gst_element_link_filtered (pipeline->priv->interleave, pipeline->priv->audioconvert, filtercaps);
	g_assert (ret);
	gst_caps_free (filtercaps);

	gst_element_connect (pipeline->priv->audioconvert, pipeline->priv->sink);

	pipeline->priv->links = NULL;
	pipeline->priv->sample = NULL;
	pipeline->priv->position = 0;
}

GType
marlin_play_pipeline_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (MarlinPlayPipelineClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinPlayPipeline), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (MARLIN_PIPELINE_TYPE,
					       "MarlinPlayPipeline",
					       &info, 0);
	}

	return type;
}

MarlinPlayPipeline *
marlin_play_pipeline_new (void)
{
	return g_object_new (MARLIN_PLAY_PIPELINE_TYPE, NULL);
}

MarlinPlayPipeline *
marlin_play_pipeline_new_from_sample (MarlinSample *sample)
{
	return g_object_new (MARLIN_PLAY_PIPELINE_TYPE,
			     "sample", sample,
			     NULL);
}

void
marlin_play_pipeline_seek (MarlinPlayPipeline *pipeline,
			   guint64 position)
{
	/* Jippy seek thing */
	marlin_sample_element_src_seek (pipeline->priv->src, GST_SEEK_METHOD_SET, position);
}

void
marlin_play_pipeline_seek_range (MarlinPlayPipeline *pipeline,
				 guint64 start,
				 guint64 end)
{
	marlin_sample_element_src_seek_range (pipeline->priv->src,
					      GST_SEEK_METHOD_SET |
					      GST_FORMAT_DEFAULT,
					      start, end);
}
