/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2003 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <marlin/marlin-sample-drawing.h>
#include <marlin/marlin-sample-selection.h>
#include <marlin/marlin-channel.h>
#include <marlin/marlin-block.h>
#undef LOCK_DEBUG
#include <marlin/marlin-read-write-lock.h>

static gboolean
get_min_max_values (MarlinChannel *channel,
		    guint64 start_frame,
		    float *min_value,
		    float *max_value)
{
	guint64 start_block_frame;
	MarlinBlock *block;
	float *frame_data;
	/* When frames_per_pixel == 1,
	   we draw current frame -1 to current frame */

	/* If start_frame - 1 is in the previous block,
	   fetch it and get the correct data. If this is the
	   first block, then just go to the next frame */

	block = marlin_channel_get_block_for_frame (channel, start_frame);
	if (block == NULL) {
		return FALSE;
	}

	READ_LOCK (block->lock);

	start_block_frame = start_frame - block->start;
	if (start_block_frame == 0) {
		MarlinBlock *tmp, *prev;

		prev = block->previous;

		if (prev == NULL) {
			READ_UNLOCK (block->lock);
			return FALSE;
		} else {
			tmp = marlin_block_previous (prev);
		}

		READ_LOCK (tmp->lock);
		frame_data = (float *) tmp->frame_data;
		*max_value = frame_data[tmp->num_frames];
		READ_UNLOCK (tmp->lock);
	} else {
		frame_data = (float *) block->frame_data;
		*max_value = frame_data[start_block_frame - 1];
	}

	frame_data = (float *) block->frame_data;
	*min_value = frame_data[start_block_frame];
	READ_UNLOCK (block->lock);

	return TRUE;
}

#define USE_PEAKS
/* #undef USE_PEAKS */

static void
get_min_max_peaks (MarlinChannel *channel,
		   guint64 start_frame,
		   int frames_per_pixel,
		   float *min_peak,
		   float *max_peak)
{
	MarlinBlock *block;
	int i, o;

	*min_peak = 0.0;
	*max_peak = 0.0;

#ifdef USE_PEAKS
	if (frames_per_pixel < MARLIN_FRAMES_PER_PEAK)
#endif
	{
		float *frame_data;

		block = marlin_channel_get_block_for_frame (channel, start_frame);
		READ_LOCK (block->lock);
		
		frame_data = (float *) block->frame_data;

		/* Scan through the real sample data to get the peaks */
		for (i = 0, o = start_frame; i < frames_per_pixel; i++, o++) {
			guint64 frame_in_block;

			if (o > block->end) {
				/* Get the next block and reset o to the
				   first frame in the block */
				READ_UNLOCK (block->lock);

				block = marlin_block_next (block);
				g_assert (block != NULL);

				READ_LOCK (block->lock);

				o = block->start;
				frame_data = (float *) block->frame_data;
			}

			frame_in_block = o - block->start;
			*min_peak = MIN (*min_peak, frame_data[frame_in_block]);
			*max_peak = MAX (*max_peak, frame_data[frame_in_block]);
		}
		READ_UNLOCK (block->lock);
	}
#ifdef USE_PEAKS
	else {
		int start_peak, peak_range;
		MarlinPeak *peak_data;

		/* FIXME: Can I cache this somehow? */
		block = marlin_channel_get_peak_for_frame (channel, start_frame);
		if (block == NULL) {
			g_print ("Start_Frame: %llu\n", start_frame);
			g_assert (block != NULL);
		}

		READ_LOCK (block->lock);

		peak_data = (MarlinPeak *) block->frame_data;

		start_peak = (start_frame - block->start) / block->scale_factor;
		peak_range = MIN (frames_per_pixel / block->scale_factor,
				  (block->end - start_frame) / block->scale_factor);

		for (i = 0, o = start_peak; i < peak_range; i++, o++) {
			guint64 peak_in_block;
			MarlinPeak peak;

			if (o * block->scale_factor > block->end) {
				/* Next block */
				READ_UNLOCK (block->lock);

				block = marlin_block_next (block);
				g_assert (block != NULL);

				READ_LOCK (block->lock);
				o = 0;
				peak_data = (MarlinPeak *) block->frame_data;
			}

			peak = peak_data[o];

			*min_peak = MIN (*min_peak, peak.low);
			*max_peak = MAX (*max_peak, peak.high);
		}

		READ_UNLOCK (block->lock);
	}
#endif
}

#define YPOS(v) CLAMP (chan_area.y + chan_height - ((((v) - vmin) * chan_height) / (vmax - vmin)), chan_area.y, chan_area.y + chan_height)

void
marlin_sample_draw (MarlinSample *sample,
		    GtkWidget *widget,
		    GdkRectangle *area,
		    GtkStateType state_type,
		    guint channel_num,
		    guint fpp,
		    int xofs,
		    float vmax,
		    float vmin)
{
	MarlinChannel *channel;
	MarlinBlock *block;
	GdkRectangle chan_area;
	guint64 start_frame;
	int quart, half, i, j, num_channels;
	int chan_height, chan_offset, extra;
	
	g_object_get (G_OBJECT (sample),
		      "channels", &num_channels,
		      NULL);

	extra = num_channels - 1;
	chan_height = (widget->allocation.height - extra) / num_channels;
	chan_offset = (chan_height * channel_num) + channel_num;

	chan_area.x = 0;
	chan_area.y = (chan_height * channel_num) + channel_num;;
	chan_area.width = widget->allocation.width;
	chan_area.height = chan_height;

	quart = (int) (chan_height / 4);
	half = quart * 2;

	start_frame = (xofs + area->x) * fpp;

	channel = marlin_sample_get_channel (sample, channel_num);

	if (channel == NULL || marlin_channel_is_ready (channel) == FALSE) {
		return;
	}
	
	for (i = 0; i < area->width; i++) {
		int xoff = 0;
		float y1, y2;
		float *frame_data;
		int sy1, sy2;

		if (fpp > 1) {
			get_min_max_peaks (channel, start_frame,
					   fpp, &y1, &y2);
		} else {
			gboolean ret;

			ret = get_min_max_values (channel, start_frame, &y1, &y2);
			if (ret == FALSE) {
				goto cont;
			}

			xoff = -1;
		}

		/* Scale the line up and adjust for centre line */
#if 0
		sy2 = chan_offset + (half - (guint) (y2 * half));
		sy1 = chan_offset + (half - (guint) (y1 * half));
#else
		sy2 = (guint) YPOS (y2);
		sy1 = (guint) YPOS (y1);
#endif

		/* Draw the line */
/* 		gdk_gc_set_clip_rectangle (widget->style->text_gc[state_type], &chan_area); */
		gdk_draw_line (widget->window,
			       widget->style->text_gc[state_type],
			       area->x + i + xoff, sy2,
			       area->x + i, sy1);
/* 		gdk_gc_set_clip_rectangle (widget->style->text_gc[state_type], NULL); */

	cont:
		start_frame += fpp;
	}		
}
