/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <marlin/marlin-sample-selection.h>

enum {
	CHANGED,
	LAST_SIGNAL
};

static GObjectClass *parent_class = NULL;
static guint signals[LAST_SIGNAL];

static void
finalize (GObject *object)
{
	MarlinSampleSelection *selection;
	GList *p;

	selection = MARLIN_SAMPLE_SELECTION (object);

	if (selection->lock == NULL) {
		return;
	}
	
	marlin_read_write_lock_lock (selection->lock,
				     MARLIN_READ_WRITE_LOCK_MODE_WRITE);
	for (p = selection->selections; p; p = p->next) {
		g_free (p->data);
	}
	g_list_free (selection->selections);

	marlin_read_write_lock_unlock (selection->lock,
				       MARLIN_READ_WRITE_LOCK_MODE_WRITE);
	marlin_read_write_lock_destroy (selection->lock);
	selection->lock = NULL;
	
	parent_class->finalize (object);
}

static void
class_init (MarlinSampleSelectionClass *klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = finalize;

	parent_class = g_type_class_peek_parent (klass);

	signals[CHANGED] = g_signal_new ("changed",
					 G_TYPE_FROM_CLASS (klass),
					 G_SIGNAL_RUN_FIRST |
					 G_SIGNAL_NO_RECURSE,
					 G_STRUCT_OFFSET (MarlinSampleSelectionClass, changed),
					 NULL, NULL,
					 g_cclosure_marshal_VOID__VOID,
					 G_TYPE_NONE, 0);
}

static void
init (MarlinSampleSelection *selection)
{
	selection->lock = marlin_read_write_lock_new ();
	selection->selections = NULL;
}

GType
marlin_sample_selection_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinSampleSelectionClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinSampleSelection), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (G_TYPE_OBJECT,
					       "MarlinSampleSelection", &info, 0);
	}

	return type;
}
			
MarlinSampleSelection *
marlin_sample_selection_new (MarlinSample *sample)
{
	MarlinSampleSelection *selection;

	selection = g_object_new (MARLIN_SAMPLE_SELECTION_TYPE, NULL);
				  
	selection->sample = sample;
	return selection;
}

/* NB: This function must be called while the parent holds
   a write lock on selection->lock */
static void
compress_selection (MarlinSampleSelection *selection)
{
	GList *p, *selections;

	if (selection->selections == NULL ||
	    selection->selections->next == NULL) {
		/* If there are 0 or 1 selections then just return */
		return;
	}

	selections = selection->selections;

	for (p = selections; p; ) {
		MarlinSampleSubSelection *sub1, *sub2;

		sub1 = p->data;
		if (p->next == NULL) {
			break; /* We've reached the end of the list */
		}
#if 0
		if (sub1->start == sub1->end) {
			GList *temp = p;
			
			g_print ("Removed tiny selection\n");
			/* Remove this selection */
			p = temp->next;
			selection->selections = g_list_remove (selection->selections, sub1);
			g_free (sub1);
			continue;
		}
#endif
		sub2 = p->next->data;

		/* Do these selections cross */
		if (sub1->end >= sub2->start) {
			/* Yes, they do.... */
			if (sub1->coverage != sub2->coverage) {
				MarlinCoverage c1, c2;

				c1 = sub1->coverage;
				c2 = sub2->coverage;

				if (c1 == MARLIN_COVERAGE_BOTH ||
				    c2 == MARLIN_COVERAGE_BOTH) {
					/* We need to make 2 selections.
					   The COVERAGE_BOTH gets the overlap
					   and the other is shrunk */

					if (c1 == MARLIN_COVERAGE_BOTH) {
						sub2->start = sub1->end;
					} else {
						sub1->end = sub2->start;
					}
				} else {
					MarlinSampleSubSelection *new_sub;
					/* We need to split into 3 selections */

					new_sub = g_new (MarlinSampleSubSelection, 1);
					new_sub->coverage = MARLIN_COVERAGE_BOTH;

					/* Two cases:
					   a) sub1 overlaps sub2
					   b) sub1 engulfs sub2
					*/

					if (sub1->end <= sub2->end) {
						new_sub->start = sub2->start;
						new_sub->initial = sub2->start;
						new_sub->end = sub1->end;

						sub1->end = new_sub->start;
						sub2->start = new_sub->end;
					} else {
						new_sub->start = sub2->start;
						new_sub->initial = sub2->start;
						new_sub->end = sub2->end;

						sub2->end = sub1->end;
						sub1->end = new_sub->start;
						sub2->start = new_sub->end;
						sub2->initial = sub1->start;
						sub2->coverage = sub1->coverage;
					}

					/* Insert before p->next */
					selection->selections = g_list_insert_before (selection->selections, p->next, new_sub);
				}
			} else {
				/* Combine and remove sub2 */
				if (sub2->end >= sub1->end) {
					sub1->end = sub2->end;
				}

				selection->selections = g_list_remove (selection->selections, sub2);
				g_free (sub2);
			}
		}

		p = p->next;
	}
}

MarlinSampleSubSelection *
marlin_sample_selection_add (MarlinSampleSelection *selection,
			     guint64 start,
			     guint64 end,
			     MarlinCoverage coverage)
{
	MarlinSampleSubSelection *sub;
	GList *p, *n, *last;
	
	g_return_if_fail (IS_MARLIN_SAMPLE_SELECTION (selection));

	/* The old switcheroo */
	if (start > end) {
		guint64 tmp;

		tmp = start;
		start = end;
		end = tmp;
	}
	
	sub = g_new (MarlinSampleSubSelection, 1);
	sub->initial = start;
	sub->start = start;
	sub->end = end;
	sub->coverage = coverage;

	marlin_read_write_lock_lock (selection->lock,
				     MARLIN_READ_WRITE_LOCK_MODE_WRITE);

	if (selection->selections == NULL) {
		selection->selections = g_list_append (NULL, sub);
	} else {
		last = NULL;
		for (p = selection->selections; p; p = p->next) {
			MarlinSampleSubSelection *tmp = p->data;
			
			/* Find the last end < start */
			if (tmp->start <= start) {
				last = p;
			} else {
				break;
			}
		}
		
		if (last != NULL) {
			n = g_list_alloc ();
			/* Insert our list...
			   Why is there no g_list_insert_after? */
			n->data = sub;
			
			n->next = last->next;
			last->next = n;
			if (n->next != NULL) {
				n->next->prev = n;
			}
			
			n->prev = last;
		} else {
			selection->selections = g_list_prepend (selection->selections, sub);
		}
	}

	marlin_read_write_lock_unlock (selection->lock,
				       MARLIN_READ_WRITE_LOCK_MODE_WRITE);

	g_signal_emit (G_OBJECT (selection), signals[CHANGED], 0);

	return sub;
}

void
marlin_sample_selection_remove (MarlinSampleSelection *selection,
				MarlinSampleSubSelection *sub)
{
	GList *p;

	g_return_if_fail (IS_MARLIN_SAMPLE_SELECTION (selection));
	g_return_if_fail (sub != NULL);

	marlin_read_write_lock_lock (selection->lock,
				     MARLIN_READ_WRITE_LOCK_MODE_WRITE);
	selection->selections = g_list_remove (selection->selections, sub);
	marlin_read_write_lock_unlock (selection->lock,
				       MARLIN_READ_WRITE_LOCK_MODE_WRITE);

	g_free (sub);

	g_signal_emit (G_OBJECT (selection), signals[CHANGED], 0);
}

gboolean
marlin_sample_selection_frame_in_selection (MarlinSampleSelection *selection,
					    MarlinSampleSubSelection **sub,
					    guint64 frame,
					    MarlinCoverage coverage)
{
	GList *p;
	
	g_return_if_fail (IS_MARLIN_SAMPLE_SELECTION (selection));

	if (sub != NULL) {
		*sub = NULL;
	}
	
	marlin_read_write_lock_lock (selection->lock,
				     MARLIN_READ_WRITE_LOCK_MODE_READ);
	for (p = selection->selections; p; p = p->next) {
		MarlinSampleSubSelection *subsel = p->data;
		
		if (frame >= subsel->start && frame <= subsel->end) {

			switch (subsel->coverage) {
			case MARLIN_COVERAGE_BOTH:
				/* Don't care about coverage...*/
				marlin_read_write_lock_unlock (selection->lock,
							       MARLIN_READ_WRITE_LOCK_MODE_READ);

				if (sub != NULL) {
					*sub = subsel;
				}
				
				return TRUE;

			case MARLIN_COVERAGE_LEFT:
			case MARLIN_COVERAGE_RIGHT:
				/* If the two coverages match
				   then we've found one, otherwise
				   continue */
				if (coverage == subsel->coverage) {
					marlin_read_write_lock_unlock (selection->lock,
								       MARLIN_READ_WRITE_LOCK_MODE_READ);
					if (sub != NULL) {
						*sub = subsel;
					}
					
					return TRUE;
				}

				break;

			default:
				break;
			}
		}
	}

	marlin_read_write_lock_unlock (selection->lock,
				       MARLIN_READ_WRITE_LOCK_MODE_READ);
	return FALSE;
}

void
marlin_sample_selection_clear (MarlinSampleSelection *selection)
{
	GList *p, *l;
	
	g_return_if_fail (IS_MARLIN_SAMPLE_SELECTION (selection));

	marlin_read_write_lock_lock (selection->lock,
				     MARLIN_READ_WRITE_LOCK_MODE_WRITE);
	l = selection->selections;
	selection->selections = NULL;

	/* The selection has changed, unlock and emit a signal */
	marlin_read_write_lock_unlock (selection->lock,
				       MARLIN_READ_WRITE_LOCK_MODE_WRITE);
	g_signal_emit (G_OBJECT (selection), signals[CHANGED], 0);

	/* Now we can safely free the old selection list without anyone
	   waiting on this potentially expensive operation */
	for (p = l; p; p = p->next) {
		g_free (p->data);
	}
	g_list_free (l);
}

void
marlin_sample_selection_changed (MarlinSampleSelection *selection,
				 gboolean compress)
{

	g_return_if_fail (IS_MARLIN_SAMPLE_SELECTION (selection));

	if (compress) {
		marlin_read_write_lock_lock (selection->lock,
					     MARLIN_READ_WRITE_LOCK_MODE_WRITE);
		compress_selection (selection);
		marlin_read_write_lock_unlock (selection->lock,
					       MARLIN_READ_WRITE_LOCK_MODE_WRITE);
	}
	
	g_signal_emit (G_OBJECT (selection), signals[CHANGED], 0);
}
