/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 - 2003 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>

#include <gst/gst.h>

#include <libgnome/gnome-i18n.h>

#include <marlin-block.h>
#include <marlin-sample.h>
#include <marlin-channel.h>
#include <marlin-mt.h>
#include <marlin-msgport.h>
#include <marlin-marshal.h>
#include <marlin-operation.h>
#include <marlin-sample-element.h>
#include <marlin-read-write-lock.h>
#include <marlin-sample-selection.h>
#include <marlin-gst-extras.h>
#include <marlin-marker-model.h>
#include <marlin-undo-manager.h>

#define MAX_NUM_CHANNELS 16

enum {
	PROP_0,
	PROP_TOTAL,
	PROP_SAMPLE_RATE,
	PROP_FILENAME,
	PROP_REALNAME,
	PROP_CHANNELS,
	PROP_DEPTH,
	PROP_LENGTH,
	PROP_DIRTY,
	PROP_WRITABLE,
	PROP_LOADING,
	PROP_SELECTION,
	PROP_MARKER_MODEL,
	PROP_TAGLIST,
	PROP_ENCODER,
	PROP_UNDO_MANAGER,
};
	
struct _MarlinSamplePrivate {
	MarlinReadWriteLock *lock; /* RWLock protecting data */
 	GPtrArray *channel_data; /* Array of the channel data */

	/* File IO information */
	char *filename; /* The filename that this was created from */
	char *realname; /* The realname of the sample */
	gboolean dirty; /* Has the sample been edited since the last save? */
	gboolean writable; /* Is the filename writable? */
	gboolean revertable; /* Has the sample got revert info? */
	
	/* Sample information */
	MarlinSampleHeader *header; /* Header for the sample */

	/* Selection information */
	/* FIXME: Should this be in a view? */
	MarlinSampleSelection *selection; /* The selections active on this
					     sample */

	/* Model for the markers */
	MarlinMarkerModel *markers;

	GstTagList *taglist; /* The tag list */

	GstElement *encoder; /* The encoder pipeline that was used
				to save last time */

	MarlinUndoManager *undo;
};

static MarlinObjectClass *parent_class = NULL;

static char *tmp_dir = NULL;

static void
finalize (GObject *object)
{
	MarlinSample *sample;
	MarlinSamplePrivate *priv;
	int i;

	sample = MARLIN_SAMPLE (object);
	priv = sample->priv;

	if (priv == NULL) {
		return;
	}

	/* Lock ... */
	WRITE_LOCK (priv->lock);
	g_print ("Destroying sample %p\n", sample);

	for (i = 0; i < priv->channel_data->len; i++) {
		marlin_channel_free (priv->channel_data->pdata[i]);
	}
	
	g_free (priv->filename);

	g_free (priv->header);

	g_object_unref (G_OBJECT (priv->selection));
	g_object_unref (G_OBJECT (priv->markers));

	if (priv->taglist != NULL) {
		gst_structure_free (priv->taglist);
	}
	
	WRITE_UNLOCK (priv->lock);
	marlin_read_write_lock_destroy (priv->lock);
	
	g_free (priv);
	sample->priv = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}
#if 0
struct _cue_marks {
	guint64 position;
	char *name;

	/* FIXME: when Marlin handles ranges, add it here */
};

static char *
get_name_from_caps (GList *caps_list,
		    guint32 id)
{
	GList *p;

	for (p = caps_list; p; p = p->next) {
		GstCaps *caps = p->data;
		guint32 new_id;

		new_id = get_int_from_caps (caps, "identifier");
		if (id == new_id) {
			return get_string_from_caps (caps, "name");
		}
	}

	return NULL;
}

static GList *
match_cue_labels (GList *cues,
		  GList *labels,
		  GList *ltxts)
{
	GList *markers = NULL, *p;
	guint32 id, position;

	for (p = cues; p; p = p->next) {
		GstCaps *cue_caps = p->data;
		struct _cue_marks *cm;
		char *name;
		
		id = get_int_from_caps (cue_caps, "identifier");
		position = get_int_from_caps (cue_caps, "position");

		name = get_name_from_caps (labels, id);
		if (name == NULL) {
			get_name_from_caps (ltxts, id);
		}

		cm = g_new (struct _cue_marks, 1);
		cm->position = position;
		cm->name = name;

		markers = g_list_append (markers, cm);
	}

	return markers;
}

static void
parse_metadata (MarlinSample *sample)
{
	MarlinSamplePrivate *priv;
	GList *cues, *labels, *ltxts, *markers, *p;
	int i;

	priv = sample->priv;

	cues = get_glist_from_caps (priv->metadata, "cues");
	if (cues == NULL) {
		/* If we don't have cuepoints there's no point continuing */
		return;
	}

	ltxts = get_glist_from_caps (priv->metadata, "ltxts");
	/* Don't use ltxts yet - they're kinda scary */
	
	labels = get_glist_from_caps (priv->metadata, "labels");
	markers = match_cue_labels (cues, labels, ltxts);

	marlin_marker_model_clear (priv->markers);
	
	for (i = 0, p = markers; p; p = p->next, i++) {
		struct _cue_marks *cm = p->data;

		marlin_marker_model_add_marker (priv->markers,
						(guint64) cm->position,
						cm->name);
		g_free (cm->name);
		g_free (cm);
	}
	g_list_free (markers);
}
#endif

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinSample *sample;
	MarlinSamplePrivate *priv;
	MarlinSampleHeader *header;
	MarlinMarkerModel *model;
	GstStructure *taglist;
	GstElement *encoder;
	char *basename;
	int i, old, dc;
	guint64 frames;
	guint sample_rate;
	GError *error = NULL;

	sample = MARLIN_SAMPLE (object);
	priv = sample->priv;
	header = priv->header;
	
	switch (prop_id) {
	case PROP_TOTAL:
		WRITE_LOCK (priv->lock);

		frames = g_value_get_uint64 (value);
		if (header->number_of_frames == frames) {
			WRITE_UNLOCK (priv->lock);
			return;
		}

		header->number_of_frames = frames;
		if (header->sample_rate != 0) {
			header->length = (float) header->number_of_frames / header->sample_rate;
		}

		WRITE_UNLOCK (priv->lock);

		marlin_object_notify (G_OBJECT (sample), "total-frames");
		break;

	case PROP_SAMPLE_RATE:
		WRITE_LOCK (priv->lock);
		
		sample_rate = g_value_get_uint (value);
		if (header->sample_rate == sample_rate) {
			WRITE_UNLOCK (priv->lock);
			return;
		}

		header->sample_rate = sample_rate;
		if (header->sample_rate != 0) {
			header->length = (float) header->number_of_frames / header->sample_rate;
		}

		WRITE_UNLOCK (priv->lock);
		marlin_object_notify (G_OBJECT (sample), "sample-rate");
		break;

	case PROP_FILENAME:
		WRITE_LOCK (priv->lock);
		
		if (priv->filename != NULL) {
			g_free (priv->filename);
		}
		priv->filename = g_strdup (g_value_get_string (value));

		if (access (priv->filename, W_OK) == 0) {
			priv->writable = TRUE;
		} else {
			priv->writable = FALSE;
		}
		
		WRITE_UNLOCK (priv->lock);
		break;

	case PROP_REALNAME:
		WRITE_LOCK (priv->lock);

		if (priv->realname != NULL) {
			g_free (priv->realname);
		}
		priv->realname = g_strdup (g_value_get_string (value));

		WRITE_UNLOCK (priv->lock);
		break;
		
	case PROP_CHANNELS:
		WRITE_LOCK (priv->lock);

		old = header->channels;
		header->channels = g_value_get_uint (value);
		
		dc = old - header->channels;
		if (dc == 0) {
			WRITE_UNLOCK (priv->lock);
			break;
		} else if (dc > 0) {
			for (i = old; i > header->channels; i--) {
				marlin_channel_free (priv->channel_data->pdata[i - 1]);
				g_ptr_array_remove_index (priv->channel_data,
							  i - 1);
			}
		} else {
			for (i = 0; i < abs (dc); i++) {
				MarlinChannel *channel;
				
				channel = marlin_channel_new (sample, priv->filename, &error);
				if (channel == NULL) {
					g_warning ("Error making channel");
					WRITE_UNLOCK (priv->lock);
					return;
				}

				g_ptr_array_add (priv->channel_data, channel);
			}
		}
		
		WRITE_UNLOCK (priv->lock);
		marlin_object_notify (G_OBJECT (sample), "channels");
		break;

	case PROP_DEPTH:
		WRITE_LOCK (priv->lock);
		header->depth = g_value_get_int (value);
		WRITE_UNLOCK (priv->lock);
		break;

	case PROP_DIRTY:
		WRITE_LOCK (priv->lock);
		priv->dirty = g_value_get_boolean (value);
		WRITE_UNLOCK (priv->lock);
		marlin_object_notify (G_OBJECT (sample), "dirty");

		if (priv->dirty == FALSE) {
			/* Store the new saved data structures 
			   for the revert */
			marlin_object_notify (G_OBJECT (sample), "revertable");
		}
		break;

	case PROP_MARKER_MODEL:
		model = g_value_get_object (value);
		
		if (priv->markers == model) {
			/* Don't need to do anything */
			return;
		}

		g_object_unref (G_OBJECT (priv->markers));
		priv->markers = model;
		g_object_ref (G_OBJECT (priv->markers));
		
		break;

	case PROP_TAGLIST:
		taglist = g_value_get_pointer (value);
		if (taglist == priv->taglist) {
			return;
		}
		
		if (priv->taglist != NULL) {
			gst_structure_free (priv->taglist);
		}

		priv->taglist = taglist;
		break;

	case PROP_ENCODER:
		encoder = g_value_get_object (value);

		if (priv->encoder == encoder) {
			/* Don't need to do anything */
			return;
		}

		g_object_unref (G_OBJECT (priv->encoder));
		priv->encoder = encoder;
		g_object_ref (G_OBJECT (priv->encoder));

		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinSample *sample;
	MarlinSampleHeader *header;
	char *title = NULL;
	
	sample = MARLIN_SAMPLE (object);
	header = sample->priv->header;

	READ_LOCK (sample->priv->lock);
	switch (prop_id) {
	case PROP_TOTAL:
		g_value_set_uint64 (value, header->number_of_frames);
		break;

	case PROP_SAMPLE_RATE:
		g_value_set_uint (value, header->sample_rate);
		break;

	case PROP_FILENAME:
		g_value_set_string (value, sample->priv->filename);
		break;

	case PROP_REALNAME:
		if (sample->priv->taglist != NULL) {
			gboolean ret;

			ret = gst_tag_list_get_string (sample->priv->taglist,
						       GST_TAG_TITLE,
						       &title);
			if (ret && title != NULL && *title != 0) {
				g_value_set_string (value, title);
				g_free (title);
			} else {
				char *realname;

				realname = g_path_get_basename (sample->priv->filename);
				g_value_set_string (value, realname);
				g_free (realname);
			}
		} else {
			char *realname;
			
			realname = g_path_get_basename (sample->priv->filename);
			g_value_set_string (value, realname);
			g_free (realname);
		}
		break;
		
	case PROP_CHANNELS:
		g_value_set_uint (value, header->channels);
		break;

	case PROP_DEPTH:
		g_value_set_int (value, header->depth);
		break;

	case PROP_LENGTH:
		g_value_set_float (value, header->length);
		break;
		
	case PROP_DIRTY:
		g_value_set_boolean (value, sample->priv->dirty);
		break;

	case PROP_WRITABLE:
		g_value_set_boolean (value, sample->priv->writable);
		break;
		
	case PROP_SELECTION:
		g_value_set_object (value, sample->priv->selection);
		break;

	case PROP_MARKER_MODEL:
		g_value_set_object (value, sample->priv->markers);
		break;

	case PROP_TAGLIST:
		g_value_set_pointer (value, sample->priv->taglist);
		break;

	case PROP_ENCODER:
		g_value_set_pointer (value, sample->priv->encoder);
		break;

	case PROP_UNDO_MANAGER:
		g_value_set_object (value, sample->priv->undo);
		break;

	default:
		break;
	}

	READ_UNLOCK (sample->priv->lock);
}

static void
class_init (GObjectClass *klass)
{
	klass->finalize = finalize;
	klass->set_property = set_property;
	klass->get_property = get_property;
	
	parent_class = g_type_class_peek_parent (klass);

	/* Properties */
	g_object_class_install_property (klass,
					 PROP_TOTAL,
					 g_param_spec_uint64 ("total_frames",
							      "",
							      "",
							      0,
							      G_MAXUINT64,
							      0,
							      G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_SAMPLE_RATE,
					 g_param_spec_uint ("sample_rate",
							    "",
							    "",
							    0,
							    G_MAXUINT,
							    0,
							    G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_FILENAME,
					 g_param_spec_string ("filename",
							      "",
							      "",
							      "Untitled",
							      G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_REALNAME,
					 g_param_spec_string ("name",
							      "", "",
							      "Untitled",
							      G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_CHANNELS,
					 g_param_spec_uint ("channels",
							    "",
							    "",
							    1,
							    MAX_NUM_CHANNELS,
							    2,
							    G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_DEPTH,
					 g_param_spec_int ("depth",
							   "",
							   "",
							   1,
							   G_MAXINT,
							   1,
							   G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_LENGTH,
					 g_param_spec_float ("length",
							     "",
							     "",
							     0,
							     G_MAXFLOAT,
							     0,
							     G_PARAM_READABLE));
	g_object_class_install_property (klass,
					 PROP_DIRTY,
					 g_param_spec_boolean ("dirty",
							       "",
							       "",
							       FALSE,
							       G_PARAM_READWRITE));

	g_object_class_install_property (klass,
					 PROP_WRITABLE,
					 g_param_spec_boolean ("writable",
							       "", "",
							       FALSE,
							       G_PARAM_READABLE));
	g_object_class_install_property (klass,
					 PROP_SELECTION,
					 g_param_spec_object ("selection",
							      "",
							      "",
							      MARLIN_SAMPLE_SELECTION_TYPE,
							      G_PARAM_READABLE));
	g_object_class_install_property (klass,
					 PROP_MARKER_MODEL,
					 g_param_spec_object ("markers",
							      "", "",
							      MARLIN_MARKER_MODEL_TYPE,
							      G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_TAGLIST,
					 g_param_spec_pointer ("taglist",
							       "", "",
							       G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_ENCODER,
					 g_param_spec_object ("encoder",
							      "", "",
							      GST_TYPE_ELEMENT,
							      G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_UNDO_MANAGER,
					 g_param_spec_object ("undo-manager",
							      "",
							      "",
							      MARLIN_UNDO_MANAGER_TYPE,
							      G_PARAM_READABLE));
							       
}

static void
init (MarlinSample *sample)
{
	sample->priv = g_new0 (MarlinSamplePrivate, 1);

	/* Create the lock */
	sample->priv->lock = marlin_read_write_lock_new ();
	
	sample->priv->channel_data = g_ptr_array_new ();
	sample->priv->header = g_new0 (MarlinSampleHeader, 1);
	
	sample->priv->selection = marlin_sample_selection_new (sample);
	sample->priv->undo = marlin_undo_manager_new ();

	sample->priv->markers = marlin_marker_model_new ();
	sample->priv->realname = NULL;

	sample->priv->undo = marlin_undo_manager_new ();

	sample->priv->taglist = gst_tag_list_new ();

	if (tmp_dir == NULL) {
		const char *tmp;

		tmp = g_getenv ("MARLIN_TMP_DIR");
		if (tmp != NULL) {
			tmp_dir = g_strdup (tmp);
		} else {
			tmp_dir = g_strdup (g_get_tmp_dir ());
		}
	}
}

GType
marlin_sample_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinSampleClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinSample), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (MARLIN_OBJECT_TYPE,
					       "MarlinSample",
					       &info, 0);
	}

	return type;
}

/**
 * marlin_sample_new:
 *
 * Creates a new stereo sample at 44100hz
 *
 * Returns: A newly created #MarlinSample.
 */
MarlinSample *
marlin_sample_new (void)
{
	MarlinSample *sample;

	sample = g_object_new (MARLIN_SAMPLE_TYPE,
			       "filename", "untitled",
			       "channels", 2,
			       "sample-rate", 44100,
			       NULL);
	
	return sample;
}

MarlinSample *
marlin_sample_new_from_sample_with_range (MarlinSample *src,
					  guint64 start_frame,
					  guint64 finish_frame,
					  MarlinCoverage coverage,
					  GError **error)
{
	MarlinSample *sample;
	MarlinChannel *src_channel, *dest_channel;
	guint sample_rate, channels;
	gboolean ret;
	guint64 total_frames;
	int i;
	
	g_return_val_if_fail (IS_MARLIN_SAMPLE (src), NULL);

	READ_LOCK (src->priv->lock);
	sample_rate = src->priv->header->sample_rate;
	channels = src->priv->header->channels;
	total_frames = src->priv->header->number_of_frames;
	READ_UNLOCK (src->priv->lock);
	
	sample = g_object_new (MARLIN_SAMPLE_TYPE,
			       "filename", "untitled",
			       "sample_rate", sample_rate,
			       "channels", coverage == MARLIN_COVERAGE_BOTH ? channels : 1,
			       NULL);

	switch (coverage) {
	case MARLIN_COVERAGE_BOTH:
		for (i = 0; i < channels; i++) {

			src_channel = src->priv->channel_data->pdata[i];
			dest_channel = sample->priv->channel_data->pdata[i];

			ret = marlin_channel_copy_data (src_channel, 
							dest_channel,
							start_frame, 
							finish_frame, error);
			if (ret == FALSE) {
				return NULL;
			}
		}
		break;

	case MARLIN_COVERAGE_LEFT:
		src_channel = src->priv->channel_data->pdata[0];
		dest_channel = sample->priv->channel_data->pdata[0];

		ret = marlin_channel_copy_data (src_channel, dest_channel,
						start_frame, finish_frame,
						error);
		if (ret == FALSE) {
			return NULL;
		}

		break;

	case MARLIN_COVERAGE_RIGHT:
		src_channel = src->priv->channel_data->pdata[1];
		dest_channel = sample->priv->channel_data->pdata[1];

		ret = marlin_channel_copy_data (src_channel, dest_channel,
						start_frame, finish_frame,
						error);
		if (ret == FALSE) {
			return NULL;
		}
		break;

	default:
		g_assert_not_reached ();
		break;
	}

	g_object_set (G_OBJECT (sample),
		      "total_frames", (guint64) finish_frame - start_frame,
		      NULL);
	
	return sample;
}

MarlinSample *
marlin_sample_new_from_sample (MarlinSample *src,
			       GError **error)
{
	guint64 number_of_frames;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (src), NULL);

	READ_LOCK (src->priv->lock);
	number_of_frames = src->priv->header->number_of_frames;
	READ_UNLOCK (src->priv->lock);
	
	return marlin_sample_new_from_sample_with_range (src, (guint64) 0,
							 number_of_frames, 
							 MARLIN_COVERAGE_BOTH,
							 error);
}

MarlinChannel *
marlin_sample_get_channel (MarlinSample *sample,
			   guint channel_num)
{
	MarlinChannel *channel;
	
	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), NULL);
	g_return_val_if_fail (channel_num < sample->priv->header->channels, NULL);

	READ_LOCK (sample->priv->lock);
	channel = g_ptr_array_index (sample->priv->channel_data, channel_num);
	READ_UNLOCK (sample->priv->lock);

	return channel;
}

gboolean
marlin_sample_generate_peaks (MarlinSample *sample,
			      MarlinOperation *operation,
			      GError **error)
{
	int i;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	READ_LOCK (sample->priv->lock);
	for (i = 0; i < sample->priv->channel_data->len; i++) {
		marlin_channel_generate_peaks (sample->priv->channel_data->pdata[i], operation, error);
	}
	READ_UNLOCK (sample->priv->lock);

	return TRUE;
}

gboolean
marlin_sample_ensure_mmapped (MarlinSample *sample,
			      MarlinOperation *operation,
			      GError **error)
{
	gboolean ret;
	int i;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	READ_LOCK (sample->priv->lock);
	for (i = 0; i < sample->priv->channel_data->len; i++) {
		ret = marlin_channel_ensure_mmapped (sample->priv->channel_data->pdata[i], operation, error);
		if (ret == FALSE) {
			READ_UNLOCK (sample->priv->lock);
			return FALSE;
		}
	}
	READ_UNLOCK (sample->priv->lock);

	return TRUE;
}

struct _delete_range_closure {
	MarlinSample *sample;
};

static void
delete_range_undo (gpointer data)
{
	struct _delete_range_closure *c = data;
	MarlinSample *sample = c->sample;
	MarlinChannel *channel;

	channel = sample->priv->channel_data->pdata[0];
	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      "total_frames", channel ? channel->frames : (guint64) 0,
		      NULL);
}

static void
delete_range_destroy (gpointer data)
{
	g_free (data);
}

gboolean
marlin_sample_delete_range (MarlinSample *sample,
			    MarlinOperation *operation,
			    guint64 start,
			    guint64 finish,
			    MarlinUndoContext *ctxt,
			    GError **error)
{
	MarlinChannel *channel;
	struct _delete_range_closure *c;
	MarlinUndoable *u;
	int i;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	c = g_new (struct _delete_range_closure, 1);
	c->sample = sample;

	u = marlin_undoable_new (delete_range_undo,
				 delete_range_undo,
				 delete_range_destroy,
				 c);
	marlin_undo_context_add (ctxt, u);

	WRITE_LOCK (sample->priv->lock);

	for (i = 0; i < sample->priv->channel_data->len; i++) {
		marlin_channel_delete_range (sample->priv->channel_data->pdata[i], start, finish, ctxt);
	}

	channel = sample->priv->channel_data->pdata[0];

	marlin_marker_model_remove_markers_in_range (sample->priv->markers,
						     start, finish, ctxt);
	marlin_marker_model_move_markers_after (sample->priv->markers,
						start, -(finish - start), ctxt);
	WRITE_UNLOCK (sample->priv->lock);

	g_object_set (G_OBJECT (sample),
		      "total-frames", channel ? channel->frames : (guint64) 0,
		      "dirty", TRUE,
		      NULL);

	return TRUE;
}

gboolean
marlin_sample_clear_range (MarlinSample *sample,
			   MarlinOperation *operation,
			   guint64 start,
			   guint64 finish,
			   MarlinCoverage coverage,
			   MarlinUndoContext *ctxt,
			   GError **error)
{
	MarlinChannel *channel;
	int i;
	gboolean ret;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	WRITE_LOCK (sample->priv->lock);

	switch (coverage) {
	case MARLIN_COVERAGE_BOTH:
		for (i = 0; i < sample->priv->channel_data->len; i++) {
			ret = marlin_channel_clear_range (sample->priv->channel_data->pdata[i], start, finish, ctxt, error);
			if (ret == FALSE) {
				WRITE_UNLOCK (sample->priv->lock);
				return FALSE;
			}
		}

		break;

	case MARLIN_COVERAGE_LEFT:
		ret = marlin_channel_clear_range (sample->priv->channel_data->pdata[0], start, finish, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		break;

	case MARLIN_COVERAGE_RIGHT:
		if (sample->priv->header->channels != 2) {
			g_warning ("Editting right channel on mono sample?");
			return;
		}
		
		ret = marlin_channel_clear_range (sample->priv->channel_data->pdata[1], start, finish, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		break;
	}

	WRITE_UNLOCK (sample->priv->lock);
	
	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      NULL);

	return TRUE;
}

struct _crop_range_closure {
	MarlinSample *sample;
};

static void
crop_range_undo (gpointer data)
{
	struct _crop_range_closure *c = data;
	MarlinSample *sample = c->sample;
	MarlinChannel *channel;

	channel = sample->priv->channel_data->pdata[0];
	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      "total_frames", channel ? channel->frames : (guint64) 0,
		      NULL);
}

static void
crop_range_destroy (gpointer data)
{
	g_free (data);
}

gboolean
marlin_sample_crop_range (MarlinSample *sample,
			  MarlinOperation *operation,
			  guint64 start,
			  guint64 finish,
			  MarlinUndoContext *ctxt,
			  GError **error)
{
	MarlinChannel *channel;
	struct _crop_range_closure *c;
	MarlinUndoable *u;
	int i;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	c = g_new (struct _crop_range_closure, 1);
	c->sample = sample;

	u = marlin_undoable_new (crop_range_undo,
				 crop_range_undo,
				 crop_range_destroy,
				 c);
	marlin_undo_context_add (ctxt, u);

	WRITE_LOCK (sample->priv->lock);

	marlin_marker_model_remove_markers_in_range (sample->priv->markers,
						     (guint64) 0, start, ctxt);
	marlin_marker_model_remove_markers_in_range (sample->priv->markers,
						     finish,
						     sample->priv->header->number_of_frames, ctxt);
	marlin_marker_model_move_markers_after (sample->priv->markers, start,
						-start, ctxt);

	for (i = 0; i < sample->priv->channel_data->len; i++) {
		marlin_channel_crop_range (sample->priv->channel_data->pdata[i], start, finish, ctxt);
	}

	channel = sample->priv->channel_data->pdata[0];
	
	WRITE_UNLOCK (sample->priv->lock);

	/* This is an ugly hack... FIXMEFIXME */
	c = g_new (struct _crop_range_closure, 1);
	c->sample = sample;

	u = marlin_undoable_new (crop_range_undo,
				 crop_range_undo,
				 crop_range_destroy,
				 c);
	marlin_undo_context_add (ctxt, u);

	g_object_set (G_OBJECT (sample),
		      "total-frames", channel ? channel->frames : (guint64) 0,
		      "dirty", TRUE,
		      NULL);

	return TRUE;
}

struct _swap_channels_closure {
	MarlinSample *sample;
};

static void
swap_channels_undo (gpointer data)
{
	struct _swap_channels_closure *s = data;

	marlin_sample_swap_channels (s->sample, NULL, NULL, NULL);
}

static void
swap_channels_destroy (gpointer data)
{
	g_free (data);
}
	
gboolean
marlin_sample_swap_channels (MarlinSample *sample,
			     MarlinOperation *operation,
			     MarlinUndoContext *ctxt,
			     GError **error)
{
	MarlinSamplePrivate *priv;
	MarlinChannel *tmp;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	priv = sample->priv;
	WRITE_LOCK (priv->lock);

	if (priv->header->channels != 2) {
		WRITE_UNLOCK (priv->lock);

		if (*error != NULL) {
			/* Make GError */
		}
		
		return;
	}

	tmp = priv->channel_data->pdata[0];
	priv->channel_data->pdata[0] = priv->channel_data->pdata[1];
	priv->channel_data->pdata[1] = tmp;

	WRITE_UNLOCK (priv->lock);

	if (ctxt) {
		struct _swap_channels_closure *c;
		MarlinUndoable *u;

		c = g_new (struct _swap_channels_closure, 1);
		c->sample = sample;

		u = marlin_undoable_new (swap_channels_undo,
					 swap_channels_undo,
					 swap_channels_destroy,
					 c);

		marlin_undo_context_add (ctxt, u);
	}

	if (priv->header->number_of_frames != 0) {
		g_object_set (G_OBJECT (sample),
			      "dirty", TRUE,
			      NULL);
	}

	return TRUE;
}

gboolean
marlin_sample_mix (MarlinSample *dest,
		   MarlinSample *src,
		   double src_db,
		   double dest_db,
		   guint64 start_frame,
		   guint64 end_frame,
		   gboolean clip,
		   MarlinOperation *operation,
		   MarlinUndoContext *ctxt,
		   GError **error)
{
	int chan;
	MarlinSamplePrivate *s_priv, *d_priv;
	MarlinChannel *channel;
	
	g_return_val_if_fail (IS_MARLIN_SAMPLE (dest), FALSE);
	g_return_val_if_fail (IS_MARLIN_SAMPLE (src), FALSE);

	s_priv = src->priv;
	d_priv = dest->priv;

	if (s_priv->header->channels != d_priv->header->channels) {
		g_warning ("Channel mismatch");

		/* New GError */
		return FALSE;
	}

	WRITE_LOCK (s_priv->lock);
	READ_LOCK (d_priv->lock);
	
	for (chan = 0; chan < s_priv->header->channels; chan++) {
		MarlinChannel *s_chan, *d_chan;

		s_chan = s_priv->channel_data->pdata[chan];
		d_chan = d_priv->channel_data->pdata[chan];

		marlin_channel_mix (d_chan, s_chan, src_db, dest_db, start_frame, end_frame, clip, operation, ctxt, error);
	}

	channel = d_priv->channel_data->pdata[0];
	
	WRITE_UNLOCK (s_priv->lock);
	READ_UNLOCK (d_priv->lock);

	g_object_set (G_OBJECT (dest),
 		      "total-frames", channel->frames,
		      "dirty", TRUE,
		      NULL);

	return TRUE;
}

struct _insert_closure {
	MarlinSample *sample;
};

static void
insert_undo (gpointer data)
{
	struct _insert_closure *c = data;
	MarlinSample *sample = c->sample;
	MarlinChannel *channel;

	channel = sample->priv->channel_data->pdata[0];
	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      "total_frames", channel->frames,
		      NULL);
}

static void
insert_destroy (gpointer data)
{
	g_free (data);
}

gboolean
marlin_sample_insert (MarlinSample *dest,
		      MarlinSample *src,
		      guint64 position,
		      MarlinCoverage coverage,
		      MarlinUndoContext *ctxt,
		      GError **error)
{
	guint src_channels, dest_channels, i;
	MarlinChannel *channel;
	struct _insert_closure *c;
	MarlinUndoable *u;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (src), FALSE);
	g_return_val_if_fail (IS_MARLIN_SAMPLE (dest), FALSE);

	src_channels = src->priv->header->channels;
	dest_channels = dest->priv->header->channels;

	if (src_channels != dest_channels) {
		g_warning ("Channel mismatch");
		return;
	}

	READ_LOCK (src->priv->lock);
	WRITE_LOCK (dest->priv->lock);
	
	for (i = 0; i < src_channels; i++) {
		MarlinChannel *s_channel = src->priv->channel_data->pdata[i];
		MarlinChannel *d_channel = dest->priv->channel_data->pdata[i];
		gboolean ret;

		ret = marlin_channel_insert (d_channel, s_channel, position, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (dest->priv->lock);
			WRITE_UNLOCK (src->priv->lock);
			return FALSE;
		}
	}

	WRITE_UNLOCK (dest->priv->lock);
	READ_UNLOCK (src->priv->lock);

	c = g_new (struct _insert_closure, 1);
	c->sample = dest;

	u = marlin_undoable_new (insert_undo,
				 insert_undo,
				 insert_destroy,
				 c);

	marlin_undo_context_add (ctxt, u);

	channel = dest->priv->channel_data->pdata[0];
	
	g_object_set (G_OBJECT (dest),
		      "total_frames", channel->frames,
		      "dirty", TRUE,
		      NULL);

	return TRUE;
}

gboolean
marlin_sample_invert_range (MarlinSample *sample,
			    guint64 start,
			    guint64 finish,
			    MarlinCoverage coverage,
			    MarlinUndoContext *ctxt,
			    GError **error)
{
	int i;
	gboolean ret;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	WRITE_LOCK (sample->priv->lock);
	
	switch (coverage) {
	case MARLIN_COVERAGE_BOTH:
		for (i = 0; i < sample->priv->header->channels; i++) {
			ret = marlin_channel_invert (sample->priv->channel_data->pdata[i], start, finish, ctxt, error);
			if (ret == FALSE) {
				WRITE_UNLOCK (sample->priv->lock);
				return FALSE;
			}
		}
		break;

	case MARLIN_COVERAGE_LEFT:
		ret = marlin_channel_invert (sample->priv->channel_data->pdata[0],
					     start, finish, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		break;

	case MARLIN_COVERAGE_RIGHT:
		if (sample->priv->header->channels != 2) {
			g_warning ("Attempting to invert RIGHT in a mono sample");
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}
		ret = marlin_channel_invert (sample->priv->channel_data->pdata[1],
					     start, finish, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		break;
	}

	WRITE_UNLOCK (sample->priv->lock);
	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      NULL);

	return TRUE;
}

gboolean
marlin_sample_adjust_volume_range (MarlinSample *sample,
				   float db,
				   guint64 start,
				   guint64 finish,
				   MarlinCoverage coverage,
				   MarlinUndoContext *ctxt,
				   GError **error)
{
	int i;
	gboolean ret;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	WRITE_LOCK (sample->priv->lock);

	switch (coverage) {
	case MARLIN_COVERAGE_BOTH:
		for (i = 0; i < sample->priv->header->channels; i++) {
			ret = marlin_channel_adjust_volume (sample->priv->channel_data->pdata[i], db, start, finish, ctxt, error);

			if (ret == FALSE) {
				WRITE_UNLOCK (sample->priv->lock);
				return FALSE;
			}
		}
		break;

	case MARLIN_COVERAGE_LEFT:
		ret = marlin_channel_adjust_volume (sample->priv->channel_data->pdata[0], db, start, finish, ctxt, error);

		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		break;

	case MARLIN_COVERAGE_RIGHT:
		if (sample->priv->header->channels != 2) {
			g_warning ("Attempting to adjust volume on RIGHT in a mono sample.");
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		ret = marlin_channel_adjust_volume (sample->priv->channel_data->pdata[1], db, start, finish, ctxt, error);
		
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		break;
	}

	WRITE_UNLOCK (sample->priv->lock);

	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      NULL);

	return TRUE;
}

struct _insert_silence_closure {
	MarlinSample *sample;
};

static void
insert_silence_undo (gpointer closure)
{
	struct _insert_silence_closure *isc = closure;
	MarlinSample *sample = isc->sample;
	MarlinChannel *channel;

	channel = sample->priv->channel_data->pdata[0];
	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      "total_frames", channel->frames,
		      NULL);
}

static void
insert_silence_destroy (gpointer closure)
{
	g_free (closure);
}

gboolean
marlin_sample_insert_silence (MarlinSample *sample,
			      guint64 position,
			      guint64 length,
			      MarlinUndoContext *ctxt,
			      GError **error)
{
	MarlinChannel *channel;
	int i;
	gboolean ret;
	MarlinUndoable *undo;
	struct _insert_silence_closure *isc;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	isc = g_new (struct _insert_silence_closure, 1);
	isc->sample = sample;

	undo = marlin_undoable_new (insert_silence_undo,
				    insert_silence_undo,
				    insert_silence_destroy,
				    isc);

	marlin_undo_context_add (ctxt, undo);

	WRITE_LOCK (sample->priv->lock);

	for (i = 0; i < sample->priv->header->channels; i++) {
		ret = marlin_channel_insert_silence (sample->priv->channel_data->pdata[i],
						     position, length, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}
	}

	marlin_marker_model_move_markers_after (sample->priv->markers,
						position, (gint64) length, ctxt);
	WRITE_UNLOCK (sample->priv->lock);

	isc = g_new (struct _insert_silence_closure, 1);
	isc->sample = sample;

	undo = marlin_undoable_new (insert_silence_undo,
				    insert_silence_undo,
				    insert_silence_destroy,
				    isc);

	marlin_undo_context_add (ctxt, undo);

	channel = sample->priv->channel_data->pdata[0];
	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      "total_frames", channel->frames,
		      NULL);

	return TRUE;
}

struct _remove_channel_closure {
	MarlinSample *sample;
	MarlinChannel *channel;
};

static void
remove_channel_undo (gpointer data)
{
	struct _remove_channel_closure *c = data;
	MarlinChannel *channel;
	gboolean ret;

	g_object_set (G_OBJECT (c->sample),
		      "channels", 2,
		      NULL);

	/* The newly added channel is always the second one */
	channel = c->sample->priv->channel_data->pdata[1];
	ret = marlin_channel_copy_data (c->channel, channel, 0, c->channel->frames - 1, NULL);
	if (ret == FALSE) {
		g_warning ("marlin_channel_copy_data failed");
	}

	g_object_set (G_OBJECT (c->sample),
		      "dirty", TRUE, 
		      NULL);
}

static void
remove_channel_redo (gpointer data)
{
	struct _remove_channel_closure *c = data;

	g_object_set (G_OBJECT (c->sample),
		      "channels", 1,
		      "dirty", TRUE,
		      NULL);
}

static void
remove_channel_destroy (gpointer data)
{
	struct _remove_channel_closure *c = data;

	marlin_channel_free (c->channel);
	g_free (c);
}

/**
 * marlin_sample_remove_channel:
 * sample:
 * channel:
 * mix:
 * operation:
 * error:
 *
 * Removes the channel indicated by @channel from @sample. If @mix is TRUE
 * then the data in @channel is mixed into the remaining channel.
 *
 * Returns: TRUE on success, FALSE on error with further details in @error.
 */
gboolean
marlin_sample_remove_channel (MarlinSample *sample,
			      MarlinChannelPosition channel,
			      gboolean mix,
			      MarlinOperation *operation,
			      MarlinUndoContext *ctxt,
			      GError **error)
{
	struct _remove_channel_closure *c;
	MarlinUndoable *u;
	MarlinChannel *chan;
	gboolean ret;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	WRITE_LOCK (sample->priv->lock);

	if (sample->priv->header->channels == 1) {
		WRITE_UNLOCK (sample->priv->lock);
		return TRUE;
	}

	if (channel == MARLIN_CHANNEL_LEFT) {
		WRITE_UNLOCK (sample->priv->lock);
		/* Swap the channels around */
		ret = marlin_sample_swap_channels (sample, operation, ctxt, error);

		if (ret == FALSE) {
			return;
		}

		WRITE_LOCK (sample->priv->lock);
	}

	if (mix) {
		MarlinChannel *src, *dest;

		dest = sample->priv->channel_data->pdata[MARLIN_CHANNEL_LEFT];
		src = sample->priv->channel_data->pdata[MARLIN_CHANNEL_RIGHT];

		ret = marlin_channel_mix (dest, src, 0.0, 0.0, (guint64) 0, src->frames, FALSE, operation, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}
	}

	WRITE_UNLOCK (sample->priv->lock);

	if (ctxt) {
		c = g_new (struct _remove_channel_closure, 1);
		c->sample = sample;
		
		/* Create a new channel and copy the old one */
		chan = sample->priv->channel_data->pdata[1];
		c->channel = marlin_channel_new (sample, sample->priv->filename, NULL);
		marlin_channel_copy_data (chan, c->channel, 0, chan->frames - 1, NULL);
		u = marlin_undoable_new (remove_channel_undo,
					 remove_channel_redo,
					 remove_channel_destroy,
					 c);
		
		marlin_undo_context_add (ctxt, u);
	}

	/* Can't be dirty if there's no samples */
	if (sample->priv->header->number_of_frames == 0) {
		g_object_set (G_OBJECT (sample), 
			      "channels", 1,
			      NULL);
	} else {
		g_object_set (G_OBJECT (sample),
			      "channels", 1,
			      "dirty", TRUE,
			      NULL);
	}
}

struct _add_channel_closure {
	MarlinSample *sample;
	MarlinChannel *channel;
	gboolean clone;
};

static void
add_channel_undo (gpointer data)
{
	struct _add_channel_closure *c = data;

	g_object_set (G_OBJECT (c->sample),
		      "channels", 1,
		      NULL);
}

static void
add_channel_redo (gpointer data)
{
	struct _add_channel_closure *c = data;

	g_object_set (G_OBJECT (c->sample),
		      "channels", 2,
		      NULL);

	if (c->clone) {
		MarlinChannel *channel;
		gboolean ret;
		
		channel = c->sample->priv->channel_data->pdata[1];
		ret = marlin_channel_copy_data (c->channel, channel, 0, c->channel->frames - 1, NULL);
		if (ret == FALSE) {
			g_warning ("marlin_channel_copy_data failed");
		}
	}
}

static void
add_channel_destroy (gpointer data)
{
	struct _add_channel_closure *c = data;

	if (c->channel) {
		marlin_channel_free (c->channel);
	}
	g_free (data);
}

/**
 * marlin_sample_add_channel:
 * sample:
 * channel:
 * clone:
 * operation:
 * error:
 *
 * Adds a channel to @sample in the position indicated by @channel. If @clone
 * is TRUE, then the data from the other channel is cloned in the new channel.
 *
 * Returns: TRUE on success, FALSE on error, with extra details in @error.
 */
gboolean
marlin_sample_add_channel (MarlinSample *sample,
			   MarlinChannelPosition channel,
			   gboolean clone,
			   MarlinOperation *operation,
			   MarlinUndoContext *ctxt,
			   GError **error)
{
	gboolean ret;
	struct _add_channel_closure *c = NULL;
	MarlinUndoable *u;

	g_return_val_if_fail (IS_MARLIN_SAMPLE (sample), FALSE);

	if (sample->priv->header->channels == 2) {
		return TRUE;
	}

	/* Add our new channel */
	g_object_set (G_OBJECT (sample),
		      "channels", 2,
		      NULL);

	if (ctxt) {
		c = g_new (struct _add_channel_closure, 1);
		c->sample = sample;
		c->clone = clone;
		c->channel = NULL;

		u = marlin_undoable_new (add_channel_undo,
					 add_channel_redo,
					 add_channel_destroy,
					 c);

		marlin_undo_context_add (ctxt, u);
	}

	WRITE_LOCK (sample->priv->lock);

	if (clone == FALSE) {
		MarlinChannel *src;

		src = sample->priv->channel_data->pdata[MARLIN_CHANNEL_LEFT];

		/* Fill the new channel with 0 data */
		ret = marlin_channel_insert_silence (sample->priv->channel_data->pdata[MARLIN_CHANNEL_RIGHT],
					       (guint64) 0, src->frames, ctxt, error);
		if (ret == FALSE) {
			WRITE_UNLOCK (sample->priv->lock);
			return FALSE;
		}

		WRITE_UNLOCK (sample->priv->lock);

		if (channel == MARLIN_CHANNEL_LEFT) {
			ret = marlin_sample_swap_channels (sample, operation, ctxt, error);
			if (ret == FALSE) {
				return FALSE;
			}
		}
	} else {
		MarlinChannel *src, *dest;
		
		src = sample->priv->channel_data->pdata[MARLIN_CHANNEL_LEFT];
		dest = sample->priv->channel_data->pdata[MARLIN_CHANNEL_RIGHT];
		
		ret = marlin_channel_copy_data (src, dest, (guint64) 0, src->frames - 1, error);

		if (ctxt) {
			gboolean ret2;

			c->channel = marlin_channel_new (sample, sample->priv->filename, NULL);
			ret2 = marlin_channel_copy_data (src, c->channel, 0,
							 src->frames - 1, NULL);
			if (ret2 == FALSE) {
				g_warning ("marlin_channel_copy_data failed to make UNDO");
			}
		}

		WRITE_UNLOCK (sample->priv->lock);
		if (ret == FALSE) {
			return FALSE;
		}
	}

	/* Can't be dirty if there's no data */
	if (sample->priv->header->number_of_frames != 0) {
		g_object_set (G_OBJECT (sample),
			      "dirty", TRUE,
			      NULL);
	}

	return TRUE;
}

/**
 * marlin_sample_read_lock
 * sample:
 *
 * Locks @sample for reading.
 */
void
marlin_sample_read_lock (MarlinSample *sample)
{
	g_return_if_fail (IS_MARLIN_SAMPLE (sample));

	READ_LOCK (sample->priv->lock);
}

/**
 * marlin_sample_read_unlock
 * sample:
 *
 * Unlocks @sample
 */
void
marlin_sample_read_unlock (MarlinSample *sample)
{
	g_return_if_fail (IS_MARLIN_SAMPLE (sample));

	READ_UNLOCK (sample->priv->lock);
}
