/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <gtk/gtkdialog.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkstock.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkprogressbar.h>

#include <marlin/marlin-sample.h>
#include <marlin/marlin-utils.h>

#include "marlin-window.h"
#include "marlin-progress-window.h"
#include "marlin-progress-window-icons.h"

/* The delay between the window being created and it being shown...
   Allows short operations to not have to open a window at all */
#define DELAY_TIME 1200

static GdkPixbuf *empty_jar_pixbuf = NULL, *full_jar_pixbuf = NULL;

static char *
make_eta_string (const char *first,
		 const char *second)
{
	return g_strdup_printf ("<span size=\"smaller\">%s: %s</span>",
				first, second);
}

static void
update_icon (MarlinProgressWindow *window,
	     double fraction)
{
	GdkPixbuf *pixbuf;
	int position;

	position = gdk_pixbuf_get_height (empty_jar_pixbuf) * (1 - fraction);

	if (position == window->progress_jar_position) {
		return;
	}

	window->progress_jar_position = position;

	pixbuf = gdk_pixbuf_copy (empty_jar_pixbuf);
	gdk_pixbuf_copy_area (full_jar_pixbuf,
			      0, position,
			      gdk_pixbuf_get_width (pixbuf),
			      gdk_pixbuf_get_height (pixbuf) - position,
			      pixbuf,
			      0, position);

	marlin_window_set_icon (MARLIN_WINDOW (window->parent_window), pixbuf);
	g_object_unref (pixbuf);
}

static void
operation_progress (MarlinOperation *operation,
		    int progress,
		    MarlinProgressWindow *window)
{
	char *text, *title, *t;
	double dt, length;
	int left;
	
	if (strstr (window->format, "%") != NULL) {
		title = g_strdup_printf (window->format, window->filename);
	} else {
		title = g_strdup (window->format);
	}
	
	text = g_strdup_printf("%s: (%d%%)", title, progress);
	gtk_label_set_text (GTK_LABEL (window->label), text);
	g_free (title);
	g_free (text);

	left = 100 - progress;
	
	dt = g_timer_elapsed (window->timer, NULL);

	length = (dt / (double) progress) * left;

	t = marlin_ms_to_pretty_time ((int)length * 1000);

	text = make_eta_string (_("Estimated time remaining"), t);
	gtk_label_set_markup (GTK_LABEL (window->eta), text);
	g_free (t);
	g_free (text);
	
	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (window->progress_bar),
				       (float)progress / 100.0);

	update_icon (window, (float) progress / 100.0);
}

static void
response_callback (GtkDialog *dialog,
		   guint response_id,
		   MarlinProgressWindow *window)
{
	/* There's only one button, duh! */
	marlin_operation_cancel (window->operation);
}

static void
operation_notify (MarlinOperation *operation,
		  GParamSpec *pspec,
		  MarlinProgressWindow *window)
{
	const char *name;
	
	name = g_param_spec_get_name (pspec);
	if (strcmp (name, "description-R") == 0) {
		char *desc;
		
		g_object_get (G_OBJECT (operation),
			      "description", &desc,
			      NULL);

 		gtk_label_set (GTK_LABEL (window->label), desc);
		g_free (window->format);
		window->format = desc;
	}
}

static gboolean
show_progress_window (MarlinProgressWindow *window)
{
	gtk_widget_show (window->dialog);

	window->visible_id = 0;
	return FALSE;
}

MarlinProgressWindow *
marlin_progress_window_new (MarlinSample *sample,
			    MarlinOperation *operation,
			    GtkWindow *parent_window,
			    const char *title_format)
{
	GtkWidget *hbox, *icon, *vbox;
	char *title, *text;
	MarlinProgressWindow *window;
	GdkCursor *cursor;
	
	window = g_new0 (MarlinProgressWindow, 1);

	if (empty_jar_pixbuf == NULL) {
		empty_jar_pixbuf = gdk_pixbuf_new_from_inline (-1, progress_jar_empty_icon, FALSE, NULL);
		full_jar_pixbuf = gdk_pixbuf_new_from_inline (-1, progress_jar_full_icon, FALSE, NULL);
	}

	window->sample = sample;
	window->operation = operation;
	window->op_id = g_signal_connect (G_OBJECT (window->operation), "notify",
					  G_CALLBACK (operation_notify), window);
			  
	window->parent_window = GTK_WIDGET (parent_window);
	g_object_get (G_OBJECT (sample),
		      "filename", &window->filename,
		      NULL);
	
	window->format = g_strdup (title_format);
	
	title = g_strdup_printf (title_format, window->filename);
	window->dialog = gtk_dialog_new_with_buttons (title, parent_window,
						      GTK_DIALOG_DESTROY_WITH_PARENT |
						      GTK_DIALOG_NO_SEPARATOR,
						      GTK_STOCK_CANCEL,
						      GTK_RESPONSE_CANCEL, NULL);
	gtk_widget_realize (window->dialog);
	cursor = gdk_cursor_new (GDK_WATCH);
	gdk_window_set_cursor (window->dialog->window, cursor);

	g_signal_connect (G_OBJECT (window->dialog), "response",
			  G_CALLBACK (response_callback), window);
			  
/* 	gtk_container_set_border_width (GTK_CONTAINER (GTK_DIALOG (window->dialog)->vbox), MARLIN_DIALOG_BORDER_WIDTH); */
	gtk_window_set_transient_for (GTK_WINDOW (window->dialog), parent_window);

	window->signal_id = g_signal_connect (G_OBJECT (operation), "progress",
					      G_CALLBACK (operation_progress), window);

	hbox = gtk_hbox_new (FALSE, 12);
 	gtk_container_set_border_width (GTK_CONTAINER (hbox), 12);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window->dialog)->vbox),
			    hbox, TRUE, TRUE, 0);
	gtk_widget_show (hbox);
	icon = gtk_image_new_from_stock (GTK_STOCK_DIALOG_INFO,
					 GTK_ICON_SIZE_DIALOG);
	gtk_misc_set_alignment (GTK_MISC (icon), 0.0, 0.0);
	gtk_box_pack_start (GTK_BOX (hbox), icon, FALSE, FALSE, 0);
	gtk_widget_show (icon);

	vbox = gtk_vbox_new (FALSE, 2);
	gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 0);
	gtk_widget_show (vbox);
	
	text = g_strdup_printf ("%s: (0%%)", title);
	window->label = gtk_label_new (text);
	gtk_misc_set_alignment (GTK_MISC (window->label), 0.0, 0.0);
	gtk_label_set_line_wrap (GTK_LABEL (window->label), TRUE);
	g_free (title);
	g_free (text);
	gtk_box_pack_start (GTK_BOX (vbox), window->label, FALSE, FALSE, 0);
	gtk_widget_show (window->label);

	text = make_eta_string (_("Estimated time remaining"), _("Unknown"));
	window->eta = gtk_label_new (text);
	gtk_label_set_use_markup (GTK_LABEL (window->eta), TRUE);
	gtk_misc_set_alignment (GTK_MISC (window->eta), 0.0, 0.5);
	g_free (text);

	gtk_box_pack_start (GTK_BOX (vbox), window->eta, FALSE, FALSE, 0);
	gtk_widget_show (window->eta);
	
	window->progress_bar = gtk_progress_bar_new ();
	gtk_box_pack_start (GTK_BOX (vbox), window->progress_bar,
			    FALSE, FALSE, 0);

	window->timer = g_timer_new ();

	gtk_widget_show (window->progress_bar);

	window->visible_id = g_timeout_add (DELAY_TIME, (GSourceFunc *) show_progress_window, window);

	/* Set icon for parent window as the progress is set as being
	   transient for it, so it shouldn't show up in the taskbar */
	marlin_window_set_icon (MARLIN_WINDOW (window->parent_window), empty_jar_pixbuf);

	return window;
}

void
marlin_progress_window_destroy (MarlinProgressWindow *window)
{
	gtk_widget_destroy (window->dialog);

	g_signal_handler_disconnect (G_OBJECT (window->operation),
				     window->signal_id);
	g_signal_handler_disconnect (G_OBJECT (window->operation),
				     window->op_id);

	if (window->visible_id > 0) {
		g_source_remove (window->visible_id);
	}

	g_free (window->filename);
	g_free (window->format);

	marlin_window_set_icon (MARLIN_WINDOW (window->parent_window), NULL);

	g_timer_destroy (window->timer);
	g_free (window);
}
