/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002-2004 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <time.h>

#include <uuid/uuid.h>

#include <marlin-window.h>

#include <glib.h>

#include <gdk/gdkkeysyms.h>

#include <gtk/gtk.h>

#include <libgnome/gnome-i18n.h>

#include <libgnomevfs/gnome-vfs-utils.h>
#include <libgnomevfs/gnome-vfs-mime.h>

#include <gconf/gconf-client.h>

#ifdef HAVE_MEDIA_PROFILES
#include <profiles/gnome-media-profiles.h>
#endif

#include <marlin/marlin-sample.h>
#include <marlin/marlin-sample-selection.h>
#include <marlin/marlin-sample-view.h>
#include <marlin/marlin-operation.h>
#include <marlin/marlin-overview-bar.h>
#include <marlin/marlin-play-pipeline.h>
#include <marlin/marlin-load-pipeline.h>
#include <marlin/marlin-save-pipeline.h>
#include <marlin/marlin-file-utils.h>
#include <marlin/marlin-utils.h>
#include <marlin/marlin-program.h>
#include <marlin/marlin-position-spinner.h>
#include <marlin/marlin-marker-view.h>
#include <marlin/marlin-marker-model.h>
#include <marlin/marlin-a11y-helper.h>
#include <marlin/marlin-x-utils.h>
#include <marlin/marlin-cross-fader.h>
#include <marlin/marlin-undo-manager.h>

#include <egg-recent-model.h>
#include <egg-recent-view.h>

#include <dialogs/marlin-information-dialog.h>
#include <dialogs/marlin-adjust-channel-dialogs.h>
#include <dialogs/marlin-mp3-dialog.h>
#include <dialogs/marlin-vorbis-dialog.h>

#include <marlin-toolbar.h>

#include "main.h"
#include "marlin-window.h"
#include "marlin-recent.h"
#include "marlin-recent-view.h"

typedef enum _MarlinWindowPositionDisplay {
	AS_FRAMES,
	AS_TIME_LONG,
	AS_SECONDS,
	AS_TIME_FRAMES,
	LAST_PD
} MarlinWindowPositionDisplay;

enum {
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_SAMPLE,
};

enum {
	DND_TYPE_TEXT_URI_LIST
};

static GtkTargetEntry drop_types[] = {
	{ "text/uri-list", 0, DND_TYPE_TEXT_URI_LIST }
};

static int num_drop_types = sizeof (drop_types) / sizeof (drop_types[0]);

typedef enum _MarlinState {
	MARLIN_STATE_STOPPED,
	MARLIN_STATE_PLAYING,
	MARLIN_STATE_PAUSED,
	MARLIN_STATE_RECORDING
} MarlinState;

struct _MarlinWindowPrivate {
	GtkWidget *main_vbox;
	GtkWidget *menu_dock;
	GtkWidget *menubar;

	Toolbar *toolbar;
	GList *toolbars;
	
	MarlinSample *sample;
	guint notify_id;
	
	MarlinSampleSelection *selection;
	guint selection_id;

	MarlinSampleView *view;
	MarlinProgressWindow *progress_window;
	MarlinProgressWindow *save_progress_window;
	MarlinOverviewBar *overview_bar;

	MarlinUndoManager *undo;
	guint undo_changed_id;

	GtkWidget *marker_view, *vruler, *hscrollbar;
	GtkWidget *message_status;
	guint help_message_cid;

	MarlinMarkerModel *marker_model;
	guint marker_added_id, marker_removed_id;
	
	MarlinWindowPositionDisplay position_display;
	GtkWidget *position_status;
	GtkWidget *sel_start_status;
	GtkWidget *sel_end_status;

	/* UI stuff */
	GtkAccelGroup *accel_group;
	GtkActionGroup *ui_action, *popup_action;
	GObject *ui_merge;

	MarlinRecentView *recent_view;
	
	MarlinPlayPipeline *play_pipeline;
	
	MarlinState state;
	guint64 initial; /* The initial position of the cursor when play
			    starts */

	/* Start finish for looping and selections */
	guint64 start, finish;
	gboolean repeating;

	/* Tell the op_handler what to do */
	gboolean close_on_op_complete;

	guint clip_id; /* Signal handler for listening to the clipboard */

	GtkWidget *select_region_dialog;
	GtkWidget *move_cursor_dialog;
	GtkWidget *properties_dialog;
	GtkWidget *xfade_dialog;
	GtkWidget *adjust_channels_dialog;
	
	time_t last_save; /* Time of the last save, in seconds */
};

static void set_path_sensitive (GtkActionGroup *ag,
				const char *name,
				gboolean sensitive);
void marlin_window_set_file_sensitive (MarlinWindow *window,
				       gboolean sensitive);
void marlin_window_set_media_bar_sensitive (MarlinWindow *window,
					    gboolean sensitive);
void marlin_window_set_view_sensitive (MarlinWindow *window,
				       gboolean sensitive);
void marlin_window_set_edit_sensitive (MarlinWindow *window,
				       gboolean sensitive);
void marlin_window_set_selection_sensitive (MarlinWindow *window,
					    gboolean sensitive);
void marlin_window_set_process_sensitive (MarlinWindow *window,
					  gboolean sensitive);
void marlin_window_set_process_name (MarlinWindow *window);
void marlin_window_set_edit_name (MarlinWindow *window);
void marlin_window_pause_sample (MarlinWindow *window);
void marlin_window_save_sample_as (MarlinWindow *window);

static GtkWindowClass *parent_class = NULL;
static guint signals[LAST_SIGNAL];

static GdkPixbuf *window_icon = NULL;

#define DEFAULT_WIDTH 600
#define DEFAULT_HEIGHT 330

static void
finalize (GObject *object)
{
	MarlinWindow *window;
	MarlinWindowPrivate *priv;

	window = MARLIN_WINDOW (object);
	priv = window->priv;

	if (priv == NULL) {
		return;
	}

	if (priv->sample != NULL) {
		g_signal_handler_disconnect (G_OBJECT (priv->sample),
					     priv->notify_id);
		g_object_unref (G_OBJECT (priv->sample));
	}

	if (priv->selection != NULL) {
		g_signal_handler_disconnect (G_OBJECT (priv->selection),
					     priv->selection_id);
		g_object_unref (G_OBJECT (priv->selection));
	}

	if (priv->marker_model != NULL) {
		g_signal_handler_disconnect (G_OBJECT (priv->marker_model),
					     priv->marker_added_id);
		g_signal_handler_disconnect (G_OBJECT (priv->marker_model),
					     priv->marker_removed_id);
		g_object_unref (G_OBJECT (priv->marker_model));
	}

	g_object_unref (G_OBJECT (priv->play_pipeline));

	if (priv->undo != NULL) {
		g_signal_handler_disconnect (G_OBJECT (priv->undo),
					     priv->undo_changed_id);
		g_object_unref (G_OBJECT (priv->undo));
	}

	if (priv->select_region_dialog != NULL) {
		gtk_widget_destroy (priv->select_region_dialog);
	}

	if (priv->move_cursor_dialog != NULL) {
		gtk_widget_destroy (priv->move_cursor_dialog);
	}

	if (priv->properties_dialog != NULL) {
		gtk_widget_destroy (priv->properties_dialog);
	}
	
	g_free (priv);
	window->priv = NULL;
	
	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
dispose (GObject *object)
{
	MarlinWindow *window = MARLIN_WINDOW (object);

	if (window->priv->clip_id > 0) {
		MarlinProgram *prog = marlin_program_get_default ();
		g_signal_handler_disconnect (G_OBJECT (prog),
					     window->priv->clip_id);
		window->priv->clip_id = 0;
	}

	G_OBJECT_CLASS (parent_class)->dispose (object);
}
 
static void
marlin_window_sample_notify (MarlinSample *sample,
			     const char *name,
			     MarlinWindow *window)
{
	guint fpp;

	if (strcmp (name, "total-frames") == 0 ||
	    strcmp (name, "dirty") == 0) {
		g_object_get (G_OBJECT (window->priv->view),
			      "frames_per_pixel", &fpp,
			      NULL);

		g_object_set (G_OBJECT (window->priv->overview_bar),
			      "frames_per_page", (guint64) fpp * GTK_WIDGET (window->priv->view)->allocation.width,
			      NULL);

		marlin_window_set_media_bar_sensitive (window, TRUE);
		marlin_window_set_file_sensitive (window, TRUE);
		marlin_window_set_title (window, sample);
		marlin_window_set_process_name (window);
		marlin_window_set_process_sensitive (window, TRUE);
		marlin_window_set_edit_sensitive (window, TRUE);
		marlin_window_set_selection_sensitive (window, TRUE);
	}

	if (strcmp (name, "channels") == 0) {
		marlin_window_set_process_name (window);
	}
}

static void
selection_changed (MarlinSampleSelection *selection,
		   MarlinWindow *window)
{
	MarlinSampleSubSelection *sub;

	if (window->priv->selection->selections == NULL) {
		gtk_statusbar_pop (GTK_STATUSBAR (window->priv->sel_start_status), 0);
		gtk_statusbar_pop (GTK_STATUSBAR (window->priv->sel_end_status), 0);
	} else {
		char *s_display, *e_display;
		guint rate;
		guint64 ms;
		
		sub = window->priv->selection->selections->data;

		switch (window->priv->position_display) {
		case AS_FRAMES:
			s_display = g_strdup_printf ("%llu", sub->start);
			e_display = g_strdup_printf ("%llu", sub->end);
			break;
			
		case AS_TIME_LONG:
			g_object_get (G_OBJECT (window->priv->sample),
				      "sample_rate", &rate,
				      NULL);
			
			ms = marlin_frames_to_ms (sub->start, rate);
			s_display = marlin_ms_to_time_string (ms);
			
			ms = marlin_frames_to_ms (sub->end, rate);
			e_display = marlin_ms_to_time_string (ms);
			break;
			
		case AS_SECONDS:
			g_object_get (G_OBJECT (window->priv->sample),
				      "sample_rate", &rate,
				      NULL);
			
			ms = marlin_frames_to_ms (sub->start, rate);
			s_display = g_strdup_printf ("%llu", ms / 1000);
			ms = marlin_frames_to_ms (sub->end, rate);
			e_display = g_strdup_printf ("%llu", ms / 1000);
			break;
			
		default:
			break;
		}
		
		gtk_statusbar_pop (GTK_STATUSBAR (window->priv->sel_start_status), 0);
		gtk_statusbar_pop (GTK_STATUSBAR (window->priv->sel_end_status), 0);
		gtk_statusbar_push (GTK_STATUSBAR (window->priv->sel_start_status), 0, s_display);
		gtk_statusbar_push (GTK_STATUSBAR (window->priv->sel_end_status), 0, e_display);
		
		g_free (s_display);
		g_free (e_display);
	}
	
	/* Edit works out what needs to be sensitive... */
	marlin_window_set_edit_sensitive (window, TRUE);
	marlin_window_set_selection_sensitive (window, TRUE);
	marlin_window_set_process_sensitive (window, TRUE);
	marlin_window_set_view_sensitive (window, TRUE);
	marlin_window_set_process_name (window);
	marlin_window_set_edit_name (window);
}

static void
marker_added (MarlinMarkerModel *model,
	      MarlinMarker *marker,
	      MarlinWindow *window)
{
	marlin_window_set_file_sensitive (window, TRUE);
	marlin_window_set_process_sensitive (window, TRUE);
}

static void
marker_removed (MarlinMarkerModel *model,
		MarlinMarker *marker,
		MarlinWindow *window)
{
	marlin_window_set_file_sensitive (window, TRUE);
	marlin_window_set_process_sensitive (window, TRUE);
}

static void
undo_changed (MarlinUndoManager *manager,
	      MarlinWindow *window)
{
	marlin_window_set_edit_sensitive (window, TRUE);
	marlin_window_set_edit_name (window);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinWindow *window;
	MarlinWindowPrivate *priv;
	guint64 total_frames;

	window = MARLIN_WINDOW (object);
	priv = window->priv;

	switch (prop_id) {
	case PROP_SAMPLE:
		if (priv->sample != NULL) {
			g_signal_handler_disconnect (G_OBJECT (priv->sample),
						     priv->notify_id);
			g_object_unref (G_OBJECT (priv->sample));
		}

		if (priv->selection != NULL) {
			g_signal_handler_disconnect (G_OBJECT (priv->selection),
						     priv->selection_id);
			g_object_unref (G_OBJECT (priv->selection));
		}

		if (priv->marker_model != NULL) {
			g_signal_handler_disconnect (G_OBJECT (priv->marker_model),
						     priv->marker_added_id);
			g_signal_handler_disconnect (G_OBJECT (priv->marker_model), 
						     priv->marker_removed_id);
			g_object_unref (G_OBJECT (priv->marker_model));
		}

		if (priv->undo != NULL) {
			g_signal_handler_disconnect (G_OBJECT (priv->undo),
						     priv->undo_changed_id);
			g_object_unref (G_OBJECT (priv->undo));
		}

		priv->sample = g_value_get_object (value);
		g_object_ref (G_OBJECT (priv->sample));
		
		priv->notify_id = g_signal_connect (G_OBJECT (priv->sample), "safe-notify",
						    G_CALLBACK (marlin_window_sample_notify),
						    window);
		
		g_object_get (G_OBJECT (priv->sample),
			      "total_frames", &total_frames,
			      "selection", &priv->selection,
			      "markers", &priv->marker_model,
			      "undo-manager", &priv->undo,
			      NULL);
#if 0
		g_object_ref (G_OBJECT (priv->selection));
		g_object_ref (G_OBJECT (priv->marker_model));
#endif

		priv->marker_added_id = g_signal_connect (G_OBJECT (priv->marker_model),
							  "marker-added",
							  G_CALLBACK (marker_added), window);
		priv->marker_removed_id = g_signal_connect (G_OBJECT (priv->marker_model),
							    "marker-removed",
							    G_CALLBACK (marker_removed), window);

		priv->undo_changed_id = g_signal_connect (G_OBJECT (priv->undo),
							  "changed",
							  G_CALLBACK (undo_changed), window);
		g_object_set (G_OBJECT (priv->play_pipeline),
			      "sample", priv->sample,
			      NULL);
		g_object_set (G_OBJECT (priv->view),
			      "sample", priv->sample,
			      NULL);
		g_object_set (G_OBJECT (priv->overview_bar),
			      "sample", priv->sample,
			      NULL);

		g_object_set (G_OBJECT (priv->marker_view),
			      "model", priv->marker_model,
			      "sample", priv->sample,
			      NULL);
		priv->selection_id = g_signal_connect (G_OBJECT (priv->selection),
						       "changed",
						       G_CALLBACK (selection_changed),
						       window);
		
		marlin_window_set_title (window, priv->sample);
		marlin_window_set_file_sensitive (window, TRUE);
		marlin_window_set_media_bar_sensitive (window, TRUE);
		marlin_window_set_view_sensitive (window, TRUE);
		marlin_window_set_edit_sensitive (window, TRUE);
		marlin_window_set_selection_sensitive (window, TRUE);
		marlin_window_set_process_sensitive (window, TRUE);
		
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinWindow *window;

	window = MARLIN_WINDOW (object);

	switch (prop_id) {
	case PROP_SAMPLE:
		g_value_set_object (value, window->priv->sample);
		break;

	default:
		break;
	}
}

static void
class_init (MarlinWindowClass *klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS (klass);
	
	object_class->finalize = finalize;
	object_class->dispose = dispose;
	object_class->set_property = set_property;
	object_class->get_property = get_property;

	g_object_class_install_property (object_class,
					 PROP_SAMPLE,
					 g_param_spec_object ("sample", "", "",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));
							      
	
	parent_class = g_type_class_peek_parent (klass);
}

static gboolean
file_open_recent (EggRecentView *view,
		  const EggRecentItem *item,
		  MarlinWindow *window)
{
	char *uri;
	char *temp;

	uri = egg_recent_item_get_uri (item);

	temp = gnome_vfs_unescape_string (uri, "");

	if (temp != NULL) {
		char *local;

		local = gnome_vfs_get_local_path_from_uri (temp);

		marlin_open_window (local, NULL);

		g_free (local);
		g_free (temp);
	}

	g_free (uri);
	return TRUE;
}

static void
add_widget (GtkUIManager *merge,
	    GtkWidget *widget,
	    MarlinWindow *window)
{
	if (GTK_IS_MENU_SHELL (widget)) {
		window->priv->menubar = widget;
	} else {
		window->priv->toolbars = g_list_append (window->priv->toolbars, widget);
		/* FIXME: Somehow */
		/* Hack to make the toolbars not have text */
/* 		gtk_toolbar_set_style (GTK_TOOLBAR (widget), GTK_TOOLBAR_ICONS); */
	}

	gtk_box_pack_start (GTK_BOX (window->priv->menu_dock), widget,
			    FALSE, FALSE, 0);
	gtk_widget_show (widget); 
	
}

static void
menu_item_select_cb (GtkMenuItem *proxy,
		     MarlinWindow *window)
{
	GtkAction *action;
	char *message;

	action = g_object_get_data (G_OBJECT (proxy), "gtk-action");
	g_return_if_fail (action != NULL);

	g_object_get (G_OBJECT (action), 
		      "tooltip", &message, 
		      NULL);
	if (message && window->priv->message_status) {
		gtk_statusbar_push (GTK_STATUSBAR (window->priv->message_status), 
				    window->priv->help_message_cid, message);
		g_free (message);
	}
}

static void
menu_item_deselect_cb (GtkMenuItem *proxy,
		       MarlinWindow *window)
{
	if (window->priv->message_status) {
		gtk_statusbar_pop (GTK_STATUSBAR (window->priv->message_status),
				   window->priv->help_message_cid);
	}
}

static void
connect_proxy_cb (GtkUIManager *manager,
		  GtkAction *action,
		  GtkWidget *proxy,
		  MarlinWindow *window)
{
	if (GTK_IS_MENU_ITEM (proxy)) {
		g_signal_connect (proxy, "select",
				  G_CALLBACK (menu_item_select_cb), window);
		g_signal_connect (proxy, "deselect",
				  G_CALLBACK (menu_item_deselect_cb), window);
	}
}

static void
disconnect_proxy_cb (GtkUIManager *manager,
		     GtkAction *action,
		     GtkWidget *proxy,
		     MarlinWindow *window)
{
	if (GTK_IS_MENU_ITEM (proxy)) {
		g_signal_handlers_disconnect_by_func (proxy, G_CALLBACK (menu_item_select_cb), window);
		g_signal_handlers_disconnect_by_func (proxy, G_CALLBACK (menu_item_deselect_cb), window);
	}
}

static void
init (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	GtkUIManager *merge;
	EggRecentModel *model;
	GConfClient *client;
	char *xml;
	
	window->priv = g_new0 (MarlinWindowPrivate, 1);
	priv = window->priv;

	priv->toolbars = NULL;
	
	priv->main_vbox = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (priv->main_vbox);
	gtk_container_add (GTK_CONTAINER (window), priv->main_vbox);

	priv->menu_dock = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (priv->menu_dock);
	gtk_box_pack_start (GTK_BOX (priv->main_vbox), priv->menu_dock,
			    FALSE, TRUE, 0);
	
	client = gconf_client_get_default ();
	priv->position_display = gconf_client_get_int (client, 
						       "/apps/marlin/system-state/position-display",
						       NULL);
	g_object_unref (G_OBJECT (client));

	priv->close_on_op_complete = FALSE;
	priv->state = MARLIN_STATE_STOPPED;

	merge = gtk_ui_manager_new ();

	/* Connect to the merge for UI updates */
	g_signal_connect (merge, "connect_proxy",
			  G_CALLBACK (connect_proxy_cb), window);
	g_signal_connect (merge, "disconnect_proxy",
			  G_CALLBACK (disconnect_proxy_cb), window);

	/* Action stuff */
	priv->ui_action = gtk_action_group_new ("WindowActions");
	marlin_window_menu_setup (window);
	gtk_ui_manager_insert_action_group (merge, priv->ui_action, 0);

	priv->popup_action = gtk_action_group_new ("PopupsActions");
	marlin_window_menu_popup_setup (window);
	gtk_ui_manager_insert_action_group (merge, priv->popup_action, 0);
	
	priv->ui_merge = G_OBJECT (merge);
	g_signal_connect (G_OBJECT (merge), "add_widget",
			  G_CALLBACK (add_widget), window);
	xml = marlin_file ("marlin.xml");
	if (xml == NULL) {
		g_error ("Marlin not installed correctly");
	}
	
	gtk_ui_manager_add_ui_from_file (merge, xml, NULL);
	g_free (xml);

	priv->accel_group = gtk_ui_manager_get_accel_group (merge);
	gtk_window_add_accel_group (GTK_WINDOW (window), priv->accel_group);

	/* Window holds ref */
	g_object_unref (priv->accel_group);

 	priv->toolbar = toolbar_new (window);
	gtk_widget_show (GTK_WIDGET (priv->toolbar));
	gtk_box_pack_end (GTK_BOX (priv->menu_dock), GTK_WIDGET (priv->toolbar),
			  FALSE, FALSE, 0);
	
	priv->recent_view = marlin_recent_view_new (merge);
	g_signal_connect (G_OBJECT (priv->recent_view), "activate",
			  G_CALLBACK (file_open_recent), window);

	priv->last_save = time (NULL);
}

GType
marlin_window_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinWindowClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinWindow), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (GTK_TYPE_WINDOW,
					       "MarlinWindow",
					       &info, 0);
	}

	return type;
}

static void
statusbar_set_frames (GtkWidget *statusbar,
		      MarlinWindow *window,
		      guint64 frames)
{
	guint rate;
	guint64 ms;
	char *display = NULL;

	switch (window->priv->position_display) {
	case AS_FRAMES:
		display = g_strdup_printf ("%llu", frames);
		break;

	case AS_TIME_LONG:
		g_object_get (G_OBJECT (window->priv->sample),
			      "sample_rate", &rate,
			      NULL);
		
		ms = marlin_frames_to_ms (frames, rate);
		display = marlin_ms_to_time_string (ms);
		break;

	case AS_SECONDS:
		g_object_get (G_OBJECT (window->priv->sample),
			      "sample_rate", &rate,
			      NULL);

		ms = marlin_frames_to_ms (frames, rate);
		display = g_strdup_printf ("%llu", ms / 1000);
		break;

	default:
		return;
	}

	gtk_statusbar_pop (GTK_STATUSBAR (statusbar), 0);
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), 0, display);

	g_free (display);
}

static void
marlin_window_change_overview_bar_fpp (MarlinSampleView *view,
				       guint frames_per_pixel,
				       MarlinWindow *window)
{
	g_object_set (G_OBJECT (window->priv->overview_bar),
		      "frames_per_page", (guint64) frames_per_pixel * GTK_WIDGET (view)->allocation.width,
		      NULL);

	g_object_set (G_OBJECT (window->priv->marker_view),
		      "frames_per_pixel", frames_per_pixel,
		      NULL);
	
	marlin_window_set_view_sensitive (window, TRUE);
}

static void
view_vzoom_changed (MarlinSampleView *view,
		    float vmax,
		    float vmin,
		    MarlinWindow *window)
{
	marlin_window_set_view_sensitive (window, TRUE);
}

static void
marlin_window_page_start_changed (MarlinSampleView *view,
				  guint64 start_frame,
				  MarlinWindow *window)
{
	guint frames_per_pixel;

	g_object_get (G_OBJECT (view),
		      "frames_per_pixel", &frames_per_pixel,
		      NULL);

	g_object_set (G_OBJECT (window->priv->overview_bar),
		      "page-start", start_frame,
		      NULL);
}

static void
marlin_window_view_size_allocate (MarlinSampleView *view,
				  GtkAllocation *allocation,
				  MarlinWindow *window)
{
	guint frames_per_pixel;
	
	g_object_get (G_OBJECT (view),
		      "frames_per_pixel", &frames_per_pixel,		      
		      NULL);
	
	g_object_set (G_OBJECT (window->priv->overview_bar),
		      "frames_per_page", (guint64) frames_per_pixel * allocation->width,
		      NULL);
}

static void
marlin_window_overview_bar_start_changed (MarlinOverviewBar *overview_bar,
				      guint64 start,
				      MarlinWindow *window)
{
	marlin_sample_view_scroll_to (window->priv->view, start);
}

/* Grr, annoyingly ugly prototype */
static void
marlin_window_view_move_cursor (MarlinSampleView *view,
				GtkMovementStep step,
				int count,
				gboolean extend_selection,
				MarlinWindow *window);
static void
marlin_window_overview_bar_cursor_changed (MarlinOverviewBar *overview_bar,
				       guint64 position,
				       MarlinWindow *window)
{
	/* We need to turn off this signal, otherwise we get into
	   a loop of the overview_bar emitting the signal and setting the
	   view's cursor, which then gets back in here, and goes on */
	g_signal_handlers_block_matched (G_OBJECT (window->priv->view),
					 G_SIGNAL_MATCH_FUNC,
					 0, 0, NULL,
					 G_CALLBACK (marlin_window_view_move_cursor), NULL);

	g_object_set (G_OBJECT (window->priv->view),
		      "cursor_position", position,
		      NULL);

	g_signal_handlers_unblock_matched (G_OBJECT (window->priv->view),
					   G_SIGNAL_MATCH_FUNC,
					   0, 0, NULL,
					   G_CALLBACK (marlin_window_view_move_cursor), NULL);
}

static void
marlin_window_view_move_cursor (MarlinSampleView *view,
				GtkMovementStep step,
				int count,
				gboolean extend_selection,
				MarlinWindow *window)
{
	guint64 view_pos;

	g_object_get (G_OBJECT (view),
		      "cursor_position", &view_pos,
		      NULL);

	/* Change the position in the status bar */
	statusbar_set_frames (window->priv->position_status,
			      window, (guint64) view_pos);
	
	/* We need to turn off this signal, otherwise we get into
	   a loop of the overview_bar emitting the signal and setting the
	   view's cursor, which then gets back in here, and goes on */
	g_signal_handlers_block_matched (G_OBJECT (window->priv->overview_bar),
					 G_SIGNAL_MATCH_FUNC,
					 0, 0, NULL,
					 G_CALLBACK (marlin_window_overview_bar_cursor_changed), NULL);
	
	g_object_set (G_OBJECT (window->priv->overview_bar),
		      "cursor_position", (guint64) view_pos,
		      NULL);
	g_signal_handlers_unblock_matched (G_OBJECT (window->priv->overview_bar),
					   G_SIGNAL_MATCH_FUNC,
					   0, 0, NULL,
					   G_CALLBACK (marlin_window_overview_bar_cursor_changed), NULL);
}

static void
play_eos (MarlinPlayPipeline *pipeline,
	  MarlinWindow *window)
{
	marlin_pipeline_set_state (MARLIN_PIPELINE (pipeline), GST_STATE_READY);
	window->priv->state = MARLIN_STATE_STOPPED;
	
	marlin_window_move_cursor_to (window, window->priv->initial);

	if (window->priv->repeating) {
		marlin_window_play_sample (window);
	}
}

static void
position_changed (MarlinPlayPipeline *pipeline,
		  guint64 position,
		  MarlinWindow *window)
{
	marlin_window_move_cursor_to (window, position);
}

static void 
change_position_display (GtkCheckMenuItem *item,
			 MarlinWindow *window)
{
	GConfClient *client;

	if (gtk_check_menu_item_get_active (item)) {
		guint64 position;

		window->priv->position_display = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (item), "position-display"));

		client = gconf_client_get_default ();
		gconf_client_set_int (client, "/apps/marlin/system-state/position-display", window->priv->position_display, NULL);
		g_object_unref (G_OBJECT (client));

		/* Update the display */
		g_object_get (G_OBJECT (window->priv->view),
			      "cursor_position", &position,
			      NULL);

		statusbar_set_frames (window->priv->position_status, window, position);
	}
}

static char *names[LAST_PD] = {
	N_("_Frames"),
	N_("_Time"),
	N_("_Seconds"),
	N_("T_ime & Frames")
};

static void
attach_popup_menu (MarlinWindow *window,
		   GtkWidget *widget)
{
	GtkWidget *menu;
	MarlinWindowPositionDisplay pd;
	GSList *group = NULL;

	menu = gtk_menu_new ();

	for (pd = AS_FRAMES; pd < LAST_PD; pd++) {
		GtkWidget *item;
		
		item = gtk_radio_menu_item_new_with_mnemonic (group, _(names[pd]));
		if (pd == window->priv->position_display) {
			gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (item), TRUE);
		}
		g_object_set_data (G_OBJECT (item), "position-display", 
				   GINT_TO_POINTER (pd));

		group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (item));

		gtk_widget_show (item);
		gtk_menu_shell_append ((GtkMenuShell *) menu, item);
		g_signal_connect (G_OBJECT (item), "toggled",
				  G_CALLBACK (change_position_display), window);
	}

	gnome_popup_menu_attach (menu, widget, NULL);
}

static gboolean
view_key_press (MarlinSampleView *view,
		GdkEventKey *event,
		MarlinWindow *window)
{
	MarlinWindowPrivate *priv = window->priv;
	MarlinSampleSubSelection *sub;
	MarlinUndoContext *ctxt;
	guint64 position;
	
	switch (event->keyval) {
	case GDK_space:
		switch (priv->state) {
		case MARLIN_STATE_PAUSED:
		case MARLIN_STATE_STOPPED:
			marlin_window_play_sample (window);
			break;

		case MARLIN_STATE_PLAYING:
			marlin_window_pause_sample (window);
			break;

		case MARLIN_STATE_RECORDING:
			break;

		default:
			break;
		}
		
		break;

	case GDK_m:
		/* Add a marker at the cursor position */
		/* FIXME: Check this is a valid operation */
		g_object_get (G_OBJECT (view),
			      "cursor_position", &position,
			      NULL);

		ctxt = marlin_undo_manager_context_begin (priv->undo,
							  _("Add Marker"));
		marlin_marker_model_add_marker (priv->marker_model,
						position, NULL, ctxt);
		marlin_undo_manager_context_end (priv->undo, ctxt);

		break;
		
	default:
		g_print ("Other key pressed\n");
		return FALSE;
	}

	return TRUE;
}

static void
drag_data_received (GObject *object,
		    GdkDragContext *context,
		    int x,
		    int y,
		    GtkSelectionData *selection,
		    guint info,
		    guint time,
		    gpointer data)
{
	MarlinWindow *window;
	char *tmp, *filename, **filenames;
	int i;

	window = MARLIN_WINDOW (object);

	switch (info) {
	case DND_TYPE_TEXT_URI_LIST:
		tmp = g_strndup (selection->data, selection->length);
		filenames = g_strsplit (tmp, "\n", 0);
		g_free (tmp);

		for (i = 0; filenames[i] != NULL; i++) {
			filename = g_strstrip (filenames[i]);

			if (filename[0] == 0) {
				continue;
			}
			
			marlin_open_window (filename + 7, NULL);
			g_free (filename);
		}

		g_free (filenames);
		break;

	default:
		break;
	}
}

static void
clip_changed (MarlinProgram *program,
	      MarlinWindow *window)
{
	if (marlin_program_get_clipboard (program) == NULL) {
		marlin_window_set_edit_sensitive (window, FALSE);
		marlin_window_set_view_sensitive (window, FALSE);
	} else {
		marlin_window_set_edit_sensitive (window, TRUE);
		marlin_window_set_view_sensitive (window, TRUE);
	}
}

static void
attach_popups (MarlinWindow *window)
{
	GtkWidget *menu;

	menu = gtk_ui_manager_get_widget (GTK_UI_MANAGER (window->priv->ui_merge),
					  "/ViewPopup");
	if (menu == NULL) {
		g_warning ("/ViewPopup is not valid. Have you installed properly?");
		return;
	}
	
	gnome_popup_menu_attach (menu, GTK_WIDGET (window->priv->view), NULL);
}

static void
attach_marker_popup (MarlinWindow *window)
{
	GtkWidget *menu;

}

static void
marker_view_move_cursor (MarlinMarkerView *view,
			 guint64 position,
			 MarlinWindow *window)
{
	marlin_window_move_cursor_to (window, position);
}

static void
overview_bar_play_request (MarlinOverviewBar *bar,
			   guint64 start,
			   MarlinWindow *window)
{
	guint64 end;

	g_object_get (G_OBJECT (window->priv->sample),
		      "total_frames", &end,
		      NULL);

	marlin_window_play_range (window, start, end);
}

void
marlin_window_set_icon (MarlinWindow *window,
			GdkPixbuf *icon)
{
	if (window_icon == NULL) {
		char *filename;
		filename = marlin_file ("marlin/marlin-icon.png");
		
		window_icon = gdk_pixbuf_new_from_file (filename, NULL);
		g_free (filename);
	}

	if (icon == NULL) {
		icon = window_icon;
	}

	gtk_window_set_icon (GTK_WINDOW (window), icon);
}
	
MarlinWindow *
marlin_window_new (MarlinSample *sample)
{
	MarlinProgram *program = marlin_program_get_default ();
	MarlinWindow *window;
	MarlinWindowPrivate *priv;
	GtkWidget *sw, *table, *ev, *statusbar;
	guint64 frames;
	EggRecentModel *model;
	
	window = g_object_new (MARLIN_WINDOW_TYPE, NULL);
	priv = window->priv;

	gtk_window_set_default_size (GTK_WINDOW (window), DEFAULT_WIDTH, DEFAULT_HEIGHT);

	/* DND support */
	gtk_drag_dest_set (GTK_WIDGET (window), GTK_DEST_DEFAULT_ALL,
			   drop_types, num_drop_types, GDK_ACTION_COPY);
	g_signal_connect (G_OBJECT (window), "drag_data_received",
			  G_CALLBACK (drag_data_received), NULL);

	marlin_window_set_icon (window, NULL);

	/* Create the internal members */
	if (sample != NULL) {
		g_object_get (G_OBJECT (sample),
			      "total_frames", &frames,
			      "selection", &priv->selection,
			      "undo-manager", &priv->undo,
			      "markers", &priv->marker_model,
			      NULL);

#if 0
		g_object_ref (G_OBJECT (priv->selection));
		g_object_ref (G_OBJECT (priv->undo));
#endif

		priv->sample = sample;
		g_object_ref (priv->sample);
		marlin_window_set_title (window, sample);
		
		priv->notify_id = g_signal_connect (G_OBJECT (sample), "safe-notify",
						    G_CALLBACK (marlin_window_sample_notify),
						    window);

/* 		g_object_ref (G_OBJECT (priv->marker_model)); */
		
		priv->marker_added_id = g_signal_connect (G_OBJECT (priv->marker_model),
							  "marker-added",
							  G_CALLBACK (marker_added), window);
		priv->marker_removed_id = g_signal_connect (G_OBJECT (priv->marker_model),
							    "marker-removed",
							    G_CALLBACK (marker_removed), window);
		priv->undo_changed_id = g_signal_connect (G_OBJECT (priv->undo),
							  "changed",
							  G_CALLBACK (undo_changed), window);
	} else {
		frames = (guint64) 0;
		priv->sample = NULL;
		priv->selection = NULL;
		priv->marker_model = NULL;
		priv->undo = NULL;
	}

	if (priv->sample == NULL) {
		priv->play_pipeline = g_object_new (MARLIN_PLAY_PIPELINE_TYPE, NULL);
	} else {
		priv->play_pipeline = g_object_new (MARLIN_PLAY_PIPELINE_TYPE,
						    "sample", priv->sample,
						    NULL);
	}
	g_signal_connect (G_OBJECT (priv->play_pipeline), "eos",
			  G_CALLBACK (play_eos), window);
	g_signal_connect (G_OBJECT (priv->play_pipeline), "position-changed",
			  G_CALLBACK (position_changed), window);

	priv->overview_bar = g_object_new (MARLIN_OVERVIEW_BAR_TYPE,
					   "sample", priv->sample,
					   "frames_per_page", (guint64) (4096 * DEFAULT_WIDTH),
					   NULL);
	gtk_widget_show (GTK_WIDGET (priv->overview_bar));
	if (priv->selection != NULL) {
		priv->selection_id = g_signal_connect (G_OBJECT (priv->selection), "changed",
						       G_CALLBACK (selection_changed), window);
	}
	g_signal_connect (G_OBJECT (priv->overview_bar), "page-start-changed",
			  G_CALLBACK (marlin_window_overview_bar_start_changed),
			  window);
	g_signal_connect (G_OBJECT (priv->overview_bar), "cursor-changed",
			  G_CALLBACK (marlin_window_overview_bar_cursor_changed),
			  window);
	g_signal_connect (G_OBJECT (priv->overview_bar), "play-request",
			  G_CALLBACK (overview_bar_play_request), window);
	
	gtk_box_pack_start (GTK_BOX (priv->main_vbox),
			    GTK_WIDGET (priv->overview_bar),
			    FALSE, FALSE, 0);

	/* 3 rows, 2 columns:
	                  | Horizontal ruler
	   vertical ruler | sample view
	                  | Scrollbar.
	*/
	   
	table = gtk_table_new (3, 2, FALSE);
	gtk_box_pack_start (GTK_BOX (priv->main_vbox), table, TRUE, TRUE, 0);

	/* <dobey>GtkRuler kinda sucks..</dobey>.
	   TODO: Make MarlinRuler? */

	priv->marker_view = g_object_new (MARLIN_MARKER_VIEW_TYPE,
					  "model", priv->marker_model,
					  "frames_per_pixel", 4096,
					  "sample", priv->sample,
					  NULL);
	g_signal_connect (G_OBJECT (priv->marker_view), "move-cursor",
			  G_CALLBACK (marker_view_move_cursor), window);
	
	gtk_table_attach (GTK_TABLE (table), priv->marker_view,
			  1, 2, 0, 1,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL,
			  0, 0);

	priv->vruler = gtk_vruler_new ();
	gtk_ruler_draw_pos (GTK_RULER (priv->vruler));
	gtk_table_attach (GTK_TABLE (table), priv->vruler,
			  0, 1, 1, 2,
			  GTK_FILL,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	/* Create the scrollbar here so we can use it's adjustment for
	   the scrolled window */
	priv->hscrollbar = gtk_hscrollbar_new (NULL);
	
	sw = gtk_scrolled_window_new (GTK_RANGE (priv->hscrollbar)->adjustment, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
					GTK_POLICY_NEVER,
					GTK_POLICY_NEVER);

	/* Set the adjustments on the marker view */
	gtk_widget_set_scroll_adjustments (GTK_WIDGET (priv->marker_view),
					   GTK_RANGE (priv->hscrollbar)->adjustment,
					   GTK_RANGE (GTK_SCROLLED_WINDOW (sw)->vscrollbar)->adjustment);

	if (priv->sample != NULL) {
		priv->view = g_object_new (MARLIN_SAMPLE_VIEW_TYPE,
					   "frames_per_pixel", 4096,
					   "sample", priv->sample,
					   NULL);
	} else {
		priv->view = g_object_new (MARLIN_SAMPLE_VIEW_TYPE,
					   "frames_per_pixel", 4096,
					   NULL);
	}
	
	g_signal_connect (G_OBJECT (priv->view), "key-press-event",
			  G_CALLBACK (view_key_press), window);
	gtk_container_add (GTK_CONTAINER (sw), GTK_WIDGET (window->priv->view));

	gtk_table_attach (GTK_TABLE (table), sw,
			  1, 2, 1, 2,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);
	
	g_signal_connect (G_OBJECT (priv->view), "frames-per-pixel-changed",
			  G_CALLBACK (marlin_window_change_overview_bar_fpp),
			  window);
	g_signal_connect (G_OBJECT (priv->view), "page-start-changed",
			  G_CALLBACK (marlin_window_page_start_changed),
			  window);
	g_signal_connect (G_OBJECT (priv->view), "size-allocate",
			  G_CALLBACK (marlin_window_view_size_allocate),
			  window);
	g_signal_connect (G_OBJECT (priv->view), "move-cursor",
			  G_CALLBACK (marlin_window_view_move_cursor),
			  window);
	g_signal_connect (G_OBJECT (priv->view), "vzoom-changed",
			  G_CALLBACK (view_vzoom_changed), window);
	
	gtk_table_attach (GTK_TABLE (table), priv->hscrollbar,
			  1, 2, 2, 3,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL,
			  0, 0);

	attach_popups (window);
	
	gtk_widget_show_all (table);

	statusbar = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (priv->main_vbox), statusbar, FALSE, FALSE, 0);
	gtk_widget_show (statusbar);
	
	/* The status bars */
	/* Message bar */
	priv->message_status = gtk_statusbar_new ();
	priv->help_message_cid = gtk_statusbar_get_context_id (GTK_STATUSBAR (priv->message_status), "help_message");

	gtk_widget_show (priv->message_status);
	gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (priv->message_status), FALSE);

	gtk_box_pack_start (GTK_BOX (statusbar), priv->message_status, TRUE, TRUE, 0);

	/* Cursor position */
	priv->position_status = gtk_statusbar_new ();
	gtk_widget_set_size_request (priv->position_status, 100, -1);

	statusbar_set_frames (priv->position_status, window, (guint64) 0);

	ev = gtk_event_box_new ();
	gtk_container_add (GTK_CONTAINER (ev), priv->position_status);
 	attach_popup_menu (window, ev);
	gtk_widget_show_all (ev);

	gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (priv->position_status), FALSE);
	
	gtk_box_pack_start (GTK_BOX (statusbar), ev, FALSE, FALSE, 0);

	/* Selection start */
	priv->sel_start_status = gtk_statusbar_new ();
	gtk_widget_set_size_request (priv->sel_start_status, 100, -1);

	ev = gtk_event_box_new ();
	gtk_container_add (GTK_CONTAINER (ev), priv->sel_start_status);
 	attach_popup_menu (window, ev);
	gtk_widget_show_all (ev);

	gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (priv->sel_start_status), FALSE);

	gtk_box_pack_start (GTK_BOX (statusbar), ev, FALSE, FALSE, 0);
	
	priv->sel_end_status = gtk_statusbar_new ();
	gtk_widget_set_size_request (priv->sel_end_status, 100, -1);

	ev = gtk_event_box_new ();
	gtk_container_add (GTK_CONTAINER (ev), priv->sel_end_status);
	attach_popup_menu (window, ev);
	gtk_widget_show_all (ev);

	gtk_box_pack_start (GTK_BOX (statusbar), ev, FALSE, FALSE, 0);

	model = marlin_recent_get_model ();
	egg_recent_view_set_model (EGG_RECENT_VIEW (priv->recent_view), model);	
	marlin_window_set_file_sensitive (window, frames != 0);
	marlin_window_set_media_bar_sensitive (window, frames != 0);
	marlin_window_set_view_sensitive (window, frames != 0);
	marlin_window_set_edit_sensitive (window, frames != 0);
	marlin_window_set_selection_sensitive (window, frames != 0);
	marlin_window_set_process_sensitive (window, frames != 0);
	marlin_window_set_process_name (window);

	set_path_sensitive (priv->ui_action, "MediaRecord", TRUE);
	/* Connect to the clipboard signals */
	priv->clip_id = g_signal_connect (G_OBJECT (program), "clipboard-changed",
					  G_CALLBACK (clip_changed), window);

	gtk_widget_show (priv->main_vbox);
	return window;
}

static int
get_int (GConfClient *client,
	 const char *root,
	 const char *key)
{
	int ret;
	char *fullkey;

	fullkey = g_strdup_printf ("/apps/marlin/system-state/%s/%s", root, key);
	ret = gconf_client_get_int (client, fullkey, NULL);
	g_free (fullkey);

	return ret;
}

MarlinWindow *
marlin_window_new_from_state (MarlinSample *sample,
			      const char *uuid)
{
	MarlinWindow *window;
	GConfClient *client;
	int x, y, width, height, cursor_position;
	
	window = marlin_window_new (sample);

	client = gconf_client_get_default ();
	
	/* Restore state */
	x = get_int (client, uuid, "x");
	y = get_int (client, uuid, "y");
	width = get_int (client, uuid, "width");
	height = get_int (client, uuid, "height");

	gtk_window_move (GTK_WINDOW (window), x, y);
	gtk_window_resize (GTK_WINDOW (window), width, height);

	cursor_position = get_int (client, uuid, "cursor-position");
	g_object_set (G_OBJECT (window->priv->view),
		      "cursor-position", cursor_position,
		      NULL);

	g_object_unref (G_OBJECT (client));
	
	return window;
}

void
marlin_window_set_title (MarlinWindow *window,
			 MarlinSample *sample)
{
	char *title, *filename;
	gboolean dirty, writable;

	g_object_get (G_OBJECT (sample),
		      "name", &filename,
		      "dirty", &dirty,
		      "writable", &writable,
		      NULL);
	
	if (filename == NULL) {
		title = g_strdup (_("Marlin Sample Editor"));
	} else {
		title = g_strdup_printf ("%s%s%s- Marlin", filename,
					 writable ? "" : _(" (read only)"),
					 dirty ? _(" (modified)") : "");
	}

	g_free (filename);

	gtk_window_set_title (GTK_WINDOW (window), title);
	g_free (title);
}

static void
maybe_save_response (GtkDialog *dialog,
		     guint response_id,
		     MarlinWindow *window)
{
	switch (response_id) {
	case GTK_RESPONSE_NO:
		gtk_widget_destroy (GTK_WIDGET (dialog));
		gtk_widget_destroy (GTK_WIDGET (window));
		return;

	case GTK_RESPONSE_CANCEL:
		break;

	case GTK_RESPONSE_OK:
		/* save */
		window->priv->close_on_op_complete = TRUE;
		marlin_window_save_sample_as (window);
		break;

	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
}

static char *
time_to_str (time_t length)
{
	int mins;
	
	if (length < 60) {
		return g_strdup_printf (ngettext ("If you close without saving, changes from the last second will be discarded.",
						  "If you close without saving, changes from the last %d seconds will be discarded.", length), length);
	} else {
		mins = length / 60;

		return g_strdup_printf (ngettext ("If you close without saving, changes from the last minute will be discarded.",
						  "If you close without saving, changes from the last %d minutes will be discarded", mins), mins);
	}
}

static void
maybe_save_quit (MarlinWindow *window)
{
	GtkWidget *mess;
	char *title, *filename;
	char *basename, *time_str;
	char *bold;
	time_t now;
	
	g_object_get (G_OBJECT (window->priv->sample),
		      "filename", &filename,
		      NULL);

	now = time (NULL);
	time_str = time_to_str (now - window->priv->last_save);
	
	basename = g_path_get_basename (filename);

	bold = g_strdup_printf (_("Save changes made to \"%s\" before closing?"),
				basename);
	mess = gtk_message_dialog_new (GTK_WINDOW (window), 0,
				       GTK_MESSAGE_WARNING,
				       GTK_BUTTONS_NONE,
				       "<span weight=\"bold\" size=\"larger\">%s</span>\n%s",
				       bold, time_str);
	g_free (bold);
	g_signal_connect (G_OBJECT (mess), "response",
			  G_CALLBACK (maybe_save_response), window);

	marlin_add_button_to_dialog (GTK_DIALOG (mess), _("Do_n't save"), GTK_STOCK_NO, GTK_RESPONSE_NO);
	marlin_add_button_to_dialog (GTK_DIALOG (mess), _("Cancel"), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (mess), GTK_STOCK_SAVE, GTK_RESPONSE_OK);
	gtk_dialog_set_default_response (GTK_DIALOG (mess), GTK_RESPONSE_OK);

	title = g_strdup_printf (_("Save %s?"), filename);
	gtk_window_set_title (GTK_WINDOW (mess), title);
	g_free (title);
	g_free (basename);
	g_free (filename);
	g_free (time_str);

	/* Bad evil hack */
	gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (mess)->label), TRUE);
	gtk_label_set_line_wrap (GTK_LABEL (GTK_MESSAGE_DIALOG (mess)->label), FALSE);
	gtk_label_set_justify (GTK_LABEL (GTK_MESSAGE_DIALOG (mess)->label), GTK_JUSTIFY_LEFT);
	
	gtk_widget_show (mess);
}
		
gboolean
marlin_window_can_close (MarlinWindow *window)
{
	gboolean dirty;

	g_object_get (G_OBJECT (window->priv->sample),
		      "dirty", &dirty,
		      NULL);

	if (dirty) {
		maybe_save_quit (window);
	}
	
	return dirty;
}

void
marlin_window_close (MarlinWindow *window)
{
	g_return_if_fail (IS_MARLIN_WINDOW (window));
	
	if (marlin_window_can_close (window) == FALSE) {
		gtk_widget_destroy (GTK_WIDGET (window));
	} /* Otherwise the save dialog will close the window */
}

GtkActionGroup *
marlin_window_get_action_group (MarlinWindow *window)
{
	g_return_if_fail (IS_MARLIN_WINDOW (window));

	return window->priv->ui_action;
}

GtkActionGroup *
marlin_window_get_popup_action_group (MarlinWindow *window)
{
	g_return_if_fail (IS_MARLIN_WINDOW (window));

	return window->priv->popup_action;
}

void
marlin_window_move_cursor_to (MarlinWindow *window,
			      guint64 absolute_position)
{
	g_object_set (G_OBJECT (window->priv->view),
		      "cursor_position", absolute_position,
		      NULL);
}

void
marlin_window_move_cursor (MarlinWindow *window,
			   guint64 relative_position)
{
	guint64 position;
	guint fpp;

	g_object_get (G_OBJECT (window->priv->view),
		      "cursor_position", &position,
		      "frames_per_pixel", &fpp,
		      NULL);

	position += (relative_position * fpp);
	g_object_set (G_OBJECT (window->priv->view),
		      "cursor_position", (guint64) position,
		      NULL);
}

void
marlin_window_move_cursor_to_start (MarlinWindow *window)
{
	g_object_set (G_OBJECT (window->priv->view),
		      "cursor_position", (guint64) 0,
		      NULL);
}

void
marlin_window_move_cursor_to_end (MarlinWindow *window)
{
	guint fpp;
	guint64 total;

	g_object_get (G_OBJECT (window->priv->view),
		      "frames_per_pixel", &fpp,
		      NULL);
	g_object_get (G_OBJECT (window->priv->sample),
		      "total_frames", &total,
		      NULL);

	g_object_set (G_OBJECT (window->priv->view),
		      "cursor_position", (guint64) (total - fpp),
		      NULL);
}


static void
set_path_sensitive (GtkActionGroup *ag,
		    const char *name,
		    gboolean sensitive)
{
	GtkAction *action;

	action = gtk_action_group_get_action (ag, name);
	g_object_set (action, "sensitive", sensitive, NULL);
}

void
marlin_window_set_file_sensitive (MarlinWindow *window,
				  gboolean sensitive)
{
	gboolean dirty, writable;
	GtkActionGroup *ag;

	g_object_get (G_OBJECT (window->priv->sample),
		      "dirty", &dirty,
		      "writable", &writable,
		      NULL);
	ag = window->priv->ui_action;
	
	set_path_sensitive (ag, "FileSave", dirty && writable);
	set_path_sensitive (ag, "FileSaveAs", sensitive);
 	set_path_sensitive (ag, "FileProperties", sensitive);
}

void
marlin_window_set_media_bar_sensitive (MarlinWindow *window,
				       gboolean sensitive)
{
	GtkActionGroup *ag;

	ag = window->priv->ui_action;

	set_path_sensitive (ag, "MediaPrevious", sensitive);
	set_path_sensitive (ag, "MediaNext", sensitive);
	set_path_sensitive (ag, "MediaForward", sensitive);
	set_path_sensitive (ag, "MediaRewind", sensitive);

	set_path_sensitive (ag, "MediaPlay", sensitive);
	set_path_sensitive (ag, "MediaRepeat", sensitive);
	set_path_sensitive (ag, "MediaPause", sensitive);
	set_path_sensitive (ag, "MediaStop", sensitive);
}

void
marlin_window_set_view_sensitive (MarlinWindow *window,
				  gboolean sensitive)
{
	MarlinProgram *program = marlin_program_get_default ();
	GtkActionGroup *ag;
	guint64 frames;
	
	ag = window->priv->ui_action;

	if (sensitive == FALSE) {
		set_path_sensitive (ag, "ZoomIn", FALSE);
		set_path_sensitive (ag, "ZoomOut", FALSE);
		set_path_sensitive (ag, "ViewVZoomIn", FALSE);
		set_path_sensitive (ag, "ViewVZoomOut", FALSE);
	} else {
		set_path_sensitive (ag, "ZoomIn",
				    marlin_sample_view_can_zoom_in (window->priv->view));
		set_path_sensitive (ag, "ZoomOut",
				    marlin_sample_view_can_zoom_out (window->priv->view));
		set_path_sensitive (ag, "ViewVZoomIn",
				    marlin_sample_view_can_vzoom_in (window->priv->view));
		set_path_sensitive (ag, "ViewVZoomOut",
				    marlin_sample_view_can_vzoom_out (window->priv->view));
	}

	if (window->priv->selection == NULL) {
		set_path_sensitive (ag, "ShowAll", FALSE);
		set_path_sensitive (ag, "ViewCentre", FALSE);
	} else {
		set_path_sensitive (ag, "ShowAll", sensitive);
		set_path_sensitive (ag, "ViewCentre", sensitive);
	}

	if (window->priv->selection &&
	    window->priv->selection->selections == NULL) {
		set_path_sensitive (ag, "ZoomSelection", FALSE);
	} else {
		set_path_sensitive (ag, "ZoomSelection", sensitive);
	}

	if (marlin_program_get_clipboard (program) == NULL) {
		set_path_sensitive (ag, "ViewClipboard", FALSE);
	} else {
		set_path_sensitive (ag, "ViewClipboard", sensitive);
	}
}

void
marlin_window_set_edit_sensitive (MarlinWindow *window,
				  gboolean sensitive)
{
	MarlinProgram *program = marlin_program_get_default ();
	GtkActionGroup *ag;
	gboolean can_remove;
	
	ag = window->priv->ui_action;
	
	if (window->priv->selection == NULL) {
		set_path_sensitive (ag, "EditMoveCursor", FALSE);
	} else {
		set_path_sensitive (ag, "EditMoveCursor", sensitive);
	}

	if (window->priv->selection &&
	    window->priv->selection->selections == NULL) {
		set_path_sensitive (ag, "EditDelete", FALSE);
		set_path_sensitive (ag, "EditCopy", FALSE);
		set_path_sensitive (ag, "EditCut", FALSE);
		set_path_sensitive (ag, "EditCrop", FALSE);
	} else {
		MarlinSampleSubSelection *sub;
		
		sub = window->priv->selection->selections->data;
		can_remove = (sub->coverage == MARLIN_COVERAGE_BOTH);
		set_path_sensitive (ag, "EditDelete", can_remove);
		set_path_sensitive (ag, "EditCopy", sensitive);
		set_path_sensitive (ag, "EditCut", can_remove);
		set_path_sensitive (ag, "EditCrop", can_remove);
	}

	if (marlin_program_get_clipboard (program) == NULL) {
		set_path_sensitive (ag, "EditPasteNew", FALSE);
		set_path_sensitive (ag, "EditPaste", FALSE);
		set_path_sensitive (ag, "EditPasteMix", FALSE);
		set_path_sensitive (ag, "EditReplace", FALSE);
		set_path_sensitive (ag, "EditPasteXFade", FALSE);
	} else {
		set_path_sensitive (ag, "EditPasteNew", sensitive);
		set_path_sensitive (ag, "EditPaste", sensitive);
		set_path_sensitive (ag, "EditPasteMix", sensitive);
		set_path_sensitive (ag, "EditReplace", sensitive);
		set_path_sensitive (ag, "EditPasteXFade", sensitive);
	}

	if (marlin_undo_manager_can_undo (window->priv->undo)) {
		set_path_sensitive (ag, "EditUndo", sensitive);
	} else {
		set_path_sensitive (ag, "EditUndo", FALSE);
	}

	if (marlin_undo_manager_can_redo (window->priv->undo)) {
		set_path_sensitive (ag, "EditRedo", sensitive);
	} else {
		set_path_sensitive (ag, "EditRedo", FALSE);
	}
}

void
marlin_window_set_selection_sensitive (MarlinWindow *window,
				       gboolean sensitive)
{
	GtkActionGroup *ag;

	ag = window->priv->ui_action;

	if (window->priv->selection == NULL) {
		set_path_sensitive (ag, "SelectAll", FALSE);
		set_path_sensitive (ag, "SelectRegion", FALSE);
	} else {
		set_path_sensitive (ag, "SelectAll", sensitive);
		set_path_sensitive (ag, "SelectRegion", sensitive);
	}
	
	if (window->priv->selection &&
	    window->priv->selection->selections == NULL) {
		set_path_sensitive (ag, "SelectNone", FALSE);
		set_path_sensitive (ag, "SelectHalve", FALSE);
		set_path_sensitive (ag, "SelectDouble", FALSE);
		set_path_sensitive (ag, "SelectLeft", FALSE);
		set_path_sensitive (ag, "SelectRight", FALSE);
	} else {
		set_path_sensitive (ag, "SelectNone", sensitive);
		set_path_sensitive (ag, "SelectHalve", TRUE);
		set_path_sensitive (ag, "SelectDouble", TRUE);
		set_path_sensitive (ag, "SelectLeft", TRUE);
		set_path_sensitive (ag, "SelectRight", TRUE);
	}
}

void
marlin_window_set_process_sensitive (MarlinWindow *window,
				     gboolean sensitive)
{
	GtkActionGroup *ag;
	guint channels;
	MarlinMarkerModel *model;
	GList *markers;
	
	ag = window->priv->ui_action;

	g_object_get (G_OBJECT (window->priv->sample),
		      "channels", &channels,
		      "markers", &model,
		      NULL);
	
	g_object_get (G_OBJECT (model),
		      "markers", &markers,
		      NULL);

	if (window->priv->selection == NULL) {
		set_path_sensitive (ag, "ProcessMute", FALSE);
		set_path_sensitive (ag, "ProcessInvert", FALSE);
		set_path_sensitive (ag, "ProcessVolume", FALSE);
	} else {
		set_path_sensitive (ag, "ProcessMute", sensitive);
		set_path_sensitive (ag, "ProcessInvert", sensitive);
		set_path_sensitive (ag, "ProcessVolume", sensitive);
	}

	if (channels != 2) {
		set_path_sensitive (ag, "ProcessSwapChannels", FALSE);
	} else {
		set_path_sensitive (ag, "ProcessSwapChannels", sensitive);
	}

	set_path_sensitive (ag, "ProcessAdjustChannels", TRUE);

	if (markers == NULL) {
		set_path_sensitive (ag, "ProcessSplitSample", FALSE);
	} else {
		set_path_sensitive (ag, "ProcessSplitSample", sensitive);
	}
}

static void
set_path_name (GtkActionGroup *ag,
	       const char *path,
	       const char *text)
{
	GtkAction *action;

	action = gtk_action_group_get_action (ag, path);
	g_object_set (action, "label", text, NULL);
}

void
marlin_window_set_process_name (MarlinWindow *window)
{
	GtkActionGroup *ag;
	int num_chans;

	ag = window->priv->ui_action;

	if (window->priv->selection &&
	    window->priv->selection->selections == NULL) {
		/* No selection */
		set_path_name (ag, "ProcessMute", _("Mute Sample"));
		set_path_name (ag, "ProcessInvert", _("Invert Sample"));
		set_path_name (ag, "ProcessVolume", _("Adjust Sample Volume..."));
	} else {
		/* Selection */
		set_path_name (ag, "ProcessMute", _("Mute Selection"));
		set_path_name (ag, "ProcessInvert", _("Invert Selection"));
		set_path_name (ag, "ProcessVolume", _("Adjust Selection Volume..."));
	}

	g_object_get (G_OBJECT (window->priv->sample),
		      "channels", &num_chans,
		      NULL);
	if (num_chans == 1) {
		set_path_name (ag, "ProcessAdjustChannels", _("Add a Channel..."));
	} else {
		set_path_name (ag, "ProcessAdjustChannels", _("Remove a Channel..."));
	}
}

void
marlin_window_set_edit_name (MarlinWindow *window)
{
	GtkActionGroup *ag;

	ag = window->priv->ui_action;

	if (window->priv->selection &&
	    window->priv->selection->selections == NULL) {
		/* No selection */
		set_path_name (ag, "EditPasteMix", _("Mix..."));
		set_path_name (ag, "EditReplace", _("Replace"));
	} else {
		/* Selection */
		set_path_name (ag, "EditPasteMix", _("Mix into Selection..."));
		set_path_name (ag, "EditReplace", _("Replace Selection"));
	}

	if (marlin_undo_manager_can_undo (window->priv->undo)) {
		char *name;

		name = g_strdup_printf (_("Undo '%s'"),
					  marlin_undo_manager_get_undo_name (window->priv->undo));
		set_path_name (ag, "EditUndo", name);
		g_free (name);
	} else {
		set_path_name (ag, "EditUndo", _("Undo"));
	}

	if (marlin_undo_manager_can_redo (window->priv->undo)) {
		char *name;

		name = g_strdup_printf (_("Redo '%s'"),
					  marlin_undo_manager_get_redo_name (window->priv->undo));
		set_path_name (ag, "EditRedo", name);
		g_free (name);
	} else {
		set_path_name (ag, "EditRedo", _("Redo"));
	}					  
}

MarlinSample *
marlin_window_get_sample (MarlinWindow *window)
{
	g_return_val_if_fail (IS_MARLIN_WINDOW (window), NULL);
	
	return window->priv->sample;
}

void
marlin_window_play_range (MarlinWindow *window,
			  guint64 start,
			  guint64 finish)
{
	MarlinWindowPrivate *priv;

	g_return_if_fail (IS_MARLIN_WINDOW (window));
	g_return_if_fail (IS_MARLIN_PLAY_PIPELINE (window->priv->play_pipeline));
	priv = window->priv;

	if (window->priv->state == MARLIN_STATE_PLAYING) {
		marlin_pipeline_set_state (MARLIN_PIPELINE (priv->play_pipeline),
					   GST_STATE_PAUSED);
		priv->state = MARLIN_STATE_PAUSED;
	}

	priv->start = start;
	priv->finish = finish;
	
	priv->state = MARLIN_STATE_PLAYING;

	/* The pipeline state must be >= PAUSED to send an event
	   so we set it here temporarily */
	marlin_pipeline_set_state (MARLIN_PIPELINE (priv->play_pipeline),
				   GST_STATE_PAUSED);

	marlin_play_pipeline_seek_range (priv->play_pipeline, start, finish);
	marlin_pipeline_set_state (MARLIN_PIPELINE (priv->play_pipeline),
				   GST_STATE_PLAYING);
}

void
marlin_window_play_sample (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	guint64 start, finish, frames;

	g_return_if_fail (IS_MARLIN_WINDOW (window));
	g_return_if_fail (IS_MARLIN_PLAY_PIPELINE (window->priv->play_pipeline));
	priv = window->priv;

	if (window->priv->state == MARLIN_STATE_PLAYING) {
		return;
	} else if (window->priv->state == MARLIN_STATE_PAUSED) {
		window->priv->state = MARLIN_STATE_PLAYING;
		marlin_pipeline_set_state (MARLIN_PIPELINE (priv->play_pipeline),
					   GST_STATE_PLAYING);
		return;
	}
	
	/* Check if we've got a selection */
	marlin_read_write_lock_lock (priv->selection->lock,
				     MARLIN_READ_WRITE_LOCK_MODE_READ);
	if (priv->selection->selections != NULL) {
		MarlinSampleSubSelection *sub = priv->selection->selections->data;
		start = sub->start;
		priv->initial = start;
		finish = sub->end;

		g_print ("Loop selection: %llu, %llu\n", start, finish);
	} else {
		g_object_get (G_OBJECT (priv->view),
			      "cursor_position", &priv->initial,
			      NULL);
		g_object_get (G_OBJECT (priv->sample),
			      "total_frames", &finish,
			      NULL);
		start = priv->initial;
	}

	marlin_read_write_lock_unlock (priv->selection->lock,
				       MARLIN_READ_WRITE_LOCK_MODE_READ);

	marlin_window_play_range (window, start, finish);
}

void
marlin_window_repeat_sample (MarlinWindow *window)
{
	window->priv->repeating = TRUE;
 	marlin_window_play_sample (window);
}

void
marlin_window_stop_sample (MarlinWindow *window)
{
	if (window->priv->state == MARLIN_STATE_STOPPED) {
		return;
	}
	
	window->priv->state = MARLIN_STATE_STOPPED;

	marlin_pipeline_set_state (MARLIN_PIPELINE (window->priv->play_pipeline), GST_STATE_READY);
	
	/* Reset the position to the initial */
	marlin_window_move_cursor_to (window, window->priv->initial);
	window->priv->repeating = FALSE;
}

void
marlin_window_pause_sample (MarlinWindow *window)
{
	if (window->priv->state == MARLIN_STATE_PAUSED) {
		return;
	}
	
	window->priv->state = MARLIN_STATE_PAUSED;
	marlin_pipeline_set_state (MARLIN_PIPELINE (window->priv->play_pipeline), GST_STATE_PAUSED);
}

void
marlin_window_resume_sample (MarlinWindow *window)
{
}

void
marlin_window_sample_completed (MarlinWindow *window)
{
	if (window->priv->repeating) {
		window->priv->state = MARLIN_STATE_PLAYING;
	} else {
		marlin_window_stop_sample (window);
	}
}
		
/* Save yourself :) */
static void
set_int (GConfClient *client,
	 const char *root,
	 const char *key,
	 int value)
{
	char *fullkey;

	fullkey = g_strdup_printf ("%s/%s", root, key);
	gconf_client_set_int (client, fullkey, value, NULL);
	g_free (fullkey);
}

void
marlin_window_save_state (MarlinWindow *window,
			  gpointer closure)
{
	uuid_t uuid;
	char uid[100];
	char *gconf_root, *window_root, *key;
	GConfClient *client;
	GConfValue *value, *window_value;
	GSList *window_id_list;

	int x, y, width, height, cursor;
	char *filename;

	client = gconf_client_get_default ();

	if (client == NULL) {
		g_warning ("Cannot save session");
		return;
	}
	
	/* Make a unique ID for this window */
	uuid_generate (uuid);
	uuid_unparse (uuid, uid);
	
	gconf_root = "/apps/marlin/system-state/";

	key = g_strdup_printf ("%s%s", gconf_root, "window-list");
	window_value = gconf_client_get (client, key, NULL);

	if (window_value == NULL) {

		/* Create a new LIST value */
		window_value = gconf_value_new (GCONF_VALUE_LIST);
		gconf_value_set_list_type (window_value, GCONF_VALUE_STRING);
		window_id_list = NULL;
	} else {
		window_id_list = gconf_value_get_list (window_value);
	}

	/* Create a new string */
	value = gconf_value_new (GCONF_VALUE_STRING); 
	gconf_value_set_string (value, uid);

	/* Stick it at the start of the list */
	window_id_list = g_slist_prepend (window_id_list, value);

	/* Set it back in the value */

	gconf_value_set_list (window_value, window_id_list);
	gconf_client_set (client, key, window_value, NULL);
	g_free (key);

	/* Now store our state into the window_root */
	window_root = g_strdup_printf ("%s%s", gconf_root, uid);

	/* Filename */
	g_object_get (G_OBJECT (window->priv->sample),
		      "filename", &filename,
		      NULL);
	key = g_strdup_printf ("%s/filename", window_root);
	gconf_client_set_string (client, key, filename, NULL);
	g_free (key);
	
	/* Geometry */
	gtk_window_get_position (GTK_WINDOW (window), &x, &y);
	gtk_window_get_size (GTK_WINDOW (window), &width, &height);

	set_int (client, window_root, "x", x);
	set_int (client, window_root, "y", y);
	set_int (client, window_root, "width", width);
	set_int (client, window_root, "height", height);

	/* Cursor position */
	g_object_get (G_OBJECT (window->priv->view),
		      "cursor_position", &cursor,
		      NULL);

	set_int (client, window_root, "cursor-position", cursor);
	/* Save the selection list here */

	g_free (window_root);
	g_object_unref (G_OBJECT (client));
}

static void
remove_key (GConfClient *client,
	    const char *root,
	    const char *key)
{
	char *fullkey;

	fullkey = g_strdup_printf ("/apps/marlin/system-state/%s/%s", root, key);
	gconf_client_unset (client, fullkey, NULL);
	g_free (fullkey);
}

void
marlin_window_remove_state (const char *state_root)
{
	GConfClient *client;

	client = gconf_client_get_default ();

	remove_key (client, state_root, "filename");
	remove_key (client, state_root, "x");
	remove_key (client, state_root, "y");
	remove_key (client, state_root, "width");
	remove_key (client, state_root, "height");
	remove_key (client, state_root, "cursor-position");

	g_object_unref (G_OBJECT (client));
}

struct _load_op_data {
	MarlinOperation *op;
	MarlinPipeline *pipeline;
	MarlinWindow *window;
};

static void
load_op_started (MarlinOperation *operation,
		 struct _load_op_data *lod)
{
	MarlinWindow *window = lod->window;
	MarlinSample *sample = window->priv->sample;
	
	marlin_window_set_media_bar_sensitive (window, FALSE);
	window->priv->progress_window = marlin_progress_window_new (sample,
								    operation,
								    GTK_WINDOW (window),
								    _("Loading %s"));
}

static void
load_op_finished (MarlinOperation *operation,
		  struct _load_op_data *lod)
{
	MarlinWindow *window = lod->window;
	MarlinSample *sample;
	guint64 frames;

	sample = window->priv->sample;
	marlin_progress_window_destroy (window->priv->progress_window);
	window->priv->progress_window == NULL;

	marlin_window_set_title (window, sample);
	
	marlin_window_set_file_sensitive (window, TRUE);
	marlin_window_set_media_bar_sensitive (window, TRUE);
	marlin_window_set_view_sensitive (window, TRUE);
	marlin_window_set_edit_sensitive (window, TRUE);
	marlin_window_set_selection_sensitive (window, TRUE);
	marlin_window_set_process_sensitive (window, TRUE);
	marlin_window_set_process_name (window);

	if (window->priv->close_on_op_complete) {
		gtk_widget_destroy (GTK_WIDGET (window));
	}

	g_object_set (G_OBJECT (window->priv->overview_bar),
		      "sample", sample,
		      NULL);
	g_object_set (G_OBJECT (window->priv->view),
		      "sample", sample,
		      NULL);

	/* Kill the pipeline */
  	marlin_pipeline_set_state (lod->pipeline, GST_STATE_NULL);
}

static void
load_op_cancelled (MarlinOperation *operation,
		   struct _load_op_data *lod)
{
	MarlinWindow *window = lod->window;
	MarlinSample *sample;

	marlin_progress_window_destroy (window->priv->progress_window);
	window->priv->progress_window = NULL;

	sample = g_object_new (MARLIN_SAMPLE_TYPE,
			       "filename", "untitled",
			       "channels", 2,
			       "sample-rate", 44100,
			       NULL);

	g_object_set (G_OBJECT (window),
		      "sample", sample,
		      NULL);

	/* Samples need to be unreffed once they're attached to a window */
	g_object_unref (G_OBJECT (sample));

	marlin_pipeline_set_state (lod->pipeline, GST_STATE_NULL);
}

static void
load_op_error (MarlinOperation *operation,
	       GError *error,
	       const char *debug_message,
	       struct _load_op_data *lod)
{
	MarlinWindow *window = lod->window;
	MarlinSample *sample;
	GtkWidget *dialog;
	char *markup, *message, *filename;

	marlin_progress_window_destroy (window->priv->progress_window);
	window->priv->progress_window = NULL;
	
	g_object_get (G_OBJECT (window->priv->sample),
		      "filename", &filename,
		      NULL);
	
	markup = g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>\n%s",
				  _("Could not load '%s'"),
				  error->message);
	message = g_strdup_printf (markup, filename);
	g_free (markup);
	g_free (filename);

	dialog = gtk_message_dialog_new (NULL, 0,
					 GTK_MESSAGE_ERROR,
					 GTK_BUTTONS_OK,
					 message);

	gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (dialog)->label), TRUE);
	gtk_label_set_line_wrap (GTK_LABEL (GTK_MESSAGE_DIALOG (dialog)->label), FALSE);

	g_free (message);

	g_signal_connect (dialog, "response",
			  G_CALLBACK (gtk_widget_destroy), NULL);
	gtk_widget_show (dialog);

	sample = g_object_new (MARLIN_SAMPLE_TYPE,
			       "filename", "untitled",
			       "channels", 2,
			       "sample-rate", 44100,
			       NULL);

	g_object_set (G_OBJECT (window),
		      "sample", sample,
		      NULL);

	g_object_unref (G_OBJECT (sample));

	marlin_pipeline_set_state (lod->pipeline, GST_STATE_NULL);
}

static void
load_pipeline_shutdown (MarlinPipeline *pipeline,
			struct _load_op_data *lod)
{
	g_object_unref (G_OBJECT (lod->pipeline));
	g_free (lod);
}

void
marlin_window_load_file (MarlinWindow *window,
			 const char *filename)
{
	struct _load_op_data *lod;
	MarlinOperation *operation;
	MarlinLoadPipeline *pipeline;

	lod = g_new (struct _load_op_data, 1);

	lod->op = operation = marlin_operation_new ();
	lod->window = window;

	g_signal_connect (G_OBJECT (operation), "started",
			  G_CALLBACK (load_op_started), lod);
	g_signal_connect (G_OBJECT (operation), "finished",
			  G_CALLBACK (load_op_finished), lod);
	g_signal_connect (G_OBJECT (operation), "cancelled",
			  G_CALLBACK (load_op_cancelled), lod);
	g_signal_connect (G_OBJECT (operation), "error",
			  G_CALLBACK (load_op_error), lod);

	
	pipeline = marlin_load_pipeline_new (operation);
	lod->pipeline = pipeline;
	g_signal_connect (G_OBJECT (pipeline), "shutdown",
			  G_CALLBACK (load_pipeline_shutdown), lod);
	
	g_object_set (G_OBJECT (pipeline),
		      "filename", filename,
		      "sample", window->priv->sample,
		      NULL);

	/* Go */
	marlin_pipeline_set_state (MARLIN_PIPELINE (pipeline), GST_STATE_PLAYING);
}

struct _save_op_data {
	MarlinOperation *op;
	MarlinPipeline *pipeline;
	MarlinWindow *window;

	char *temp_filename;
	char *filename;
};

static void
save_op_started (MarlinOperation *operation,
		 struct _save_op_data *sod)
{
	MarlinWindow *window = sod->window;

	window->priv->save_progress_window = marlin_progress_window_new (window->priv->sample,
									 operation,
									 GTK_WINDOW (window),
									 _("Saving %s"));

	/* Show the window immediatly */
	gtk_widget_show (window->priv->save_progress_window->dialog);
}

static void
save_op_finished (MarlinOperation *operation,
		  struct _save_op_data *sod)
{
	MarlinWindow *window = sod->window;
	MarlinPipeline *pipeline;
	
	marlin_progress_window_destroy (window->priv->save_progress_window);
	window->priv->save_progress_window = NULL;

	pipeline = sod->pipeline;
	g_print ("Save-Op finished\n");

	/* Reset the time of last save */
	window->priv->last_save = time (NULL);
	
	g_object_set (G_OBJECT (window->priv->sample),
		      "dirty", FALSE,
		      NULL);

	/* Copy the new file over the old one...only remove the temp file
	   if the copy succeeded */
	if (marlin_file_copy (sod->temp_filename, sod->filename, NULL)) {
		unlink (sod->temp_filename);
	} else {
		/* FIXME: Why did it fail? Show user a decent dialog */
		g_print ("Copying file to %s failed.\nFile has been left as %s\n", sod->filename, sod->temp_filename);
	}

	marlin_pipeline_set_state (pipeline, GST_STATE_NULL);
}

static void
save_op_cancelled (MarlinOperation *operation,
		   struct _save_op_data *sod)
{
	MarlinWindow *window = sod->window;
	MarlinPipeline *pipeline = sod->pipeline;
	
	marlin_progress_window_destroy (window->priv->save_progress_window);
	window->priv->save_progress_window = NULL;

	marlin_pipeline_set_state (pipeline, GST_STATE_NULL);
}
				      
static void
save_shutdown (MarlinPipeline *pipeline,
	       struct _save_op_data *sod)
{
	g_print ("Save shutdown\n");
	g_object_unref (G_OBJECT (pipeline));

	g_free (sod->filename);
	g_free (sod->temp_filename);
	g_free (sod);
}

typedef void (* options_dialog) (MarlinWindow *window, MarlinPipeline *pipeline);
struct _maps {
	char *mimetype;
	options_dialog function;
};

static void
marlin_default_save (MarlinWindow *window,
		     MarlinPipeline *pipeline)
{
	marlin_pipeline_set_state (pipeline, GST_STATE_PLAYING);
}

struct _maps mimetypemap[] = {
	{ "audio/x-mp3", marlin_mp3_dialog_new },
	{ "audio/mpeg", marlin_mp3_dialog_new },
#if 0
	{ "application/ogg", marlin_vorbis_dialog_new },
	{ "application/x-ogg", marlin_vorbis_dialog_new },
#endif
	{ NULL, marlin_default_save },
	{ NULL, NULL }
};

static void
save_sample_as (MarlinWindow *window,
		const char *filename,
		GstElement *encoder)
{
	struct _save_op_data *sod;
	MarlinOperation *operation;
	MarlinSavePipeline *pipeline;
	const char *mimetype;
	char *path, *name;
	int i;

	sod = g_new (struct _save_op_data, 1);
	mimetype = gnome_vfs_mime_type_from_name (filename);
	
	sod->op = operation = marlin_operation_new ();
	sod->window = window;

	g_signal_connect (G_OBJECT (operation), "started",
			  G_CALLBACK (save_op_started), sod);
	g_signal_connect (G_OBJECT (operation), "finished",
			  G_CALLBACK (save_op_finished), sod);
	g_signal_connect (G_OBJECT (operation), "cancelled",
			  G_CALLBACK (save_op_cancelled), sod);
	
	sod->filename = g_strdup (filename);
	path = g_path_get_dirname (filename);
	name = g_path_get_basename (filename);

	if (path == NULL) {
		sod->temp_filename = g_strdup_printf (".%s.XXXXXX", name);
	} else {
		char *t = g_strdup_printf (".%s.XXXXXX", name);
		mktemp (t);

		sod->temp_filename = g_build_filename (path, t, NULL);
		g_free (t);
	}
	g_free (path);
	g_free (name);

	if (encoder == NULL) {
		pipeline = g_object_new (MARLIN_SAVE_PIPELINE_TYPE,
					 "operation", operation,
					 "sample", window->priv->sample,
					 "filename", sod->temp_filename,
					 "mimetype", mimetype,
					 NULL);
	} else {
		pipeline = g_object_new (MARLIN_SAVE_PIPELINE_TYPE,
					 "operation", operation,
					 "sample", window->priv->sample,
					 "filename", sod->temp_filename,
					 "encoder", encoder,
					 NULL);
	}

	sod->pipeline = (MarlinPipeline *) pipeline;
	g_signal_connect (G_OBJECT (pipeline), "shutdown",
			  G_CALLBACK (save_shutdown), sod);
	
	if (encoder != NULL) {
		/* Caller has supplied an encoder, so the parameters have
		   already been set */
		marlin_default_save (window, MARLIN_PIPELINE (pipeline));
		g_object_set (G_OBJECT (window->priv->sample),
			      "encoder", encoder,
			      NULL);
		return;
	}

	for (i = 0; mimetypemap[i].function; i++) {
		if (mimetypemap[i].mimetype == NULL) {
			marlin_default_save (window, MARLIN_PIPELINE (pipeline));
			return;
		}
		
		if (strcmp (mimetype, mimetypemap[i].mimetype) == 0) {
			mimetypemap[i].function (window, MARLIN_PIPELINE (pipeline));
		 	return;
		}
	}
} 

void
marlin_window_save_sample (MarlinWindow *window)
{
	char *filename;
	GstElement *encoder;
	
 	g_object_get (G_OBJECT (window->priv->sample),
		      "filename", &filename,
		      "encoder", &encoder,
		      NULL);

	if (encoder == NULL) {
		const char *mimetype;

		/* If there was no encoder then we need to check that we 
		   can save to this format */

		mimetype = gnome_vfs_mime_type_from_name (filename);
		if (marlin_gst_can_encode (mimetype) == FALSE) {
			GtkWidget *dialog;
			char *text;

			text = g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>\n%s",
						_("Marlin is unable to save %s\nas there is no encoder for %s."),
						_("Would you like to save as a different format?"));

			dialog = gtk_message_dialog_new (GTK_WINDOW (window),
							 GTK_DIALOG_MODAL,
							 GTK_MESSAGE_ERROR,
							 GTK_BUTTONS_NONE,
							 text,
							 filename, mimetype);
			g_free (text);
			
			/* Naughty */
			gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (dialog)->label), TRUE);
			gtk_label_set_line_wrap (GTK_LABEL (GTK_MESSAGE_DIALOG (dialog)->label), FALSE);
			gtk_dialog_add_button (GTK_DIALOG (dialog),
					       GTK_STOCK_CANCEL,
					       GTK_RESPONSE_CANCEL);
			gtk_dialog_add_button (GTK_DIALOG (dialog),
					       GTK_STOCK_SAVE_AS,
					       GTK_RESPONSE_OK);

			switch (gtk_dialog_run (GTK_DIALOG (dialog))) {
			case GTK_RESPONSE_CANCEL:
				gtk_widget_destroy (dialog);
				break;

			case GTK_RESPONSE_OK:
				gtk_widget_destroy (dialog);
				marlin_window_save_sample_as (window);
				break;
				
			default:
				break;
			}
			
			g_free (filename);
			return;
		}
	}
	
	save_sample_as (window, filename, encoder);
	g_free (filename);
}

struct _FileSaveData {
	GtkWidget *filesel;
	MarlinWindow *window;
#ifdef HAVE_MEDIA_PROFILES
	GtkWidget *profile_hbox;
	GtkWidget *type_combo, *profiles;
#endif
};

#ifdef HAVE_MEDIA_PROFILES
enum {
	EXTENSION,
	PROFILE
};
#endif

/* Returns TRUE to write file, FALSE to not */
static gboolean
ask_overwrite_dialog (GtkWindow *window,
		      const char *name)
{
	GtkWidget *overwrite;
	char *text;
	char *backup;
	
	/* Ask to overwrite */
	text = g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>\n%s",
				_("The file '%s' already exists."),
				_("Do you want to make a backup called '%s' first?"));
	
	backup = g_strdup_printf ("%s~", name);
	overwrite = gtk_message_dialog_new (window,
					    GTK_DIALOG_MODAL,
					    GTK_MESSAGE_QUESTION,
					    GTK_BUTTONS_NONE,
					    text, name, backup);
	g_free (text);
	
	/* Naughty */
	gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (overwrite)->label), TRUE);
	gtk_label_set_line_wrap (GTK_LABEL (GTK_MESSAGE_DIALOG (overwrite)->label), FALSE);
	
	gtk_dialog_add_button (GTK_DIALOG (overwrite),
			       _("Overwrite File"),
			       GTK_RESPONSE_NO);
	gtk_dialog_add_button (GTK_DIALOG (overwrite),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (overwrite),
			       _("Backup File"),
			       GTK_RESPONSE_YES);
	
	switch (gtk_dialog_run (GTK_DIALOG (overwrite))) {
	case GTK_RESPONSE_YES:
		gtk_widget_destroy (overwrite);
		
		if (marlin_file_copy (name, backup, NULL) == FALSE) {
			g_warning ("Error making backup...Not saving so that there is no data loss");
			g_free (backup);
			return FALSE;
		}
		
		g_free (backup);
		return TRUE;
		
	case GTK_RESPONSE_NO:
		gtk_widget_destroy (overwrite);
		g_free (backup);
		
		return TRUE;
		
	default:
		gtk_widget_destroy (overwrite);
		g_free (backup);
		
		return FALSE;
	}
}

/* Returns TRUE to write file, FALSE to not */
static gboolean
ask_nospace_write_dialog (GtkWindow *window,
			  const char *name)
{
	GtkWidget *overwrite;
	char *text;
	
	/* Ask to overwrite */
	text = g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>\n%s",
				_("The file '%s' already exists."),
				_("There is not enough space to create a backup.\nDo you want to overwrite the file?"));
	
	overwrite = gtk_message_dialog_new (window,
					    GTK_DIALOG_MODAL,
					    GTK_MESSAGE_QUESTION,
					    GTK_BUTTONS_NONE,
					    text, name);
	g_free (text);
	
	/* Naughty */
	gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (overwrite)->label), TRUE);
	gtk_label_set_line_wrap (GTK_LABEL (GTK_MESSAGE_DIALOG (overwrite)->label), FALSE);
	
	gtk_dialog_add_button (GTK_DIALOG (overwrite),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (overwrite),
			       _("Overwrite File"),
			       GTK_RESPONSE_YES);
	
	switch (gtk_dialog_run (GTK_DIALOG (overwrite))) {
	case GTK_RESPONSE_YES:
		gtk_widget_destroy (overwrite);
		return TRUE;
		
	default:
		gtk_widget_destroy (overwrite);
		return FALSE;
	}
}

static gboolean
is_valid_filename (GtkWindow *window,
		   const char *name)
{
	if (g_file_test (name, G_FILE_TEST_EXISTS) == FALSE) {
		return TRUE;
	}
	
	if (g_file_test (name, G_FILE_TEST_IS_DIR)) {
		GtkWidget *isdir;
		char *text;

		text = g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>\n\n%s",
					_("Unable to save file to '%s'."),
					_("'%s' is a directory."));

		isdir = gtk_message_dialog_new (window,
						0, GTK_MESSAGE_ERROR,
						GTK_BUTTONS_CLOSE,
						text,
						name, name);
		g_free (text);

		/* Naughty */
		gtk_label_set_use_markup (GTK_LABEL (GTK_MESSAGE_DIALOG (isdir)->label), TRUE);

		gtk_dialog_run (GTK_DIALOG (isdir));
		gtk_widget_destroy (isdir);
		
		return FALSE;
	}

	if (g_file_test (name, G_FILE_TEST_EXISTS)) {
		GnomeVFSFileSize size;
		GnomeVFSResult result;
		GnomeVFSURI *uri;
		char *text_uri;
		struct stat statbuf;

		text_uri = g_strdup_printf ("file://%s", name);
		uri = gnome_vfs_uri_new (text_uri);
		g_free (text_uri);

		result = gnome_vfs_get_volume_free_space (uri, &size);
		gnome_vfs_uri_unref (uri);

		if (result != GNOME_VFS_OK) {
			return ask_overwrite_dialog (window, name);
		} else {
			/* Check the there's enough space */
			if (stat (name, &statbuf) != 0) {
				g_warning ("stat on %s failed: %s", name,
					   g_strerror (errno));
				return FALSE;
			}
			
			if (statbuf.st_size >= size) {
				return ask_nospace_write_dialog (window, name);
			} else {
				return ask_overwrite_dialog (window, name);
			}
		}
	}

	return FALSE;
}
		
static void
file_sel_save_as_response (GtkDialog *filesel,
			   int response_id,
			   struct _FileSaveData *sd)
{
	MarlinWindow *window = sd->window;
	GConfClient *client;
	GstElement *encoder = NULL;
	const char *name;
	char *dirname, *d;
	int i;

	switch (response_id) {
	case GTK_RESPONSE_OK:
		name = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (filesel));
		if (is_valid_filename (GTK_WINDOW (filesel), name) == FALSE) {
			return;
		}
		
		d = g_path_get_dirname (name);
		/* Add a trailing / so that the file selector will cd into it */
		dirname = g_strdup_printf ("%s/", d);
		
		client = gconf_client_get_default ();
		gconf_client_set_string (client,
					 "/apps/marlin/system-state/save-file-directory",
					 dirname, NULL);
		g_object_unref (G_OBJECT (client));
		g_free (dirname);
		g_free (d);

#ifdef HAVE_MEDIA_PROFILES
		{
			int type;

			type = gtk_combo_box_get_active (GTK_COMBO_BOX (sd->type_combo));
			if (type == EXTENSION) {
				encoder = NULL;
			} else {
				const char *prof;
				char *pipeline;
				GstPad *src, *sink;
				GstElement *id;
				GError *err = NULL;
				GMAudioProfile *profile = gm_audio_profile_choose_get_active (sd->profiles);
				
				prof = gm_audio_profile_get_pipeline (profile);
				pipeline = g_strdup_printf ("(name=profile-encoder identity name=encoder_start ! %s ! identity name=encoder_end )", prof);
				encoder = gst_parse_launch (pipeline, &err);
				if (err != NULL) {
					g_warning ("Error parsing pipeline");
					g_free (pipeline);
					g_error_free (err);
					encoder = NULL;
				} else {			
					/* Hook up ghost pads to the start and end */
					id = gst_bin_get_by_name (GST_BIN (encoder), "encoder_start");
					g_assert (id != NULL);
					
					sink = gst_element_get_pad (id, "sink");
					gst_element_add_ghost_pad (encoder, sink, "sink-ghost");
					
					id = gst_bin_get_by_name (GST_BIN (encoder), "encoder_end");
					g_assert (id != NULL);
					
					src = gst_element_get_pad (id, "src");
					gst_element_add_ghost_pad (encoder, src, "src-ghost");
					
/* 					g_print ("Using %s as pipeline\n", pipeline); */
					g_free (pipeline);
				}
			}
		}
#endif
		/* Set the sample's filename */
		g_object_set (G_OBJECT (window->priv->sample),
			      "filename", name,
			      NULL);

		save_sample_as (window, name, encoder);
		break;
		
	case GTK_RESPONSE_CANCEL:
	default:
		break;
		
	}

	gtk_widget_destroy (GTK_WIDGET (filesel));
	g_free (sd);
}

#ifdef HAVE_MEDIA_PROFILES
static void
add_formats (struct _FileSaveData *sd)
{
	int i;
	static char *formats[] = {
		N_("From file extension"),
		N_("From profile"),
		NULL
	};

	for (i = 0; formats[i] != NULL; i++) {
		gtk_combo_box_append_text (GTK_COMBO_BOX (sd->type_combo), _(formats[i]));
	}

	gtk_combo_box_set_active (GTK_COMBO_BOX (sd->type_combo), 0);
	gtk_widget_set_sensitive (sd->profile_hbox, FALSE);
}

static void
type_combo_changed (GtkComboBox *combo,
		    struct _FileSaveData *sd)
{
	int active = gtk_combo_box_get_active (combo);

	gtk_widget_set_sensitive (sd->profile_hbox, (active == 1));
}
#endif

static void
save_dialog_add_options (struct _FileSaveData *sd)
{
#ifdef HAVE_MEDIA_PROFILES
	GtkWidget *table, *label;

	table = gtk_table_new (2, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);
	gtk_widget_show (table);

	label = gtk_label_new_with_mnemonic (_("Determine _Format:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	PACK (table, label, 0, 0, GTK_FILL);

	sd->type_combo = gtk_combo_box_new_text ();
	gtk_widget_show (sd->type_combo);
	PACK (table, sd->type_combo, 1, 0, GTK_FILL);

	g_signal_connect (G_OBJECT (sd->type_combo), "changed",
			  G_CALLBACK (type_combo_changed), sd);
	sd->profile_hbox = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (sd->profile_hbox);
	gtk_table_attach (GTK_TABLE (table), sd->profile_hbox,
			  1, 2, 1, 2,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	label = gtk_label_new_with_mnemonic (_("_Profile:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (sd->profile_hbox), label, FALSE, FALSE, 0);
	sd->profiles = gm_audio_profile_choose_new ();
	gtk_widget_show (sd->profiles);
	gtk_box_pack_start (GTK_BOX (sd->profile_hbox), sd->profiles,
			    TRUE, TRUE, 0);

	add_formats (sd);

	gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER (sd->filesel), table);
#endif
}

void
marlin_window_save_sample_as (MarlinWindow *window)
{
	struct _FileSaveData *sd;
	GConfClient *client;
	char *directory, *name, *filename, *sname;

	sd = g_new (struct _FileSaveData, 1);
	
	g_object_get (G_OBJECT (window->priv->sample),
		      "filename", &filename,
		      "name", &name,
		      NULL);

	sd->window = window;
	sd->filesel = marlin_file_save_dialog_new (GTK_WINDOW (window));
	
	sname = filename ? g_path_get_basename (filename) : g_strdup (name);
	gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (sd->filesel),
					   sname);
	g_free (sname);
	g_free (filename);
	g_free (name);

	save_dialog_add_options (sd);

	client = gconf_client_get_default ();
	directory = gconf_client_get_string (client,
					     "/apps/marlin/system-state/save-file-directory",
					     NULL);
	if (directory != NULL &&
	    *directory != 0) {
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (sd->filesel), directory);
	}
	g_free (directory);
	g_object_unref (G_OBJECT (client));
	
	g_signal_connect (G_OBJECT (sd->filesel), "response",
			  G_CALLBACK (file_sel_save_as_response), sd);
	
	gtk_widget_show (sd->filesel);
}

struct _FileOpenData {
	GtkWidget *filesel;
	MarlinWindow *window;
};

static void
file_sel_response (GtkDialog *filesel,
		   int response_id,
		   struct _FileOpenData *fd)
{
	const char *name;
	char *dirname, *d;
	int i;

	switch (response_id) {
	case GTK_RESPONSE_OK:
		name = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (filesel));
		d = g_path_get_dirname (name);
		/* Add a trailing / so that the file selector will cd into it */
		dirname = g_strdup_printf ("%s/", d);
		
		marlin_gconf_set_string ("/apps/marlin/system-state/open-file-directory", dirname);
		g_free (dirname);
		g_free (d);
		
		marlin_open_window (name, NULL);
		break;

	case GTK_RESPONSE_CANCEL:
	default:
		break;
		
	}

	gtk_widget_destroy (GTK_WIDGET (filesel));
	g_free (fd);
}

void
marlin_window_open (MarlinWindow *window)
{
	GtkWidget *options;
	static struct _FileOpenData *fd = NULL;
	char *directory;

	if (fd != NULL) {
		gdk_window_raise (fd->filesel->window);
		gdk_window_show (fd->filesel->window);
		return;
	}
	
	fd = g_new (struct _FileOpenData, 1);
	
	fd->window = window;
	fd->filesel = marlin_file_open_dialog_new (GTK_WINDOW (window));
	
	directory = marlin_gconf_get_string ("/apps/marlin/system-state/open-file-directory");
	if (directory != NULL &&
	    *directory != 0) {
#if 0
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (fd->filesel),
 						     directory);
#endif
		g_free (directory);
	}
	
	g_signal_connect (G_OBJECT (fd->filesel), "response",
			  G_CALLBACK (file_sel_response), fd);
	g_signal_connect (G_OBJECT (fd->filesel), "destroy",
			  G_CALLBACK (gtk_widget_destroyed), &fd);
	
	gtk_widget_show (fd->filesel);
}

struct _select_closure {
	MarlinWindow *window;
	guint64 start, end;
	MarlinCoverage coverage;
};

static void
select_undo (gpointer data)
{
	struct _select_closure *c = data;

	if (c->start == 0 && c->end == 0) {
		marlin_window_select_none (c->window, NULL);
	} else {
		MarlinSampleSubSelection *sub = NULL;

		WRITE_LOCK (c->window->priv->selection->lock);
		if (c->window->priv->selection->selections) {
			sub = c->window->priv->selection->selections->data;
		}

		if (sub) {
			sub->start = c->start;
			sub->end = c->end;
			sub->coverage = c->coverage;
			WRITE_UNLOCK (c->window->priv->selection->lock);
			marlin_sample_selection_changed (c->window->priv->selection,
							 FALSE);
		} else {
			WRITE_UNLOCK (c->window->priv->selection->lock);
			marlin_sample_selection_add (c->window->priv->selection,
						     c->start, c->end, c->coverage);
		}
	}
}

static void
select_all_redo (gpointer data)
{
	struct _select_closure *c = data;
	marlin_window_select_all (c->window, NULL);
}

static void
select_destroy (gpointer data)
{
	g_free (data);
}

void
marlin_window_select_all (MarlinWindow *window,
			  MarlinUndoContext *ctxt)
{
	guint64 total;

	if (ctxt) {
		struct _select_closure *c;
		MarlinUndoable *u;
		MarlinSampleSubSelection *sub;

		READ_LOCK (window->priv->selection->lock);
		sub = window->priv->selection->selections->data;

		c = g_new0 (struct _select_closure, 1);
		c->window = window;

		if (sub) {
			c->start = sub->start;
			c->end = sub->end;
			c->coverage = sub->coverage;
		} else {
			c->coverage = MARLIN_COVERAGE_BOTH;
		}

		READ_UNLOCK (window->priv->selection->lock);

		u = marlin_undoable_new (select_undo,
					 select_all_redo,
					 select_destroy,
					 c);
		marlin_undo_context_add (ctxt, u);
	}
		
	g_object_get (G_OBJECT (window->priv->sample),
		      "total_frames", &total,
		      NULL);
	marlin_sample_selection_add (window->priv->selection,
				     (guint64) 0, total,
				     MARLIN_COVERAGE_BOTH);
}

static void
select_none_redo (gpointer data)
{
	struct _select_closure *c = data;

	marlin_window_select_none (c->window, NULL);
}

void
marlin_window_select_none (MarlinWindow *window,
			   MarlinUndoContext *ctxt)
{
	if (ctxt) {
		struct _select_closure *c;
		MarlinUndoable *u;
		MarlinSampleSubSelection *sub;

		READ_LOCK (window->priv->selection->lock);
		sub = window->priv->selection->selections->data;

		c = g_new0 (struct _select_closure, 1);
		c->window = window;

		if (sub) {
			c->start = sub->start;
			c->end = sub->end;
			c->coverage = sub->coverage;
		} else {
			c->coverage = MARLIN_COVERAGE_BOTH;
		}

		READ_UNLOCK (window->priv->selection->lock);

		u = marlin_undoable_new (select_undo,
					 select_none_redo,
					 select_destroy,
					 c);
		marlin_undo_context_add (ctxt, u);
	}

	marlin_sample_selection_clear (window->priv->selection);
}

void
marlin_window_zoom_in (MarlinWindow *window)
{
	guint fpp;

	g_object_get (G_OBJECT (window->priv->view),
		      "frames_per_pixel", &fpp,
		      NULL);

	fpp /= 2;

	if (fpp < 1) {
		fpp = 1;
	}

	g_object_set (G_OBJECT (window->priv->view),
		      "frames_per_pixel", fpp,
		      NULL);
	
	marlin_window_set_view_sensitive (window, TRUE);
}

void
marlin_window_zoom_out (MarlinWindow *window)
{
	guint fpp;

	g_object_get (G_OBJECT (window->priv->view),
		      "frames_per_pixel", &fpp,
		      NULL);

	fpp *= 2;

	if (fpp > 16384) {
		fpp = 16384;
	}

	g_object_set (G_OBJECT (window->priv->view),
		      "frames_per_pixel", fpp,
		      NULL);

	marlin_window_set_view_sensitive (window, TRUE);
}

void
marlin_window_show_all (MarlinWindow *window)
{
	guint fpp;
	guint64 total;

	g_object_get (G_OBJECT (window->priv->sample),
		      "total_frames", &total,
		      NULL);

	fpp = (total / GTK_WIDGET (window->priv->view)->allocation.width);
	if (total % GTK_WIDGET (window->priv->view)->allocation.width != 0) {
		fpp++;
	}

	g_object_set (G_OBJECT (window->priv->view),
		      "frames_per_pixel", fpp,
		      NULL);
	marlin_sample_view_scroll_to (window->priv->view, (guint64) 0);
}

void
marlin_window_zoom_selection (MarlinWindow *window)
{
	guint fpp;
	guint64 length, start;
	MarlinWindowPrivate *priv;
	MarlinSampleSubSelection *sub;

	priv = window->priv;

	READ_LOCK (priv->selection->lock);
	if (priv->selection->selections == NULL) {
		READ_UNLOCK (priv->selection->lock);
		return;
	}

	sub = priv->selection->selections->data;
	length = sub->end - sub->start;
	start = sub->start;

	READ_UNLOCK (priv->selection->lock);

	fpp = (length / GTK_WIDGET (priv->view)->allocation.width);
	if (length % GTK_WIDGET (priv->view)->allocation.width != 0) {
		fpp++;
	}

	g_object_set (G_OBJECT (window->priv->view),
		      "frames_per_pixel", fpp,
		      NULL);
	marlin_sample_view_scroll_to (priv->view, start);
}

void
marlin_window_delete (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	MarlinSampleSubSelection *sub;
	MarlinUndoContext *ctxt;
	
	priv = window->priv;
	
	READ_LOCK (priv->selection->lock);
	
	if (priv->selection->selections == NULL) {
		READ_UNLOCK (priv->selection->lock);
		return;
	}
	
	sub = priv->selection->selections->data;

	/* Check we're only deleting on BOTH */
	if (sub->coverage != MARLIN_COVERAGE_BOTH) {
		GtkWidget *dialog;

		dialog = gtk_message_dialog_new (GTK_WINDOW (window),
						 GTK_DIALOG_DESTROY_WITH_PARENT |
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR,
						 GTK_BUTTONS_OK,
						 _("The delete operation can only be performed on both channels of a stereo sample."));
		g_signal_connect (G_OBJECT (dialog), "response",
				  G_CALLBACK (gtk_widget_destroy), NULL);
		gtk_widget_show (dialog);

		READ_UNLOCK (priv->selection->lock);
		return;
	}
	
	ctxt = marlin_undo_manager_context_begin (priv->undo, _("Delete Selection"));
	marlin_sample_delete_range (priv->sample, NULL, sub->start, sub->end, ctxt, NULL);
	marlin_undo_manager_context_end (priv->undo, ctxt);

	READ_UNLOCK (priv->selection->lock);
	
	/* Clear the selection */
	marlin_sample_selection_clear (priv->selection);
}

void
marlin_window_clear (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	MarlinSampleSubSelection *sub;
	MarlinUndoContext *ctxt;

	priv = window->priv;

	READ_LOCK (priv->selection->lock);
	if (priv->selection->selections == NULL) {
		guint64 frames;

		READ_UNLOCK (priv->selection->lock);

		/* Clear the whole sample */
		g_object_get (G_OBJECT (priv->sample),
			      "total_frames", &frames,
			      NULL);
		
		ctxt = marlin_undo_manager_context_begin (priv->undo,
							  _("Clear"));
		marlin_sample_clear_range (priv->sample, NULL,
					   (guint64) 0, frames - 1,
					   MARLIN_COVERAGE_BOTH, ctxt, NULL);
		marlin_undo_manager_context_end (priv->undo, ctxt);
		return;
	}

	sub = priv->selection->selections->data;

	ctxt = marlin_undo_manager_context_begin (priv->undo, _("Clear Selection"));
	marlin_sample_clear_range (priv->sample, NULL, sub->start, sub->end, sub->coverage, ctxt, NULL);
	marlin_undo_manager_context_end (priv->undo, ctxt);
	READ_UNLOCK (priv->selection->lock);

	/* Clear the selection */
	marlin_sample_selection_clear (priv->selection);
}

void
marlin_window_invert (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	MarlinSampleSubSelection *sub;
	MarlinUndoContext *ctxt;
	gboolean ret;
	GError *error = NULL;

	priv = window->priv;

	READ_LOCK (priv->selection->lock);
	if (priv->selection->selections == NULL) {
		guint64 frames;

		READ_UNLOCK (priv->selection->lock);

		/* Invert whole sample */
		g_object_get (G_OBJECT (priv->sample),
			      "total_frames", &frames,
			      NULL);

		ctxt = marlin_undo_manager_context_begin (priv->undo, _("Invert"));
		ret = marlin_sample_invert_range (priv->sample, (guint64) 0,
						  frames - 1, 
						  MARLIN_COVERAGE_BOTH, 
						  ctxt,
						  &error);
		if (ret == FALSE) {
			g_warning ("marlin_sample_invert_range failed");
		}
		marlin_undo_manager_context_end (priv->undo, ctxt);

		return;
	}

	sub = priv->selection->selections->data;

	ctxt = marlin_undo_manager_context_begin (priv->undo, _("Invert Selection"));
	ret = marlin_sample_invert_range (priv->sample, sub->start, sub->end, sub->coverage, ctxt, &error);
	if (ret == FALSE) {
		g_warning ("marlin_sample_invert_range failed");
	}
	marlin_undo_manager_context_end (priv->undo, ctxt);

	READ_UNLOCK (priv->selection->lock);
}

void
marlin_window_crop (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	MarlinSampleSubSelection *sub;
	MarlinUndoContext *ctxt;
	GError *error = NULL;
	gboolean ret;

	priv = window->priv;

	READ_LOCK (priv->selection->lock);
	if (priv->selection->selections == NULL) {
		READ_UNLOCK (priv->selection->lock);
		return;
	}

	sub = priv->selection->selections->data;

	/* Check we're only cropping on BOTH */
	/* FIXME: For discussion - Should cropping a single channel on a
	   stereo sample make a mono sample as the result? */
	if (sub->coverage != MARLIN_COVERAGE_BOTH) {
		GtkWidget *dialog;

		dialog = gtk_message_dialog_new (GTK_WINDOW (window),
						 GTK_DIALOG_DESTROY_WITH_PARENT |
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR,
						 GTK_BUTTONS_OK,
						 _("The crop operation can only be performed on both channels of a stereo sample."));
		g_signal_connect (G_OBJECT (dialog), "response",
				  G_CALLBACK (gtk_widget_destroy), NULL);
		gtk_widget_show (dialog);

		READ_UNLOCK (priv->selection->lock);
		return;
	}

	ctxt = marlin_undo_manager_context_begin (priv->undo, _("Crop Selection"));
	ret = marlin_sample_crop_range (priv->sample, NULL, sub->start, sub->end, ctxt, &error);
	if (ret == FALSE) {
		g_warning ("marlin_sample_crop_range failed");
	}
	marlin_undo_manager_context_end (priv->undo, ctxt);

	READ_UNLOCK (priv->selection->lock);

	/* Clear the selection */
	marlin_sample_selection_clear (priv->selection);
}

void
marlin_window_paste (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	MarlinCoverage coverage;
	MarlinProgram *program = marlin_program_get_default ();
	MarlinSample *clipboard;
	MarlinUndoContext *ctxt;
	guint64 position;
	gboolean ret;
	GError *error = NULL;
	
	priv = window->priv;

	g_object_get (G_OBJECT (priv->view),
		      "cursor_coverage", &coverage,
		      "cursor_position", &position,
		      NULL);

	clipboard = MARLIN_SAMPLE (marlin_program_get_clipboard (program));

	ctxt = marlin_undo_manager_context_begin (priv->undo, _("Paste"));
	ret = marlin_sample_insert (priv->sample, clipboard, position, coverage, ctxt, &error);
	if (ret == FALSE) {
		g_warning ("marlin_sample_insert failed");
	}
	marlin_undo_manager_context_end (priv->undo, ctxt);
}

struct _mixer_data {
	GtkAdjustment *src_level, *dest_level;
	MarlinSample *src, *dest;
	MarlinWindow *window;
};

static void
mixer_response_cb (GtkDialog *mixer,
		   guint response_id,
		   struct _mixer_data *md)
{
	MarlinProgram *program = marlin_program_get_default ();
	MarlinUndoContext *ctxt;
	MarlinSample *clipboard, *dest;
	MarlinWindowPrivate *priv;
	guint64 cursor_pos, end;
	gboolean clip = FALSE;
	
	priv = md->window->priv;
	switch (response_id) {
	case GTK_RESPONSE_CANCEL:
		break;

	case GTK_RESPONSE_OK:
		clipboard = MARLIN_SAMPLE (marlin_program_get_clipboard (program));
		dest = md->window->priv->sample;

		g_object_get (G_OBJECT (md->window->priv->view),
			      "cursor_position", &cursor_pos,
			      NULL);

		
		READ_LOCK (md->window->priv->selection->lock);
		if (md->window->priv->selection->selections != NULL) {
			MarlinSampleSubSelection *sub;
			
			sub = md->window->priv->selection->selections->data;
			end = sub->end;
			clip = TRUE;
			ctxt = marlin_undo_manager_context_begin (priv->undo,
								  _("Mix Into"));
		} else {
			ctxt = marlin_undo_manager_context_begin (priv->undo,
								  _("Mix"));
			end = (guint64) 0;
			clip = FALSE;
		}
		READ_UNLOCK (md->window->priv->selection->lock);
		
		marlin_sample_mix (dest, clipboard, md->src_level->value,
				   md->dest_level->value, cursor_pos,
				   end, clip,
				   NULL, ctxt, NULL);
		marlin_undo_manager_context_end (priv->undo, ctxt);
		break;

	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-paste-mix-dialog");
		return;

	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (mixer));
	g_free (md);
}

#define PACK_FULL(t, w, l, top, x, y) gtk_table_attach (GTK_TABLE ((t)), (w), \
                                             (l), (l) + 1, \
                                             (top), top + 1, \
                                             (x), (y), 0, 0)

void
marlin_window_mix (MarlinWindow *window)
{
	MarlinProgram *program = marlin_program_get_default ();
	struct _mixer_data *md;
	GtkWidget *mixer, *vbox, *table;
	GtkWidget *label, *s_scale, *d_scale;
	char *name, *title;
	
	md = g_new (struct _mixer_data, 1);
	md->src = MARLIN_SAMPLE (marlin_program_get_clipboard (program));
	md->dest = window->priv->sample;
	md->window = window;

	g_object_get (G_OBJECT (window->priv->sample),
		      "name", &name,
		      NULL);
	title = g_strdup_printf (_("Mix Clipboard: %s"), name);
	g_free (name);
	
	mixer = gtk_dialog_new ();
	gtk_window_set_transient_for (GTK_WINDOW (mixer), GTK_WINDOW (window));
	gtk_window_set_title (GTK_WINDOW (mixer), title);
	g_free (title);
	
	gtk_dialog_set_has_separator (GTK_DIALOG (mixer), FALSE);
 	gtk_window_set_default_size (GTK_WINDOW (mixer), 325, 270);
/* 	gtk_window_set_resizable (GTK_WINDOW (mixer), FALSE); */

	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (mixer)->vbox), vbox);

	table = gtk_table_new (2, 2, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE (table), 12);
	gtk_table_set_row_spacings (GTK_TABLE (table), 12);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);

	md->src_level = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, -43.1, 15.1, 0.1, 10.0, 0.1));
	md->dest_level = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, -43.1, 15.1, 0.1, 10.0, 0.1));

	s_scale = gtk_vscale_new (md->src_level);
	g_signal_connect (G_OBJECT (s_scale), "format_value",
			  G_CALLBACK (marlin_set_volume_digits), NULL);
	gtk_range_set_inverted (GTK_RANGE (s_scale), TRUE);
	gtk_widget_show (s_scale);
	PACK_FULL (table, s_scale, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND);
	
	label = gtk_label_new_with_mnemonic (_("_Clipboard Level:"));
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), s_scale);
	gtk_widget_show (label);
	PACK_FULL (table, label, 0, 0, GTK_FILL | GTK_EXPAND, GTK_FILL);

	marlin_add_paired_relations (s_scale, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);
	
	d_scale = gtk_vscale_new (md->dest_level);
	g_signal_connect (G_OBJECT (d_scale), "format-value",
			  G_CALLBACK (marlin_set_volume_digits), NULL);
	gtk_range_set_inverted (GTK_RANGE (d_scale), TRUE);
	gtk_widget_show (d_scale);
	PACK_FULL (table, d_scale, 1, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND);
	
	label = gtk_label_new_with_mnemonic (_("_Sample Level:"));
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), d_scale);
	gtk_widget_show (label);
	PACK_FULL (table, label, 1, 0, GTK_FILL | GTK_EXPAND, GTK_FILL);

	marlin_add_paired_relations (d_scale, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);
	
	gtk_dialog_add_button (GTK_DIALOG (mixer),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (mixer),
			       _("Mix"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (mixer),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (mixer), GTK_RESPONSE_OK);

	g_signal_connect (mixer, "response",
			  G_CALLBACK (mixer_response_cb), md);
	gtk_widget_show (mixer);
}

void
marlin_window_selection_to_clipboard (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	MarlinSampleSubSelection *sub;
	MarlinProgram *program = marlin_program_get_default ();
	MarlinSample *clipboard;
	GError *error = NULL;

	priv = window->priv;

	READ_LOCK (priv->selection->lock);
	if (priv->selection->selections == NULL) {
		READ_UNLOCK (priv->selection->lock);
		return;
	}

	sub = priv->selection->selections->data;
	g_print ("selection end: %llu\n", sub->end);
	clipboard = marlin_sample_new_from_sample_with_range (priv->sample,
							      sub->start,
							      sub->end,
							      sub->coverage,
							      &error);
	READ_UNLOCK (priv->selection->lock);
	if (clipboard == NULL) {
		g_warning ("marlin_sample_new_from_sample_with_range failed");
		return;
	}

	marlin_program_set_clipboard (program, G_OBJECT (clipboard));
	g_object_unref (clipboard);
}

GtkUIManager *
marlin_window_get_menu_merge (MarlinWindow *window)
{
	return GTK_UI_MANAGER (window->priv->ui_merge);
}

Toolbar *
marlin_window_get_toolbar (MarlinWindow *window)
{
	return window->priv->toolbar;
}

MarlinSampleView *
marlin_window_get_view (MarlinWindow *window)
{
	return window->priv->view;
}

/* Utility function that takes a combo box and a list of markers
   and fills the combo with the names of the markers */
static void
make_marker_combo_menu (GtkComboBox *combo,
			GList *markers)
{
	GList *m;

	for (m = markers; m; m = m->next) {
		MarlinMarker *marker = m->data;

		gtk_combo_box_append_text (combo, marker->name);
	}

	gtk_combo_box_set_active (combo, 0);
}

struct _select_region_data {
	MarlinWindow *window;
	MarlinSampleSelection *selection;
	GtkWidget *start, *finish;
	GtkWidget *start_combo, *finish_combo;
	GtkWidget *sp_hbox, *sm_hbox, *fp_hbox, *fm_hbox;
	GtkWidget *sm_combo, *fm_combo;
	
	GList *markers;
};

enum {
	SELECT_START_START,
	SELECT_START_CURSOR,
	SELECT_START_CUSTOM,
	SELECT_START_MARKER,
	SELECT_START_0
};

enum {
	SELECT_END_END,
	SELECT_END_CURSOR,
	SELECT_END_CUSTOM,
	SELECT_END_MARKER,
	SELECT_END_0
};

static void
dialog_response_cb (GtkDialog *dialog,
		    guint response_id,
		    struct _select_region_data *srd)
{
	int start_active, end_active;
	guint64 start, finish;
	MarlinPositionSpinnerDisplay display;
	MarlinMarker *marker;
	int fpp;
	
	switch (response_id) {
	case GTK_RESPONSE_OK:
		start_active = gtk_combo_box_get_active (GTK_COMBO_BOX (srd->start_combo));
		end_active = gtk_combo_box_get_active (GTK_COMBO_BOX (srd->finish_combo));

		switch (start_active) {
		case SELECT_START_START:
			start = (guint64) 0;
			break;

		case SELECT_START_CURSOR:
			g_object_get (G_OBJECT (srd->window->priv->view),
				      "cursor_position", &start,
				      NULL);
			break;
			
		case SELECT_START_CUSTOM:
			start = (guint64) GTK_SPIN_BUTTON (srd->start)->adjustment->value;
			g_object_get (G_OBJECT (srd->start),
				      "display_as", &display,
				      NULL);
			marlin_gconf_set_int ("/apps/marlin/system-state/select-range-start-display", display);
			
			break;

		case SELECT_START_MARKER:
			marker = g_list_nth_data (srd->markers, gtk_combo_box_get_active (GTK_COMBO_BOX (srd->sm_combo)));
			g_assert (marker != NULL);

			start = marker->position;
			break;

		default:
			g_assert_not_reached ();
		}

		switch (end_active) {
		case SELECT_END_END:
			g_object_get (G_OBJECT (srd->window->priv->sample),
				      "total_frames", &finish,
				      NULL);
			g_object_get (G_OBJECT (srd->window->priv->view),
				      "frames_per_pixel", &fpp,
				      NULL);

			finish -= fpp;
			break;

		case SELECT_END_CURSOR:
			g_object_get (G_OBJECT (srd->window->priv->view),
				      "cursor_position", &finish,
				      NULL);
			break;
			
		case SELECT_END_CUSTOM:
			finish = GTK_SPIN_BUTTON (srd->finish)->adjustment->value;

			g_object_get (G_OBJECT (srd->finish),
				      "display_as", &display,
				      NULL);
			marlin_gconf_set_int ("/apps/marlin/system-state/select-range-finish-display", display);
			break;

		case SELECT_END_MARKER:
			marker = g_list_nth_data (srd->markers, gtk_combo_box_get_active (GTK_COMBO_BOX (srd->fm_combo)));
			g_assert (marker != NULL);

			finish = marker->position;
			break;

		default:
			g_assert_not_reached ();
		}
		
		marlin_sample_selection_clear (srd->selection);
		marlin_sample_selection_add (srd->selection, start, finish,
					     MARLIN_COVERAGE_BOTH);
		break;
		
	case GTK_RESPONSE_CANCEL:
		break;

	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-window-select-region-dialog");
		return;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (srd);
}

#define PACK(t, w, l, top, x) gtk_table_attach (GTK_TABLE ((t)), (w), \
					     (l), (l) + 1, \
					     (top), top + 1, \
					     (x), GTK_FILL, 0, 0)

static void
select_start_changed (GtkComboBox *combo,
		      struct _select_region_data *srd)
{
	int active = gtk_combo_box_get_active (combo);

	gtk_widget_set_sensitive (srd->sm_hbox, (active == SELECT_START_MARKER));
	gtk_widget_set_sensitive (srd->sp_hbox, (active == SELECT_START_CUSTOM));
	if (active == SELECT_START_CUSTOM) {
		gtk_widget_show (srd->sp_hbox);
		gtk_widget_hide (srd->sm_hbox);
	} else if (active == SELECT_START_MARKER) {
		gtk_widget_show (srd->sm_hbox);
		gtk_widget_hide (srd->sp_hbox);
	}
}

static void
select_finish_changed (GtkComboBox *combo,
		       struct _select_region_data *srd)
{
	int active = gtk_combo_box_get_active (combo);

	gtk_widget_set_sensitive (srd->fm_hbox, (active == SELECT_END_MARKER));
	gtk_widget_set_sensitive (srd->fp_hbox, (active == SELECT_END_CUSTOM));
	if (active == SELECT_END_CUSTOM) {
		gtk_widget_show (srd->fp_hbox);
		gtk_widget_hide (srd->fm_hbox);
	} else if (active == SELECT_END_MARKER) {
		gtk_widget_show (srd->fm_hbox);
		gtk_widget_hide (srd->fp_hbox);
	}
}

static void
make_start_combo_menu (GtkComboBox *combo,
		       struct _select_region_data *srd)
{
	int i;
	static char *positions[] = {
		N_("Start of Sample"),
		N_("Cursor Position"),
		N_("Custom Position"),
		N_("Marker"),
		NULL
	};

	for (i = 0; positions[i] != NULL; i++) {
		if (i != SELECT_START_MARKER || srd->markers != NULL) {
			gtk_combo_box_append_text (combo, _(positions[i]));
		}
	}

	gtk_combo_box_set_active (combo, 0);
	gtk_widget_set_sensitive (srd->sm_hbox, FALSE);
	gtk_widget_set_sensitive (srd->sp_hbox, FALSE);
	gtk_widget_show (srd->sp_hbox);
	gtk_widget_hide (srd->sm_hbox);
}

static void
make_finish_combo_menu (GtkComboBox *combo,
			struct _select_region_data *srd)
{
	int i;
	static char *positions[] = {
		N_("End of Sample"),
		N_("Cursor Position"),
		N_("Custom Position"),
		N_("Marker"),
		NULL
	};

	for (i = 0; positions[i] != NULL; i++) {
		if (i != SELECT_END_MARKER || srd->markers != NULL) {
			gtk_combo_box_append_text (combo, _(positions[i]));
		}
	}

	gtk_combo_box_set_active (combo, 0);
	gtk_widget_set_sensitive (srd->fm_hbox, FALSE);
	gtk_widget_set_sensitive (srd->fp_hbox, FALSE);
	gtk_widget_show (srd->fp_hbox);
	gtk_widget_hide (srd->fm_hbox);
}

void
marlin_window_select_region (MarlinWindow *window)
{
	struct _select_region_data *srd;
	GtkWidget *editor;
	GtkWidget *vbox, *table, *hbox;
	GtkWidget *label;
	GtkAdjustment *adj;
	MarlinSampleSubSelection *sub;
	MarlinPositionSpinnerDisplay display;
	MarlinMarkerModel *mm;
	guint64 total;
	guint rate;
	char *title, *name;

	g_return_if_fail (IS_MARLIN_WINDOW (window));
	g_return_if_fail (window->priv->selection != NULL);
	
	if (window->priv->select_region_dialog != NULL) {
		gdk_window_show (window->priv->select_region_dialog->window);
		return;
	}
	
	srd = g_new (struct _select_region_data, 1);
	srd->window = window;
	srd->selection = window->priv->selection;
	sub = srd->selection->selections ? srd->selection->selections->data : NULL;

	g_object_get (G_OBJECT (window->priv->sample),
		      "total_frames", &total,
		      "sample_rate", &rate,
		      "name", &name,
		      "markers", &mm,
		      NULL);

	g_object_get (G_OBJECT (mm),
		      "markers", &srd->markers,
		      NULL);
	
	editor = window->priv->select_region_dialog = gtk_dialog_new ();
	gtk_dialog_set_has_separator (GTK_DIALOG (editor), FALSE);
/* 	gtk_window_set_default_size (GTK_WINDOW (editor), 325, 128); */
	gtk_window_set_resizable (GTK_WINDOW (editor), FALSE);
	gtk_window_set_transient_for (GTK_WINDOW (editor), GTK_WINDOW (window));

	title = g_strdup_printf (_("Select a Region: %s"), name);
	g_free (name);
	gtk_window_set_title (GTK_WINDOW (editor), title);
	g_free (title);

	vbox = gtk_vbox_new (FALSE, 12);
 	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (editor)->vbox), vbox);

	table = gtk_table_new (2, 4, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
	
	label = gtk_label_new_with_mnemonic (_("Selection _Start:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	gtk_widget_show (label);
	PACK (table, label, 0, 0, GTK_FILL);

	srd->start_combo = gtk_combo_box_new_text ();
	marlin_add_paired_relations (srd->start_combo, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);
	g_signal_connect (G_OBJECT (srd->start_combo), "changed",
			  G_CALLBACK (select_start_changed), srd);

	gtk_widget_show (srd->start_combo);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), srd->start_combo);

	gtk_table_attach (GTK_TABLE (table), srd->start_combo,
			  1, 3, 0, 1,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox);
	gtk_table_attach (GTK_TABLE (table), hbox,
			  1, 3, 1, 2,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0 ,0);

	srd->sp_hbox = gtk_hbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox), srd->sp_hbox, TRUE, TRUE, 0);
	gtk_widget_show (srd->sp_hbox);

	label = gtk_label_new_with_mnemonic (_("_Position:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (srd->sp_hbox), label, FALSE, FALSE, 0);
	
	display = marlin_gconf_get_int ("/apps/marlin/system-state/select-range-start-display");
	srd->start = marlin_position_spinner_new ();
	g_object_set (G_OBJECT (srd->start),
		      "max_frames", total,
		      "rate", rate,
		      "display_as", display,
		      NULL);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (srd->start), (sub ? (double) sub->start : 0.0));

	gtk_entry_set_activates_default (GTK_ENTRY (srd->start), TRUE);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), srd->start);
	gtk_widget_show (srd->start);
	gtk_box_pack_start (GTK_BOX (srd->sp_hbox), srd->start, TRUE, TRUE, 0);

	marlin_add_paired_relations (srd->start, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);

	label = marlin_position_spinner_label (MARLIN_POSITION_SPINNER (srd->start));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (srd->sp_hbox), label, FALSE, FALSE, 0);

	srd->sm_hbox = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (srd->sm_hbox);
	gtk_box_pack_start (GTK_BOX (hbox), srd->sm_hbox, TRUE, TRUE, 0);

	label = gtk_label_new_with_mnemonic (_("S_tart Marker:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (srd->sm_hbox), label, FALSE, FALSE, 0);

	srd->sm_combo = gtk_combo_box_new_text ();
	make_marker_combo_menu (GTK_COMBO_BOX (srd->sm_combo), srd->markers);
	gtk_widget_show (srd->sm_combo);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), srd->sm_combo);
	gtk_box_pack_start (GTK_BOX (srd->sm_hbox), srd->sm_combo, TRUE, TRUE, 0);

	if (srd->markers == NULL) {
		gtk_widget_hide (srd->sm_hbox);
	}

	make_start_combo_menu (GTK_COMBO_BOX (srd->start_combo), srd);
	
	label = gtk_label_new_with_mnemonic (_("Selection _End:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	gtk_widget_show (label);
	PACK (table, label, 0, 2, GTK_FILL);

	srd->finish_combo = gtk_combo_box_new_text ();
	marlin_add_paired_relations (srd->finish_combo, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);
	g_signal_connect (G_OBJECT (srd->finish_combo), "changed",
			  G_CALLBACK (select_finish_changed), srd);

	gtk_widget_show (srd->finish_combo);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), srd->finish_combo);

	gtk_table_attach (GTK_TABLE (table), srd->finish_combo,
			  1, 3, 2, 3,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox);
	gtk_table_attach (GTK_TABLE (table), hbox,
			  1, 3, 3, 4,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	srd->fp_hbox = gtk_hbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox), srd->fp_hbox, TRUE, TRUE, 0);
	gtk_widget_show (srd->fp_hbox);

	label = gtk_label_new_with_mnemonic (_("P_osition:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (srd->fp_hbox), label, FALSE, FALSE, 0);
	
	display = marlin_gconf_get_int ("/apps/marlin/system-state/select-range-finish-display");
	srd->finish = marlin_position_spinner_new ();
	g_object_set (G_OBJECT (srd->finish),
		      "max_frames", total,
		      "rate", rate,
		      "display_as", display,
		      NULL);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (srd->finish), (sub ? (double) sub->end : 0.0));

	gtk_entry_set_activates_default (GTK_ENTRY (srd->finish), TRUE);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), srd->finish);
	gtk_widget_show (srd->finish);
	gtk_box_pack_start (GTK_BOX (srd->fp_hbox), srd->finish, TRUE, TRUE, 0);

	marlin_add_paired_relations (srd->finish, ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);

	label = marlin_position_spinner_label (MARLIN_POSITION_SPINNER (srd->finish));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (srd->fp_hbox), label, FALSE, FALSE, 0);

	srd->fm_hbox = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (srd->fm_hbox);
	gtk_box_pack_start (GTK_BOX (hbox), srd->fm_hbox, TRUE, TRUE, 0);

	label = gtk_label_new_with_mnemonic (_("Start Mar_ker:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (srd->fm_hbox), label, FALSE, FALSE, 0);

	srd->fm_combo = gtk_combo_box_new_text ();
	make_marker_combo_menu (GTK_COMBO_BOX (srd->fm_combo), srd->markers);
	gtk_widget_show (srd->fm_combo);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), srd->fm_combo);
	gtk_box_pack_start (GTK_BOX (srd->fm_hbox), srd->fm_combo, TRUE, TRUE, 0);

	if (srd->markers == NULL) {
		gtk_widget_hide (srd->fm_hbox);
	}

	make_finish_combo_menu (GTK_COMBO_BOX (srd->finish_combo), srd);
	
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       _("Select _Region"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (editor), GTK_RESPONSE_OK);
	
	g_signal_connect (editor, "response",
			  G_CALLBACK (dialog_response_cb), srd);
	g_signal_connect (editor, "destroy",
			  G_CALLBACK (gtk_widget_destroyed), &window->priv->select_region_dialog);

	gtk_widget_show (editor);
}

struct _AdjustVolData {
	MarlinWindow *window;
	GtkAdjustment *adj;
};

void
adjust_volume_response (GtkDialog *dialog,
			guint response_id,
			struct _AdjustVolData *avd)
{
	MarlinWindow *window = avd->window;
	MarlinWindowPrivate *priv;
	MarlinSampleSubSelection *sub;
	MarlinUndoContext *ctxt;
	gboolean ret;
	GError *error = NULL;

	priv = window->priv;

	switch (response_id) {
	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-process-adjust-volume-dialog");
		return;

	case GTK_RESPONSE_OK:
		/* Store the db value for next time */
		marlin_gconf_set_float ("/apps/marlin/system-state/adjust-volume-db", avd->adj->value);
		
		READ_LOCK (priv->selection->lock);

		if (priv->selection->selections == NULL) {
			guint64 frames;
			
			/* Adjust whole sample */
			g_object_get (G_OBJECT (priv->sample),
				      "total_frames", &frames,
				      NULL);
			ctxt = marlin_undo_manager_context_begin (priv->undo,
								  _("Adjust Sample Volume"));

			ret = marlin_sample_adjust_volume_range (priv->sample,
								 avd->adj->value,
								 (guint64) 0, 
								 frames - 1,
								 MARLIN_COVERAGE_BOTH,
								 ctxt,
								 &error);
		} else {
			sub = priv->selection->selections->data;

			ctxt = marlin_undo_manager_context_begin (priv->undo,
								  _("Adjust Selection Volume"));

			ret = marlin_sample_adjust_volume_range (priv->sample,
								 avd->adj->value,
								 sub->start,
								 sub->end,
								 sub->coverage,
								 ctxt,
								 &error);
		}
		READ_UNLOCK (priv->selection->lock);

		if (ret == FALSE) {
			g_warning ("marlin_sample_adjust_volume_range failed");
		}

		marlin_undo_manager_context_end (priv->undo, ctxt);
		break;

	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (avd);
}

void
marlin_window_adjust_volume (MarlinWindow *window)
{
	GtkWidget *editor, *vbox, *label, *scale;
	GtkWidget *hbox, *inner_vbox;
	struct _AdjustVolData *avd;
	char *title, *name;
	float db;
	
	avd = g_new (struct _AdjustVolData, 1);
	avd->window = window;

	editor = gtk_dialog_new ();
	gtk_dialog_set_has_separator (GTK_DIALOG (editor), FALSE);
/* 	gtk_window_set_default_size (GTK_WINDOW (editor), 360, 130); */
	gtk_window_set_resizable (GTK_WINDOW (editor), FALSE);
	gtk_window_set_transient_for (GTK_WINDOW (editor), GTK_WINDOW (window));

	g_object_get (G_OBJECT (window->priv->sample),
		      "name", &name,
		      NULL);

	title = g_strdup_printf (_("Adjust Volume: %s"), name);
	g_free (name);
	
	gtk_window_set_title (GTK_WINDOW (editor), title);
	g_free (title);

	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (editor)->vbox), vbox);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	inner_vbox = gtk_vbox_new (FALSE, 6);
	gtk_widget_show (inner_vbox);
	gtk_box_pack_start (GTK_BOX (hbox), inner_vbox, TRUE, TRUE, 0);

	avd->adj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, -43.1, 15.1, 0.1, 10.0, 0.1));
	db = marlin_gconf_get_float ("/apps/marlin/system-state/adjust-volume-db");
	gtk_adjustment_set_value (avd->adj, db);
	
	scale = gtk_hscale_new (avd->adj);
	
	g_signal_connect (G_OBJECT (scale), "format_value",
			  G_CALLBACK (marlin_set_volume_digits), NULL);
	gtk_range_set_inverted (GTK_RANGE (scale), TRUE);
	gtk_widget_show (scale);
	gtk_box_pack_start (GTK_BOX (inner_vbox), scale, TRUE, TRUE, 0);

	label = gtk_label_new (_("-Infinity dB -> +15.0 dB"));
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (inner_vbox), label, FALSE, FALSE, 0);
	
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       _("_Adjust"), GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_HELP, GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (editor), GTK_RESPONSE_OK);

	g_signal_connect (G_OBJECT (editor), "response",
			  G_CALLBACK (adjust_volume_response), avd);
	gtk_widget_show (editor);
}

struct _SilenceData {
	MarlinWindow *window;
	GtkWidget *length;
	GtkWidget *position;
	GtkWidget *samples;
	GtkWidget *units;
	GtkWidget *marker_hbox;
	GtkWidget *markers;

	MarlinMarkerModel *marker_model;
	GList *marker_list;
};

enum {
	SILENCE_CURSOR_POSITION,
	SILENCE_START,
	SILENCE_END,
	SILENCE_MARKER,
	SILENCE_0
};

static void
insert_silence_response (GtkDialog *dialog,
			 guint response_id,
			 struct _SilenceData *sd)
{
	GtkAdjustment *adj;
	MarlinMarker *marker;
	int history, marker_no;
	guint64 position, num_frames;
	MarlinPositionSpinnerDisplay display;
	MarlinUndoContext *ctxt;
	gboolean ret;
	GError *error = NULL;
	
	switch (response_id) {
	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-process-insert-silence-dialog");
		return;

	case GTK_RESPONSE_OK:
		history = gtk_combo_box_get_active (GTK_COMBO_BOX (sd->position));
		marlin_gconf_set_int ("/apps/marlin/system-state/insert-silence-history", history);

		g_object_get (G_OBJECT (sd->length),
			      "display_as", &display,
			      NULL);
		marlin_gconf_set_int ("/apps/marlin/system-state/insert-silence-display", display);

		switch (history) {
		case SILENCE_CURSOR_POSITION:
			g_object_get (G_OBJECT (sd->window->priv->view),
				      "cursor_position", &position,
				      NULL);
			break;

		case SILENCE_START:
			position = (guint64) 0;
			break;

		case SILENCE_END:
			g_object_get (G_OBJECT (sd->window->priv->sample),
				      "total_frames", &position,
				      NULL);
			break;

		case SILENCE_MARKER:
			marker_no = gtk_combo_box_get_active (GTK_COMBO_BOX (sd->markers));
			marker = g_list_nth_data (sd->marker_list, marker_no);
			g_assert (marker != NULL);
			position = marker->position;
			break;
			
		default:
			g_assert_not_reached ();
			break;
		}

		g_object_get (G_OBJECT (sd->length),
			      "adjustment", &adj,
			      NULL);
		num_frames = (guint64) adj->value;

		/* Can't insert 0 frames */
		if (num_frames == 0) {
			break;
		}

		ctxt = marlin_undo_manager_context_begin (sd->window->priv->undo,
							  _("Insert Silence"));
		ret = marlin_sample_insert_silence (sd->window->priv->sample, position, num_frames, ctxt, &error);
		if (ret == FALSE) {
			g_warning ("marlin_sample_insert_silence failed");
		}
		marlin_undo_manager_context_end (sd->window->priv->undo, ctxt);
		break;

	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (sd);
}

static void
make_insert_silence_menu (GtkComboBox *combo,
			  struct _SilenceData *sd)
{
	int i, history;
	static char *positions[] = {
		N_("Cursor Position"),
		N_("Start of Sample"),
		N_("End of Sample"),
		N_("At Marker"),
		NULL
	};

	for (i = 0; positions[i] != NULL; i++) {
		if (i != SILENCE_MARKER || sd->marker_list != NULL) {
			gtk_combo_box_append_text (combo, _(positions[i]));
		}
	}
	history = marlin_gconf_get_int ("/apps/marlin/system-state/insert-silence-history");

	gtk_combo_box_set_active (combo, history);
	gtk_widget_set_sensitive (sd->marker_hbox, (history == SILENCE_MARKER));
}

static void
position_combo_changed (GtkComboBox *combo,
			struct _SilenceData *sd)
{
	int active = gtk_combo_box_get_active (combo);

	gtk_widget_set_sensitive (sd->marker_hbox, (active == SILENCE_MARKER));
}

void
marlin_window_insert_silence (MarlinWindow *window)
{
	GtkWidget *editor, *vbox, *label;
	GtkWidget *table, *spacer;
	struct _SilenceData *sd;
	char *title, *name;
	guint rate, display;
	
	sd = g_new (struct _SilenceData, 1);
	sd->window = window;

	editor = gtk_dialog_new ();
	gtk_dialog_set_has_separator (GTK_DIALOG (editor), FALSE);
	gtk_window_set_transient_for (GTK_WINDOW (editor), GTK_WINDOW (window));
/* 	gtk_window_set_default_size (GTK_WINDOW (editor), 450, 110); */
	gtk_window_set_resizable (GTK_WINDOW (editor), FALSE);

	g_object_get (G_OBJECT (window->priv->sample),
		      "name", &name,
		      "sample_rate", &rate,
		      "markers", &sd->marker_model,
		      NULL);

	g_object_get (G_OBJECT (sd->marker_model),
		      "markers", &sd->marker_list,
		      NULL);
	
	title = g_strdup_printf (_("Insert Silence: %s"), name);
	g_free (name);
	
	gtk_window_set_title (GTK_WINDOW (editor), title);
	g_free (title);

	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (editor)->vbox), vbox);

	table = gtk_table_new (2, 3, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);

	label = gtk_label_new_with_mnemonic (_("_Duration:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	PACK (table, label, 0, 0, GTK_FILL);

	sd->length = marlin_position_spinner_new ();
	
	display = marlin_gconf_get_int ("/apps/marlin/system-state/insert-silence-display");
	g_object_set (G_OBJECT (sd->length),
		      "display_as", (MarlinPositionSpinnerDisplay) display,
		      "rate", rate,
		      NULL);
	
	gtk_widget_show (sd->length);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), sd->length);
	PACK (table, sd->length, 1, 0, GTK_FILL | GTK_EXPAND);

	sd->units = marlin_position_spinner_label (MARLIN_POSITION_SPINNER (sd->length));
	gtk_widget_show (sd->units);
	PACK (table, sd->units, 2, 0, GTK_FILL);

	label = gtk_label_new_with_mnemonic (_("Insert _Position:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	PACK (table, label, 0, 1, GTK_FILL);

	sd->position = gtk_combo_box_new_text ();
	g_signal_connect (G_OBJECT (sd->position), "changed",
			  G_CALLBACK (position_combo_changed), sd);

	gtk_widget_show (sd->position);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), sd->position);

	gtk_table_attach (GTK_TABLE (table), sd->position,
			  1, 3, 1, 2,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	sd->marker_hbox = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (sd->marker_hbox);
	gtk_table_attach (GTK_TABLE (table), sd->marker_hbox,
			  1, 3, 2, 3,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);
	
	label = gtk_label_new_with_mnemonic (_("_Marker:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (sd->marker_hbox), label, FALSE, FALSE, 0);

	sd->markers = gtk_combo_box_new_text ();
	make_marker_combo_menu (GTK_COMBO_BOX (sd->markers), sd->marker_list);
	
	gtk_widget_show (sd->markers);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), sd->markers);
	
	gtk_box_pack_start (GTK_BOX (sd->marker_hbox), sd->markers, TRUE, TRUE, 0);

	if (sd->marker_list == NULL) {
		gtk_widget_hide (sd->marker_hbox);
	}

	/* Do this here, so that the marker_hbox will exist */
	make_insert_silence_menu (GTK_COMBO_BOX (sd->position), sd);
	
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       _("_Insert Silence"), GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_HELP, GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (editor), GTK_RESPONSE_OK);

	g_signal_connect (G_OBJECT (editor), "response",
			  G_CALLBACK (insert_silence_response), sd);
	
	gtk_widget_show (editor);
}

struct _move_dialog_data {
	MarlinWindow *window;
	MarlinPositionSpinner *position;
	GtkWidget *combo;
	GtkWidget *marker_combo;
	GtkWidget *marker_hbox, *position_hbox;
	
	MarlinMarkerModel *marker_model;
	GList *markers;
};

enum {
	MOVE_START,
	MOVE_END,
	MOVE_CUSTOM,
	MOVE_MARKER,
	MOVE_0
};

static void
make_move_combo_menu (GtkComboBox *combo,
		      struct _move_dialog_data *mdd)
{
	int i, history;
	static char *positions[] = {
		N_("Start of Sample"),
		N_("End of Sample"),
		N_("Custom Position"),
		N_("Marker"),
		NULL
	};

	for (i = 0; positions[i] != NULL; i++) {
		if (i != MOVE_MARKER || mdd->markers != NULL) {
			gtk_combo_box_append_text (combo, _(positions[i]));
		}
	}

	history = marlin_gconf_get_int ("/apps/marlin/system-state/move-cursor-history");
	gtk_combo_box_set_active (combo, history);
	gtk_widget_set_sensitive (mdd->marker_hbox, (history == MOVE_MARKER));
	if (history == MOVE_MARKER) {
		gtk_widget_hide (mdd->position_hbox);
		gtk_widget_show (mdd->marker_hbox);
	} else {
		gtk_widget_hide (mdd->marker_hbox);
		gtk_widget_show (mdd->position_hbox);
	}
}

static void
move_cursor_response_cb (GtkDialog *dialog,
			 guint response_id,
			 struct _move_dialog_data *mdd)
{
	GtkAdjustment *adj;
	MarlinPositionSpinnerDisplay display;
	guint64 position;
	int marker_no, fpp;
	MarlinMarker *marker;
	
	switch (response_id) {
	case GTK_RESPONSE_OK:

		switch (gtk_combo_box_get_active (GTK_COMBO_BOX (mdd->combo))) {
		case MOVE_START:
			position = (guint64) 0;
			break;

		case MOVE_END:
			g_object_get (G_OBJECT (mdd->window->priv->view),
				      "frames_per_pixel", &fpp,
				      NULL);
			
			g_object_get (G_OBJECT (mdd->window->priv->sample),
				      "total_frames", &position,
				      NULL);

			position -= fpp;
			break;

		case MOVE_CUSTOM:
			g_object_get (G_OBJECT (mdd->position),
				      "display_as", &display,
				      NULL);
			marlin_gconf_set_int ("/apps/marlin/system-state/move-cursor-dialog-display", display);
			adj = GTK_SPIN_BUTTON (mdd->position)->adjustment;
			position = (guint64) adj->value;
			break;

		case MOVE_MARKER:
			marker_no = gtk_combo_box_get_active (GTK_COMBO_BOX (mdd->marker_combo));
			marker = g_list_nth_data (mdd->markers, marker_no);
			g_assert (marker != NULL);
			position = marker->position;
			break;

		default:
			g_assert_not_reached ();
		}
			
		g_object_set (G_OBJECT (mdd->window->priv->view),
			      "cursor-position", position,
			      NULL);
		
		break;

	case GTK_RESPONSE_CANCEL:
		break;

	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-move-cursor-dialog");
		return;

	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (mdd);
}

static void
move_combo_changed (GtkComboBox *combo,
		    struct _move_dialog_data *mdd)
{
	int active = gtk_combo_box_get_active (combo);

	gtk_widget_set_sensitive (mdd->marker_hbox, (active == MOVE_MARKER));
	gtk_widget_set_sensitive (mdd->position_hbox, (active == MOVE_CUSTOM));
	if (active == MOVE_CUSTOM) {
		gtk_widget_show (mdd->position_hbox);
		gtk_widget_hide (mdd->marker_hbox);
	} else if (active == MOVE_MARKER) {
		gtk_widget_show (mdd->marker_hbox);
		gtk_widget_hide (mdd->position_hbox);
	}
}

void
marlin_window_move_cursor_dialog (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	struct _move_dialog_data *mdd;
	GtkWidget *editor;
	GtkWidget *vbox, *table, *hbox, *label, *entry;
	GtkAdjustment *adj;
	guint64 position, num_frames;
	MarlinPositionSpinnerDisplay display;
	char *name, *title;
	guint rate;
	
	g_return_if_fail (IS_MARLIN_WINDOW (window));
	
	priv = window->priv;

	if (priv->move_cursor_dialog != NULL) {
		gtk_window_present (GTK_WINDOW (priv->move_cursor_dialog));
		return;
	}

	mdd = g_new (struct _move_dialog_data, 1);
	mdd->window = window;

	g_object_get (G_OBJECT (priv->sample),
		      "total_frames", &num_frames,
		      "sample_rate", &rate,
		      "name", &name,
		      "markers", &mdd->marker_model,
		      NULL);

	g_object_get (G_OBJECT (mdd->marker_model),
		      "markers", &mdd->markers,
		      NULL);
	
	editor = priv->move_cursor_dialog = gtk_dialog_new ();
	gtk_dialog_set_has_separator (GTK_DIALOG (editor), FALSE);
/* 	gtk_window_set_default_size (GTK_WINDOW (editor), 325, 95); */
	gtk_window_set_resizable (GTK_WINDOW (editor), FALSE);
	gtk_window_set_transient_for (GTK_WINDOW (editor), GTK_WINDOW (window));
	title = g_strdup_printf (_("Move Cursor: %s"), name);
	g_free (name);
	
	gtk_window_set_title (GTK_WINDOW (editor), title);
	g_free (title);

	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (editor)->vbox), vbox);

	table = gtk_table_new (2, 3, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
	
	label = gtk_label_new_with_mnemonic ("Move cursor _to:");
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	PACK (table, label, 0, 0, GTK_FILL);

	mdd->combo = gtk_combo_box_new_text ();
	/* FIXME: the combo is controller for other things too */
	marlin_add_paired_relations (GTK_WIDGET (mdd->combo), ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);
	
	g_signal_connect (G_OBJECT (mdd->combo), "changed",
			  G_CALLBACK (move_combo_changed), mdd);
	
	gtk_widget_show (mdd->combo);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), mdd->combo);
	
	gtk_table_attach (GTK_TABLE (table), mdd->combo,
			  1, 3, 0, 1,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	mdd->position_hbox = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (mdd->position_hbox);
	gtk_table_attach (GTK_TABLE (table), mdd->position_hbox,
			  1, 3, 1, 2,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	label = gtk_label_new_with_mnemonic (_("_Position:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (mdd->position_hbox), label, FALSE, FALSE, 0);

	mdd->position = (MarlinPositionSpinner *) marlin_position_spinner_new ();

	g_object_get (G_OBJECT (priv->view),
		      "cursor_position", &position,
		      NULL);

	gtk_spin_button_set_value (GTK_SPIN_BUTTON (mdd->position), (double) position);

	marlin_add_paired_relations (GTK_WIDGET (mdd->position), ATK_RELATION_LABELLED_BY,
				     label, ATK_RELATION_LABEL_FOR);

	display = marlin_gconf_get_int ("/apps/marlin/system-state/move-cursor-dialog-display");
	g_object_set (G_OBJECT (mdd->position),
		      "max_frames", num_frames,
		      "display_as", (MarlinPositionSpinnerDisplay) display,
		      "rate", rate,
		      NULL);
	
	gtk_entry_set_activates_default (GTK_ENTRY (mdd->position), TRUE);
	gtk_widget_show (GTK_WIDGET (mdd->position));
	gtk_box_pack_start (GTK_BOX (mdd->position_hbox), GTK_WIDGET (mdd->position), TRUE, TRUE, 0);

	label = marlin_position_spinner_label (mdd->position);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (mdd->position_hbox), label, FALSE, FALSE, 0);

	mdd->marker_hbox = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (mdd->marker_hbox);
	gtk_table_attach (GTK_TABLE (table), mdd->marker_hbox,
			  1, 3, 2, 3,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL | GTK_EXPAND,
			  0, 0);

	label = gtk_label_new_with_mnemonic (_("Mar_ker:"));
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (mdd->marker_hbox), label, FALSE, FALSE, 0);

	mdd->marker_combo = gtk_combo_box_new_text ();
	make_marker_combo_menu (GTK_COMBO_BOX (mdd->marker_combo), mdd->markers);

	gtk_widget_show (mdd->marker_combo);
	gtk_label_set_mnemonic_widget (GTK_LABEL (label), mdd->marker_combo);
	gtk_box_pack_start (GTK_BOX (mdd->marker_hbox), mdd->marker_combo,
			    TRUE, TRUE, 0);

	if (mdd->markers == NULL) {
		gtk_widget_hide (mdd->marker_hbox);
	}

	/* Do this here so that the marker_hbox will exist */
	make_move_combo_menu (GTK_COMBO_BOX (mdd->combo), mdd);
	
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       _("_Move cursor"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (editor),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (editor), GTK_RESPONSE_OK);
	
	g_signal_connect (editor, "response",
			  G_CALLBACK (move_cursor_response_cb), mdd);
	g_signal_connect (editor, "destroy",
			  G_CALLBACK (gtk_widget_destroyed), &priv->move_cursor_dialog);
	gtk_widget_show (editor);
}

void
marlin_window_replace (MarlinWindow *window)
{
	MarlinWindowPrivate *priv;
	MarlinProgram *program = marlin_program_get_default ();
	MarlinSample *clipboard;
	MarlinCoverage coverage;
	MarlinUndoContext *ctxt;
	guint64 position, end;
	
	priv = window->priv;

	clipboard = MARLIN_SAMPLE (marlin_program_get_clipboard (program));

	READ_LOCK (priv->selection->lock);
	if (priv->selection->selections != NULL) {
		MarlinSampleSubSelection *sub = priv->selection->selections->data;

		end = sub->end;
	} else {
		g_object_get (G_OBJECT (priv->sample),
			      "total_frames", &end,
			      NULL);

		/* end is one less than the total number of frames */
		end--;
	}
	READ_UNLOCK (priv->selection->lock);
	
	g_object_get (G_OBJECT (priv->view),
		      "cursor_position", &position,
		      NULL);

	ctxt = marlin_undo_manager_context_begin (priv->undo, _("Replace"));
	/* Replace is really a "mix with the dest at 0%" */
	marlin_sample_mix (priv->sample, clipboard, 0.0, MARLIN_INFINITE_DB,
			   position, end, TRUE, NULL, ctxt, NULL);
	marlin_undo_manager_context_end (priv->undo, ctxt);
}

struct _scale_closure {
	MarlinWindow *window;
	double scale;
};

static void
scale_selection_undo (gpointer data)
{
	struct _scale_closure *c = data;
	marlin_window_selection_scale (c->window, 1.0 / c->scale, NULL);
}

static void
scale_selection_redo (gpointer data)
{
	struct _scale_closure *c = data;
	marlin_window_selection_scale (c->window, c->scale, NULL);
}

static void
scale_selection_destroy (gpointer data)
{
	g_free (data);
}

void
marlin_window_selection_scale (MarlinWindow *window,
			       double scale,
			       MarlinUndoContext *ctxt)
{
	MarlinSampleSubSelection *sub;
	guint64 len, frames;
	MarlinUndoable *u;
	struct _scale_closure *c;

	if (ctxt) {
		c = g_new (struct _scale_closure, 1);
		c->window = window;
		c->scale = scale;
		
		u = marlin_undoable_new (scale_selection_undo,
					 scale_selection_redo,
					 scale_selection_destroy,
					 c);
		
		marlin_undo_context_add (ctxt, u);
	}

	g_object_get (G_OBJECT (window->priv->sample),
		      "total_frames", &frames,
		      NULL);
	WRITE_LOCK (window->priv->selection->lock);
	sub = window->priv->selection->selections->data;

	len = (sub->end - sub->start) * scale;
	
	sub->end = MIN (sub->start + len, frames - 1);
	WRITE_UNLOCK (window->priv->selection->lock);
	
	marlin_sample_selection_changed (window->priv->selection, FALSE);
}

struct _shift_selection {
	MarlinWindow *window;
};

static void
shift_left_undo (gpointer data)
{
	struct _shift_selection *c = data;
	marlin_window_selection_shift_right (c->window, NULL);
}

static void
shift_left_redo (gpointer data)
{
	struct _shift_selection *c = data;
	marlin_window_selection_shift_left (c->window, NULL);
}

static void
shift_left_destroy (gpointer data)
{
	g_free (data);
}

void
marlin_window_selection_shift_left (MarlinWindow *window,
				    MarlinUndoContext *ctxt)
{
	MarlinSampleSubSelection *sub;
	guint64 len;

	if (ctxt) {
		struct _shift_selection *c;
		MarlinUndoable *u;

		c = g_new (struct _shift_selection, 1);
		c->window = window;

		u = marlin_undoable_new (shift_left_undo,
					 shift_left_redo,
					 shift_left_destroy,
					 c);
		marlin_undo_context_add (ctxt, u);
	}

	WRITE_LOCK (window->priv->selection->lock);
	sub = window->priv->selection->selections->data;

	len = (sub->end - sub->start);
	if ( (gint64) (sub->start - len) < 0) {
		sub->start = (guint64) 0;
	} else {
		sub->start -= len;
	}
	sub->end = sub->start + len;
	WRITE_UNLOCK (window->priv->selection->lock);

	marlin_sample_selection_changed (window->priv->selection, FALSE);
}

static void
shift_right_undo (gpointer data)
{
	struct _shift_selection *c = data;
	marlin_window_selection_shift_left (c->window, NULL);
}

static void
shift_right_redo (gpointer data)
{
	struct _shift_selection *c = data;
	marlin_window_selection_shift_right (c->window, NULL);
}

static void
shift_right_destroy (gpointer data)
{
	g_free (data);
}

void
marlin_window_selection_shift_right (MarlinWindow *window,
				     MarlinUndoContext *ctxt)
{
	MarlinSampleSubSelection *sub;
	guint64 len, frames;

	if (ctxt) {
		struct _shift_selection *c;
		MarlinUndoable *u;

		c = g_new (struct _shift_selection, 1);
		c->window = window;

		u = marlin_undoable_new (shift_right_undo,
					 shift_right_redo,
					 shift_right_destroy,
					 c);
		marlin_undo_context_add (ctxt, u);
	}

	g_object_get (G_OBJECT (window->priv->sample),
		      "total_frames", &frames,
		      NULL);

	WRITE_LOCK (window->priv->selection->lock);
	sub = window->priv->selection->selections->data;

	len = (sub->end - sub->start);

	if ((sub->end + len) >= frames) {
		sub->end = frames - 1;
	} else {
		sub->end += len;
	}
	sub->start = sub->end - len;
	WRITE_UNLOCK (window->priv->selection->lock);

	marlin_sample_selection_changed (window->priv->selection, FALSE);
}

static void
properties_response (GtkWidget *dialog,
		     guint32 response_id,
		     MarlinWindow *window)
{
	window->priv->properties_dialog = NULL;
	gtk_widget_destroy (dialog);
}

static char *
get_string_from_taglist (GstTagList *taglist,
			 const char *tag)
{
	char *str;

	if (taglist == NULL) {
		return g_strdup (_("Unknown"));
	}
	
	if (gst_tag_list_get_string (taglist, tag, &str)) {
		return str;
	} else {
		return g_strdup (_("Unknown"));
	}
}

void
marlin_window_show_properties (MarlinWindow *window)
{
	MarlinWindowPrivate *priv = window->priv;
	GtkWidget *pd, *table, *vbox, *label, *sep;
	guint64 frames;
	guint sample_rate, num_channels;
	char *name, *filename, *title, *str;
	GstStructure *taglist;
	int trackno;
	
	if (priv->properties_dialog != NULL) {
		gtk_window_present (GTK_WINDOW (priv->properties_dialog));
		return;
	} 

	priv->properties_dialog = marlin_information_dialog_new (window->priv->sample, window);
	g_signal_connect (G_OBJECT (priv->properties_dialog), "destroy",
			  G_CALLBACK (gtk_widget_destroyed), &priv->properties_dialog);
	gtk_widget_show (priv->properties_dialog);
}

void
marlin_window_scroll_to_cursor (MarlinWindow *window)
{
	MarlinWindowPrivate *priv = window->priv;
	guint64 position;
	
	g_object_get (G_OBJECT (priv->view),
		      "cursor_position", &position,
		      NULL);
	marlin_sample_view_scroll_to (priv->view, position);
}

void
marlin_window_vzoom_in (MarlinWindow *window)
{
	g_return_if_fail (IS_MARLIN_WINDOW (window));
	
	marlin_sample_view_vzoom_in (MARLIN_SAMPLE_VIEW (window->priv->view));
}

void
marlin_window_vzoom_out (MarlinWindow *window)
{
	g_return_if_fail (IS_MARLIN_WINDOW (window));

	marlin_sample_view_vzoom_out (MARLIN_SAMPLE_VIEW (window->priv->view));
}

/* This should probably be in another file,
   actually most of the stuff in marlin-window.c should */
struct _xfade_dialog_data {
	MarlinWindow *window;
	GtkWidget *xfade;
};

static void
xfade_response_cb (GtkDialog *dialog,
		   guint response_id,
		   struct _xfade_dialog_data *xfd)
{
	MarlinCrossFaderFade *src_fade, *dest_fade;
	MarlinProgram *program = marlin_program_get_default ();
	MarlinSample *src, *dest;
	guint64 start_pos;
	
	switch (response_id) {
	case GTK_RESPONSE_OK:
		src = MARLIN_SAMPLE (marlin_program_get_clipboard (program));
		dest = xfd->window->priv->sample;

		g_object_get (G_OBJECT (xfd->xfade),
			      "src_fade", &src_fade,
			      "dest_fade", &dest_fade,
			      NULL);

		break;

	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-xfade-dialog");
		return;

	case GTK_RESPONSE_CANCEL:
	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (xfd);
}

void
marlin_window_xfade (MarlinWindow *window)
{
	struct _xfade_dialog_data *xfd;
	MarlinWindowPrivate *priv;
	MarlinProgram *program = marlin_program_get_default ();
	MarlinSample *clipboard;
	GtkWidget *dialog, *vbox;
	char *name, *title;
	guint64 start_pos;
	
	g_return_if_fail (IS_MARLIN_WINDOW (window));

	priv = window->priv;

	clipboard = MARLIN_SAMPLE (marlin_program_get_clipboard (program));
	
	if (priv->xfade_dialog != NULL) {
		gtk_window_present (GTK_WINDOW (priv->xfade_dialog));
		return;
	}
	
	xfd = g_new (struct _xfade_dialog_data, 1);
	xfd->window = window;

	g_object_get (G_OBJECT (priv->sample),
		      "name", &name,
		      NULL);

	g_object_get (G_OBJECT (priv->view),
		      "cursor_position", &start_pos,
		      NULL);
	
	dialog = priv->xfade_dialog = gtk_dialog_new ();
	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
	gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
	title = g_strdup_printf (_("Crossfade Sample: %s"), name);
	g_free (name);

	gtk_window_set_title (GTK_DIALOG (dialog), title);
	g_free (title);

	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), vbox);

	xfd->xfade = g_object_new (MARLIN_CROSS_FADER_TYPE,
				   "start_frame", start_pos,
				   NULL);
	gtk_widget_show (xfd->xfade);
	gtk_box_pack_start (GTK_BOX (vbox), xfd->xfade, FALSE, FALSE, 0);

	gtk_dialog_add_button (GTK_DIALOG (dialog),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (dialog),
			       _("Crossfade Samples"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (dialog),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

	g_signal_connect (dialog, "response",
			  G_CALLBACK (xfade_response_cb), xfd);
	g_signal_connect (dialog, "destroy",
			  G_CALLBACK (gtk_widget_destroyed),
			  &priv->xfade_dialog);
	
	gtk_widget_show (dialog);
}

void 
marlin_window_change_channels (MarlinWindow *window)
{
	MarlinWindowPrivate *priv = window->priv;
	int num_channels;

	if (priv->adjust_channels_dialog != NULL) {
		gtk_window_present (GTK_WINDOW (priv->adjust_channels_dialog));
		return;
	}

	g_object_get (G_OBJECT (window->priv->sample),
		      "channels", &num_channels,
		      NULL);

	if (num_channels == 2) {
		priv->adjust_channels_dialog = marlin_remove_channel_dialog_new (priv->sample,
										 GTK_WINDOW (window));
	} else if (num_channels == 1) {
		priv->adjust_channels_dialog = marlin_add_channel_dialog_new (priv->sample,
									      GTK_WINDOW (window));
	} else {
		g_warning ("Too many channels");
	}

	g_signal_connect (G_OBJECT (priv->adjust_channels_dialog), "destroy",
			  G_CALLBACK (gtk_widget_destroyed),
			  &priv->adjust_channels_dialog);
	gtk_widget_show (priv->adjust_channels_dialog);
}

void
marlin_window_split_sample (MarlinWindow *window)
{
	MarlinSample *original = window->priv->sample;
	MarlinSample *sample;
	MarlinMarkerModel *model;
	GList *markers, *m;
	guint64 last_pos = 0, num_frames;
	GError *error = NULL;

	g_object_get (G_OBJECT (original),
		      "markers", &model,
		      "total_frames", &num_frames,
		      NULL);

	g_object_get (G_OBJECT (model),
		      "markers", &markers,
		      NULL);

	if (markers == NULL) {
		return;
	}

	for (m = g_list_reverse (markers); m; m = m->next) {
		MarlinMarker *marker = m->data;

		sample = marlin_sample_new_from_sample_with_range (original,
								   last_pos,
								   marker->position - 1,
								   MARLIN_COVERAGE_BOTH,
								   &error);
		if (sample == NULL) {
			g_warning ("marlin_sample_new_from_sample_with_range failed");
			return;
		}

		g_object_set (G_OBJECT (sample),
			      "name", marker->name,
			      "dirty", TRUE,
			      NULL);
		marlin_open_window_with_sample (sample);

		last_pos = marker->position;
	}

	/* There is one more to split */
	sample = marlin_sample_new_from_sample_with_range (original, last_pos,
							   num_frames - 1,
							   MARLIN_COVERAGE_BOTH,
							   &error);
	if (sample == NULL) {
		g_warning ("marlin_sample_new_from_sample_with_range_failed");
		return;
	}

	g_object_set (G_OBJECT (sample),
		      "dirty", TRUE,
		      NULL);
	marlin_open_window_with_sample (sample);
}

void
marlin_window_undo (MarlinWindow *window)
{
	marlin_undo_manager_undo (window->priv->undo);
}

void
marlin_window_redo (MarlinWindow *window)
{
	marlin_undo_manager_redo (window->priv->undo);
}
